part of datagrid;

/// Handles the sizing for all the columns in the [SfDataGrid].
///
/// You can override any available methods in this class to calculate the
/// column width based on your requirement and set the instance to the
/// [SfDataGrid.columnSizer].
///
/// ``` dart
/// class CustomGridColumnSizer extends ColumnSizer {
///   // Calculate width for column when ColumnWidthMode is auto.
///   @override
///   double calculateAllCellsWidth(GridColumn column, [bool isAuto = false]) {
///     return super.calculateAllCellsWidth(column, isAuto);
///   }
///
///   // Calculate width for column when ColumnWidthMode is header.
///   @override
///   double calculateColumnHeaderWidth(GridColumn column, [bool setWidth = true]) {
///     return super.calculateColumnHeaderWidth(column, setWidth);
///   }
///
///   // Calculate width for column when ColumnWidthMode is cells.
///   @override
///   double calculateAllCellsExceptHeaderWidth(GridColumn column,
///       [bool setWidth = true]) {
///     return super.calculateAllCellsExceptHeaderWidth(column, setWidth);
///   }
/// }
///
/// final CustomGridColumnSizer _customGridColumnSizer = CustomGridColumnSizer();
///
/// @override
/// Widget build(BuildContext context) {
///   return Scaffold(
///     body: SfDataGrid(
///       source: _employeeDataSource,
///       columnSizer: _customGridColumnSizer,
///       columnWidthMode: ColumnWidthMode.auto,
///       columns: [
///         GridNumericColumn(mappingName: 'id', headerText: 'ID'),
///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
///         GridTextColumn(mappingName: 'designation', headerText: 'Designation'),
///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary')
///      ],
///     ),
///   );
/// }
/// ```
class ColumnSizer {
  int _textLength = 0;
  double _previousColumnWidth = 0.0;
  GridColumn _autoFillColumn;
  bool _isColumnSizerLoadedInitially = false;

  _DataGridStateDetails _dataGridStateDetails;

  void _initialRefresh(double availableWidth) {
    (this._dataGridStateDetails().container.columnWidths as _LineSizeCollection)
        .suspendUpdates();
    _refresh(availableWidth);
    (this._dataGridStateDetails().container.columnWidths as _LineSizeCollection)
        .resumeUpdates();
  }

  void _refresh(double availableWidth) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    var sizerColumns = dataGridSettings.columns
        .where((column) => column.columnWidthMode != ColumnWidthMode.none);

    if (dataGridSettings.columnWidthMode != null || sizerColumns.length > 0) {
      this._sizerColumnWidth(availableWidth);
    }
    dataGridSettings.container.updateScrollBars();
  }

  void _sizerColumnWidth(double viewPortWidth) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    double totalColumnSize = 0.0;
    var calculatedColumns = new List<GridColumn>();

    _autoFillColumn = _getColumnToFill();

    // Hide Hidden columns
    var hiddenColumns =
        dataGridSettings.columns.where((column) => column.isHidden);
    for (var column in hiddenColumns) {
      var index = _GridIndexResolver.resolveToScrollColumnIndex(
          dataGridSettings, dataGridSettings.columns.indexOf(column));
      dataGridSettings.container.columnWidths.setHidden(index, index, true);
      calculatedColumns.add(column);
    }

    // Set width based on Column.Width
    var widthColumns = dataGridSettings.columns
        .skipWhile((column) => column.isHidden)
        .where((column) => !(column.width).isNaN);
    for (var column in widthColumns) {
      totalColumnSize += _setColumnWidth(column, column.width);
      calculatedColumns.add(column);
    }

    // Set width based on widget column
    var widgetColumns = dataGridSettings.columns
        .skipWhile((column) => (column.isHidden ||
            !(column.columnWidthMode == ColumnWidthMode.auto ||
                column.columnWidthMode == ColumnWidthMode.cells)))
        .where((column) => column._cellType == "Widget" && column.width.isNaN);
    for (var column in widgetColumns) {
      totalColumnSize +=
          _setColumnWidth(column, dataGridSettings.defaultColumnWidth);
      calculatedColumns.add(column);
    }

    // Set width based on cells
    var cellColumns = dataGridSettings.columns
        .skipWhile(
            (column) => (column.isHidden || widgetColumns.contains(column)))
        .where((column) =>
            column.columnWidthMode == ColumnWidthMode.cells &&
            column.width.isNaN);
    for (var column in cellColumns) {
      if (column._autoWidth.isNaN) {
        var columnwidth = _getWidthBasedOnColumn(column, ColumnWidthMode.cells);
        totalColumnSize += columnwidth;
        _setAutoWidth(column, columnwidth);
      } else
        totalColumnSize += _setColumnWidth(column, column._autoWidth);
      calculatedColumns.add(column);
    }

    // Set width based on header
    var headerColumns = dataGridSettings.columns
        .skipWhile((column) => (column.isHidden))
        .where((column) =>
            column.columnWidthMode == ColumnWidthMode.header &&
            column.width.isNaN);
    for (GridColumn column in headerColumns) {
      totalColumnSize += _getWidthBasedOnColumn(column, ColumnWidthMode.header);
      calculatedColumns.add(column);
    }

    // Set width based on auto and lastColumnFill
    var lastColumn =
        dataGridSettings.columns.lastWhere((column) => !column.isHidden);
    List<GridColumn> autoColumns = dataGridSettings.columns
        .where((column) =>
            column.columnWidthMode == ColumnWidthMode.auto &&
            !(column.isHidden || widgetColumns.contains(column)) &&
            (column.width).isNaN)
        .toList();
    List<GridColumn> lastColumnFill = dataGridSettings.columns
        .skipWhile((column) => calculatedColumns.contains(column))
        .where((col) =>
            (col.columnWidthMode == ColumnWidthMode.lastColumnFill) &&
            !_isLastFillColum(col))
        .toList();
    autoColumns = (autoColumns + lastColumnFill).toSet().toList();

    for (var column in autoColumns) {
      if ((column._autoWidth).isNaN) {
        var columnWidth = _getWidthBasedOnColumn(column, ColumnWidthMode.auto);
        totalColumnSize += columnWidth;
        _setAutoWidth(column, columnWidth);
      } else
        totalColumnSize += _setColumnWidth(column, column._autoWidth);
      calculatedColumns.add(column);
    }
    _setWidthBasedOnGrid(
        totalColumnSize, calculatedColumns, lastColumn, viewPortWidth);
    _autoFillColumn = null;
  }

  GridColumn _getColumnToFill() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    var column = dataGridSettings.columns.lastWhere(
        (c) =>
            !c.isHidden &&
            (c.width).isNaN &&
            (c.columnWidthMode == ColumnWidthMode.lastColumnFill),
        orElse: () => null);
    if (column != null)
      return column;
    else {
      if (dataGridSettings.columnWidthMode == ColumnWidthMode.lastColumnFill) {
        var lastColumn = dataGridSettings.columns
            .lastWhere((c) => !c.isHidden && (c.width).isNaN);
        if (lastColumn == null) return null;
        if (lastColumn.columnWidthMode == ColumnWidthMode.none)
          return lastColumn;
      }
    }
    return null;
  }

  void _setWidthBasedOnGrid(
      double totalColumnSize,
      List<GridColumn> calculatedColumns,
      GridColumn lastColumn,
      double viewPortWidth) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    for (var column in dataGridSettings.columns) {
      if (calculatedColumns.contains(column)) continue;

      if (column._cellType == 'Widget') {
        if (dataGridSettings.columnWidthMode == ColumnWidthMode.auto ||
            dataGridSettings.columnWidthMode == ColumnWidthMode.cells) {
          totalColumnSize +=
              _setColumnWidth(column, dataGridSettings.defaultColumnWidth);
          calculatedColumns.add(column);
          continue;
        }
      }

      if (column.columnWidthMode == ColumnWidthMode.fill ||
          _isLastFillColum(column)) continue;

      switch (dataGridSettings.columnWidthMode) {
        case ColumnWidthMode.cells:
          if (column._autoWidth.isNaN) {
            var columnWidth =
                _getWidthBasedOnColumn(column, ColumnWidthMode.cells);
            totalColumnSize += columnWidth;
            _setAutoWidth(column, columnWidth);
          } else
            totalColumnSize += _setColumnWidth(column, column._autoWidth);
          calculatedColumns.add(column);
          break;
        case ColumnWidthMode.header:
          totalColumnSize +=
              _getWidthBasedOnColumn(column, ColumnWidthMode.header);
          calculatedColumns.add(column);
          break;
        case ColumnWidthMode.auto:
        case ColumnWidthMode.lastColumnFill:
          if ((column._autoWidth).isNaN) {
            var columnWidth =
                _getWidthBasedOnColumn(column, ColumnWidthMode.auto);
            totalColumnSize += columnWidth;
            _setAutoWidth(column, columnWidth);
          } else
            totalColumnSize += _setColumnWidth(column, column._autoWidth);
          calculatedColumns.add(column);
          break;
        case ColumnWidthMode.none:
          if (!column.isHidden) {
            totalColumnSize += _setColumnWidth(column,
                dataGridSettings.container.columnWidths.defaultLineSize);
            calculatedColumns.add(column);
          }
          break;
        default:
          break;
      }
    }

    List<GridColumn> remainingColumns = [];

    for (var column in dataGridSettings.columns) {
      if (!calculatedColumns.contains(column)) {
        remainingColumns.add(column);
      }
    }

    double remainingColumnWidths = viewPortWidth - totalColumnSize;

    if (remainingColumnWidths > 0 &&
        (totalColumnSize != 0 ||
            (totalColumnSize == 0 && remainingColumns.length == 1) ||
            (dataGridSettings.columns.any(
                    (col) => col.columnWidthMode == ColumnWidthMode.fill) ||
                dataGridSettings.columnWidthMode == ColumnWidthMode.fill)))
      _setFillWidth(remainingColumnWidths, remainingColumns);
    else
      _setRemainingColumnsWidth(remainingColumns);
  }

  void _setFillWidth(
      double remainingColumnWidth, List<GridColumn> remainingColumns) {
    var removedColumns = List<GridColumn>();
    var columns = remainingColumns.toList();
    var totalRemainingFillValue = remainingColumnWidth;

    double removedWidth = 0;
    GridColumn fillColumn;
    bool isRemoved;
    while (columns.length > 0) {
      isRemoved = false;
      removedWidth = 0;
      double fillWidth =
          (totalRemainingFillValue / columns.length).floorToDouble();
      var column = columns.first;
      if (column == _autoFillColumn &&
          (column.columnWidthMode == ColumnWidthMode.lastColumnFill ||
              (_dataGridStateDetails().columnWidthMode ==
                  ColumnWidthMode.lastColumnFill))) {
        columns.remove(column);
        fillColumn = column;
        continue;
      }
      double computedWidth = _setColumnWidth(column, fillWidth);
      if (fillWidth != computedWidth && fillWidth > 0) {
        isRemoved = true;
        columns.remove(column);
        for (var removedColumn in removedColumns) {
          if (!columns.contains(removedColumn)) {
            removedWidth += removedColumn.actualWidth;
            columns.add(removedColumn);
          }
        }
        removedColumns.clear();
        totalRemainingFillValue += removedWidth;
      }
      column._actualWidth = computedWidth;
      totalRemainingFillValue = totalRemainingFillValue - computedWidth;
      if (!isRemoved) {
        columns.remove(column);
        if (!removedColumns.contains(column)) removedColumns.add(column);
      }
    }

    if (fillColumn != null) {
      double columnWidth = 0;
      if ((fillColumn._autoWidth).isNaN) {
        columnWidth = _getWidthBasedOnColumn(fillColumn, ColumnWidthMode.auto);
        _setAutoWidth(fillColumn, columnWidth);
      } else
        columnWidth = fillColumn._autoWidth;
      if (totalRemainingFillValue < columnWidth)
        _setColumnWidth(fillColumn, columnWidth);
      else
        _setColumnWidth(fillColumn, totalRemainingFillValue);
    }
  }

  void _setRemainingColumnsWidth(List<GridColumn> remainingColumns) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    for (var column in remainingColumns) {
      if (_isLastFillColum(column)) {
        if (column.columnWidthMode == ColumnWidthMode.lastColumnFill ||
            (dataGridSettings.columnWidthMode ==
                ColumnWidthMode.lastColumnFill))
          _getWidthBasedOnColumn(column, ColumnWidthMode.auto);
        else
          _setNoneWidth(
              column, dataGridSettings.container.columnWidths.defaultLineSize);
      } else if (!_isFillColumn(column)) {
        _setNoneWidth(
            column, dataGridSettings.container.columnWidths.defaultLineSize);
      }
    }
  }

  bool _isFillColumn(GridColumn column) {
    return !(column.width).isNaN
        ? false
        : column.columnWidthMode == ColumnWidthMode.none
            ? this._dataGridStateDetails().columnWidthMode ==
                ColumnWidthMode.fill
            : column.columnWidthMode == ColumnWidthMode.fill;
  }

  double _setNoneWidth(GridColumn column, double width) {
    return _setColumnWidth(column, width);
  }

  bool _isLastFillColum(GridColumn column) {
    if (column == _autoFillColumn) return true;
    return false;
  }

  void _setAutoWidth(GridColumn column, double width) {
    if (column != null) column._autoWidth = width;
  }

  double _getWidthBasedOnColumn(
      GridColumn column, ColumnWidthMode columnWidthMode) {
    double width;
    switch (columnWidthMode) {
      case ColumnWidthMode.cells:
        width = calculateAllCellsExceptHeaderWidth(column);
        return _setColumnWidth(column, width);
      case ColumnWidthMode.header:
        width = calculateColumnHeaderWidth(column);
        return _setColumnWidth(column, width);
      case ColumnWidthMode.auto:
        width = calculateAllCellsWidth(column, true);
        return _setColumnWidth(column, width);
      default:
        break;
    }
    return 0.0;
  }

  /// Calculates the width for the column to fit the content including the
  /// header cell when [SfDataGrid.columnWidthMode] is [ColumnWidthMode.auto].
  double calculateAllCellsWidth(GridColumn column, [bool isAuto = false]) {
    double headerWidth = calculateColumnHeaderWidth(column, false);
    double cellWidth = calculateAllCellsExceptHeaderWidth(column, false);
    double width;
    if (cellWidth > headerWidth)
      width = cellWidth;
    else
      width = headerWidth;
    if (isAuto) return width;
    return _getColumnWidth(column, width);
  }

  /// Calculates the width for the column based on the header text when
  /// [SfDataGrid.columnWidthMode] is [ColumnWidthMode.header].
  double calculateColumnHeaderWidth(GridColumn column, [bool setWidth = true]) {
    double width = this._getHeaderCellWidth(column).roundToDouble();
    if (setWidth) column._actualWidth = width;
    return width;
  }

  /// Calculates the width for the column to fit the content in cells except
  /// header cell when [SfDataGrid.columnWidthMode] is [ColumnWidthMode.cells].
  double calculateAllCellsExceptHeaderWidth(GridColumn column,
      [bool setWidth = true]) {
    if (setWidth) {
      column._actualWidth = this._getCellWidth(column);
      return column.actualWidth;
    } else
      return this._getCellWidth(column);
  }

  double _setColumnWidth(GridColumn column, double columnWidth) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    var columnIndex = dataGridSettings.columns.indexOf(column);
    var width = _getColumnWidth(column, columnWidth);
    column._actualWidth = width;
    dataGridSettings.container.columnWidths[columnIndex] = column.actualWidth;
    return width;
  }

  double _getColumnWidth(GridColumn column, double columnWidth) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    var columnIndex = dataGridSettings.columns.indexOf(column);
    if (column.width < column.actualWidth) {
      return columnWidth;
    }
    double width = dataGridSettings.container.columnWidths[columnIndex][0];

    var resultWidth = _checkWidthConstraints(column, columnWidth, width);
    return resultWidth;
  }

  double _checkWidthConstraints(
      GridColumn column, double width, double columnWidth) {
    if (!(column.minimumWidth).isNaN || !(column.maximumWidth).isNaN) {
      if (!(column.maximumWidth).isNaN) {
        if (!(width).isNaN && column.maximumWidth > width) {
          columnWidth = width;
        } else {
          columnWidth = column.maximumWidth;
        }
      }

      if (!(column.minimumWidth).isNaN) {
        if (!(width).isNaN && column.minimumWidth < width) {
          if (width > column.maximumWidth)
            columnWidth = column.maximumWidth;
          else
            columnWidth = width;
        } else {
          columnWidth = column.minimumWidth;
        }
      }
    } else {
      if (!(width).isNaN) columnWidth = width;
    }
    return columnWidth;
  }

  double _getCellWidth(GridColumn column) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    double resultWidth = 0;
    var recordCount = dataGridSettings.source.dataSource.length;
    if (recordCount == 0) return double.nan;
    _textLength = 0;
    _previousColumnWidth = 0;
    int stringLength = 0;
    var isInDefaultMode = dataGridSettings.columnWidthCalculationMode ==
        ColumnWidthCalculationMode.textSize;
    Object index;
    int firstVisibleIndex, lastVisibleIndex;

    if (dataGridSettings.columnWidthCalculationRange ==
        ColumnWidthCalculationRange.visibleRows) {
      var visibleLines =
          dataGridSettings.container.scrollRows.getVisibleLines();
      firstVisibleIndex =
          visibleLines[visibleLines.firstBodyVisibleIndex].lineIndex - 1;
      lastVisibleIndex =
          dataGridSettings.container.scrollRows.lastBodyVisibleLineIndex - 1;
    } else {
      firstVisibleIndex = 0;
      lastVisibleIndex = dataGridSettings.source.dataSource.length - 1;
    }
    for (int rowIndex = firstVisibleIndex;
        rowIndex <= lastVisibleIndex;
        rowIndex++) {
      if (isInDefaultMode) {
        var textWidth = this.getCellWidth(column, rowIndex);
        if (textWidth.toString().isEmpty) continue;

        if (resultWidth < textWidth) resultWidth = textWidth;
      } else {
        var currentRowIndex =
            _GridIndexResolver.resolveToRecordIndex(dataGridSettings, rowIndex);
        String text = this._getDisplayText(currentRowIndex, column);
        if (text.length >= stringLength) {
          stringLength = text.length;
          index = rowIndex;
        }
      }
    }

    if (!isInDefaultMode) {
      var textWidth = this.getCellWidth(column, index);
      resultWidth = textWidth;
    }

    _textLength = 0;
    _previousColumnWidth = 0;
    return resultWidth;
  }

  /// Gets the width of the cell to calculate the width of the specified column.
  double getCellWidth(GridColumn column, int rowIndex) {
    double textWidth = 0.0;
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    var currentRowIndex =
        _GridIndexResolver.resolveToRecordIndex(dataGridSettings, rowIndex);
    var formattedText = this._getDisplayText(currentRowIndex, column);
    if (formattedText != null && formattedText.length >= _textLength ||
        _previousColumnWidth >= column.actualWidth) {
      textWidth = _measureTextWidth(formattedText, column, rowIndex);
      _textLength = formattedText.length;
      _previousColumnWidth = column.actualWidth;
    }

    return textWidth.roundToDouble();
  }

  /// Gets the height of the cell to calculate the width of the specified
  /// column.
  double getCellHeight(GridColumn column, int rowIndex) {
    double textHeight = 0.0;
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    var currentRowIndex =
        _GridIndexResolver.resolveToRecordIndex(dataGridSettings, rowIndex);
    var formattedText = this._getDisplayText(currentRowIndex, column);
    if (formattedText != null && formattedText.length >= _textLength ||
        _previousColumnWidth >= column.actualWidth) {
      textHeight = _measureTextHeight(formattedText, column, rowIndex);
      _textLength = formattedText.length;
      _previousColumnWidth = column.actualWidth;
    }

    return textHeight.roundToDouble();
  }

  /// Gets the row height to fit the row based on the content.
  ///
  /// The following code snippet shows how to use [getAutoRowHeight] method,
  ///
  /// ``` dart
  /// final ColumnSizer _columnSizer = ColumnSizer();
  ///
  /// @override
  ///  Widget build(BuildContext context) {
  ///  return Scaffold(
  ///       body: SfDataGrid(
  ///           source: _employeeDatasource,
  ///           columnSizer: _columnSizer,
  ///           onQueryRowHeight: (rowIndex) {
  ///             double height = _columnSizer.getAutoRowHeight(rowIndex);
  ///             return height;
  ///           },
  ///           columns: <GridColumn>[
  ///         GridTextColumn(mappingName: 'id')
  ///           ..softWrap = true
  ///           ..overflow = TextOverflow.clip
  ///           ..headerText = 'ID',
  ///         GridTextColumn(mappingName: 'contactName')
  ///           ..softWrap = true
  ///           ..overflow = TextOverflow.clip
  ///           ..headerText = 'Contact Name',
  ///         GridTextColumn(mappingName: 'companyName')
  ///           ..softWrap = true
  ///           ..overflow = TextOverflow.clip
  ///           ..headerText = 'Company Name',
  ///         GridTextColumn(mappingName: 'city')
  ///           ..softWrap = true
  ///           ..overflow = TextOverflow.clip
  ///           ..headerText = 'City',
  ///         GridTextColumn(mappingName: 'address')
  ///           ..softWrap = true
  ///           ..overflow = TextOverflow.clip
  ///           ..headerText = 'Address',
  ///         GridTextColumn(mappingName: 'designation')
  ///          ..softWrap = true
  ///           ..overflow = TextOverflow.clip
  ///           ..headerText = 'Designation',
  ///         GridTextColumn(mappingName: 'country')
  ///           ..softWrap = true
  ///           ..overflow = TextOverflow.clip
  ///           ..headerText = 'Country',
  ///       ]));
  /// }
  /// ```
  double getAutoRowHeight(int rowIndex,
      {bool canIncludeHiddenColumns, List<String> excludedColumns}) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    double resultHeight = -1;
    GridColumn gridColumn;
    int stringLength = 0;
    var isInDefaultMode = dataGridSettings.columnWidthCalculationMode ==
        ColumnWidthCalculationMode.textSize;
    for (int columnIndex = 0;
        columnIndex < dataGridSettings.columns.length;
        columnIndex++) {
      var column = dataGridSettings.columns[columnIndex];
      if (column.isHidden && !canIncludeHiddenColumns) continue;

      if (excludedColumns != null) {
        if (excludedColumns.contains(column.mappingName)) continue;
      }

      var recordCount = dataGridSettings.source.dataSource.length;
      if (recordCount == 0) return double.nan;
      if (isInDefaultMode) {
        var textHeight = getCellHeight(column, rowIndex);
        if (textHeight.toString().isEmpty) continue;

        if (resultHeight < textHeight) resultHeight = textHeight;
      } else {
        var text = this._getDisplayText(rowIndex, column);
        if (text != null && text.length >= stringLength) {
          stringLength = text.length;
          gridColumn = column;
        }
      }
    }

    if (!isInDefaultMode) {
      var textHeight = this.getCellHeight(gridColumn, rowIndex);
      resultHeight = textHeight;
    }
    _textLength = 0;
    _previousColumnWidth = 0;
    return resultHeight;
  }

  double _getHeaderCellWidth(GridColumn column) {
    var rowCount = this._dataGridStateDetails().headerLineCount;
    double resultWidth = 0;
    for (int rowIndex = 0; rowIndex < rowCount; rowIndex++) {
      var text = column.headerText ?? column.mappingName;
      if (text.isEmpty || text == null) return 0;
      var textWidth = _measureHeaderTextWidth(text, column, rowIndex);
      var width = textWidth;
      if (resultWidth < width) resultWidth = width;
    }
    return resultWidth;
  }

  String _getDisplayText(int rowIndex, GridColumn column) {
    var text = this
        ._dataGridStateDetails()
        .source
        .getCellValue(rowIndex, column.mappingName);
    String formattedText = column.getFormattedValue(text);
    if (formattedText != null) {
      return formattedText;
    }
    return '';
  }

  double _measureHeaderTextWidth(
      String displayText, GridColumn column, int rowIndex) {
    TextStyle headerTextStyle = column.headerStyle?.textStyle ??
        this._dataGridStateDetails().dataGridThemeData.headerStyle?.textStyle;
    return _calculateTextSize(
            column, displayText, headerTextStyle, double.infinity, rowIndex)
        .width;
  }

  double _measureTextWidth(
      String displayText, GridColumn column, int rowIndex) {
    TextStyle cellTextStyle = column.cellStyle?.textStyle ??
        this._dataGridStateDetails().dataGridThemeData.cellStyle?.textStyle;
    return _calculateTextSize(
            column, displayText, cellTextStyle, double.infinity, rowIndex)
        .width;
  }

  double _measureTextHeight(
      String displayText, GridColumn column, int rowIndex) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    TextStyle cellTextStyle = column.cellStyle?.textStyle ??
        dataGridSettings.dataGridThemeData.cellStyle?.textStyle;
    var columnIndex = dataGridSettings.columns.indexOf(column);
    int scrollColumnIndex = _GridIndexResolver.resolveToScrollColumnIndex(
        dataGridSettings, columnIndex);
    double columnWidth =
        dataGridSettings.container.columnWidths[scrollColumnIndex][0];
    var gridLinesVisibility = dataGridSettings.gridLinesVisibility;
    Size gridLineStrokeWidthSize = this._getGridLineStrokeWidthSize(
        gridLinesVisibility, rowIndex, columnIndex);
    // Removing padding and stroke width values also to the column hight calculation
    double actualColumnWidth = columnWidth -
        (column.padding.horizontal + gridLineStrokeWidthSize.width);
    return _calculateTextSize(
            column, displayText, cellTextStyle, actualColumnWidth, rowIndex)
        .height;
  }

  Size _getGridLineStrokeWidthSize(
      GridLinesVisibility gridLinesVisibility, int rowIndex, int columnIndex) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    Size gridLineStrokeWidthSize = Size.zero;
    double gridLineStrokeWidth =
        dataGridSettings.dataGridThemeData?.gridLineStrokeWidth;
    if (gridLinesVisibility == GridLinesVisibility.both) {
      if (columnIndex == 0 && rowIndex == 0) {
        // Adding border line stroke also to the column width calculation
        gridLineStrokeWidth *= 2;
        gridLineStrokeWidthSize = Size(
            (gridLineStrokeWidthSize.width + gridLineStrokeWidth),
            (gridLineStrokeWidthSize.height + gridLineStrokeWidth));
      }
      if (rowIndex == 0 && columnIndex > 0) {
        gridLineStrokeWidthSize = Size(
            (gridLineStrokeWidthSize.width + gridLineStrokeWidth),
            (gridLineStrokeWidthSize.height + (2 * gridLineStrokeWidth)));
      }
      if (columnIndex == 0 && rowIndex > 0) {
        gridLineStrokeWidthSize = Size(
            (gridLineStrokeWidthSize.width + (2 * gridLineStrokeWidth)),
            (gridLineStrokeWidthSize.height + gridLineStrokeWidth));
      }
      if (columnIndex > 0 && rowIndex > 0) {
        gridLineStrokeWidthSize = Size(
            (gridLineStrokeWidthSize.width + gridLineStrokeWidth),
            (gridLineStrokeWidthSize.height + gridLineStrokeWidth));
      }
    } else if (gridLinesVisibility == GridLinesVisibility.vertical) {
      if (columnIndex == 0)
        // Adding border line stroke also to the column width calculation
        gridLineStrokeWidth *= 2;
      gridLineStrokeWidthSize = Size(
          (gridLineStrokeWidthSize.width + gridLineStrokeWidth),
          gridLineStrokeWidthSize.height);
    } else if (gridLinesVisibility == GridLinesVisibility.horizontal) {
      if (rowIndex == 0)
        // Adding border line stroke also to the column width calculation
        gridLineStrokeWidth *= 2;
      gridLineStrokeWidthSize = Size(gridLineStrokeWidthSize.width,
          (gridLineStrokeWidthSize.height + gridLineStrokeWidth));
    }
    return gridLineStrokeWidthSize;
  }

  Size _calculateTextSize(GridColumn column, String text, TextStyle style,
      double width, int rowIndex) {
    Size textSize = Size.zero;
    var columnIndex = this._dataGridStateDetails().columns.indexOf(column);
    var gridLinesVisibility = this._dataGridStateDetails().gridLinesVisibility;
    Size gridLineStrokeWidthSize = this._getGridLineStrokeWidthSize(
        gridLinesVisibility, rowIndex, columnIndex);
    final TextPainter textPainter = TextPainter(
        text: TextSpan(
          text: text,
          style: style,
        ),
        maxLines: column.maxLines,
        textDirection: TextDirection.ltr)
      ..layout(minWidth: 0, maxWidth: width);
    textSize = textPainter.size;

    if (column.padding != null) {
      textSize = Size((textSize.width + column.padding.horizontal),
          (textSize.height + column.padding.vertical));
    }
    return Size((textSize.width + gridLineStrokeWidthSize.width),
        (textSize.height + gridLineStrokeWidthSize.height));
  }

  @override
  bool operator ==(other) =>
      identical(this, other) ||
      other is ColumnSizer && runtimeType == other.runtimeType;

  @override
  int get hashCode => super.hashCode;
}
