part of datagrid;

/// Provides the base functionalities for all the column types in [SfDataGrid].
class GridColumn {
  double _autoWidth = double.nan;

  /// How the column widths are determined.
  ///
  /// This takes higher priority than [SfDataGrid.columnWidthMode].
  ///
  /// Defaults to [ColumnWidthMode.none]
  ///
  /// Also refer [ColumnWidthMode]
  ColumnWidthMode get columnWidthMode => this._columnWidthMode;
  ColumnWidthMode _columnWidthMode = ColumnWidthMode.none;

  /// How the column widths are determined.
  ///
  /// This takes higher priority than [SfDataGrid.columnWidthMode].
  ///
  /// Defaults to [ColumnWidthMode.none]
  ///
  /// Also refer [ColumnWidthMode]
  set columnWidthMode(ColumnWidthMode newValue) =>
      this._columnWidthMode = newValue;

  /// The name to map the data member in the underlying data object of
  /// datasource.
  ///
  /// Defaults to null
  String get mappingName => this._mappingName;
  String _mappingName;

  /// The name to map the data member in the underlying data object of
  ///  datasource.
  ///
  /// If [headerText] is not set, [mappingName] will be displayed as text
  /// in header cell.
  ///
  /// Defaults to null
  set mappingName(String newValue) {
    if (this._mappingName == newValue) return;

    this._mappingName = newValue;
  }

  /// The cell type of the column which denotes renderer associated with column.
  String get cellType => this._cellType;
  String _cellType;

  /// How the text in cells except header cell is aligned.
  ///
  /// Defaults to null
  Alignment get textAlignment => this._textAlignment;
  Alignment _textAlignment = Alignment.centerLeft;

  /// How the text in cells except header cell is aligned.
  ///
  /// Defaults to null
  set textAlignment(Alignment value) => this._textAlignment = value;

  /// How the text in header cell is aligned.
  ///
  /// Defaults to null
  Alignment get headerTextAlignment => this._headerTextAlignment;
  Alignment _headerTextAlignment = Alignment.center;

  /// How the text in header cell is aligned.
  ///
  /// Defaults to null
  set headerTextAlignment(Alignment value) => this._headerTextAlignment = value;

  /// An optional maximum number of lines for the text to span, wrapping
  /// if necessary in cells except header cell.
  ///
  /// Defaults to null
  int get maxLines => this._maxLines;
  int _maxLines;

  /// An optional maximum number of lines for the text to span, wrapping
  /// if necessary in cells except header cell.
  ///
  /// Defaults to null
  set maxLines(int newValue) => this._maxLines = newValue;

  /// How visual overflow should be handled in cells except header cell.
  ///
  /// Defaults to null
  ///
  /// Also refer [TextOverflow]
  TextOverflow get overflow => this._overflow;
  TextOverflow _overflow;

  /// How visual overflow should be handled in cells except header cell.
  ///
  /// Defaults to null
  ///
  /// Also refer [TextOverflow]
  set overflow(TextOverflow newValue) => this._overflow = newValue;

  /// Whether the text should break at soft line breaks.
  ///
  /// Defaults to false
  bool get softWrap => this._softWrap;
  bool _softWrap = false;

  /// Whether the text should break at soft line breaks.
  ///
  /// Defaults to false
  set softWrap(bool value) => _softWrap = value;

  /// The actual display width of the column when auto fitted based on
  /// [SfDataGrid.columnWidthMode] or [columnWidthMode].
  ///
  /// Defaults to [double.nan]
  double get actualWidth => this._actualWidth;
  double _actualWidth = double.nan;

  /// The text which displays in header cell.
  ///
  /// Defaults to null.
  String get headerText => this._headerText;
  String _headerText;

  /// The text which displays in header cell.
  ///
  /// Defaults to null.
  set headerText(String newValue) => this._headerText = newValue;

  /// The minimum width of the column.
  ///
  /// The column width could not be set or auto sized lesser than the
  /// [minimumWidth]
  ///
  /// Defaults to [double.nan]
  double get minimumWidth => this._minimumWidth;
  double _minimumWidth = double.nan;

  /// The minimum width of the column.
  ///
  /// The column width could not be set or auto sized lesser than the
  /// [minimumWidth]
  ///
  /// Defaults to [double.nan]
  set minimumWidth(double newValue) => this._minimumWidth = newValue;

  /// The maximum width of the column.
  ///
  /// The column width could not be set or auto sized greater than the
  /// [maximumWidth]
  ///
  /// Defaults to [double.nan]
  double get maximumWidth => this._maximumWidth;
  double _maximumWidth = double.nan;

  /// The maximum width of the column.
  ///
  /// The column width could not be set or auto sized greater than the
  /// [maximumWidth]
  ///
  /// Defaults to [double.nan]
  set maximumWidth(double newValue) => this._maximumWidth = newValue;

  /// The width of the column.
  ///
  /// If value is lesser than [minimumWidth], then [minimumWidth]
  /// is set to [width].
  /// Otherwise, If value is greater than [maximumWidth], then the
  /// [maximumWidth] is set to [width].
  ///
  /// Defaults to [double.nan]
  double get width => this._width;
  double _width = double.nan;

  /// The width of the column.
  ///
  /// If value is lesser than [minimumWidth], then [minimumWidth]
  /// is set to [width].
  /// Otherwise, If value is greater than [maximumWidth], then the
  /// [maximumWidth] is set to [width].
  ///
  /// Defaults to [double.nan]
  set width(double newValue) => this._width = newValue;

  /// The amount of space between the contents of the cell and the cell's
  /// border.
  ///
  /// This is also applicable for header cells.
  ///
  /// Defaults to 16
  EdgeInsetsGeometry get padding => this._padding;
  EdgeInsetsGeometry _padding = EdgeInsets.all(16);

  /// The amount of space between the contents of the cell and the cell's
  /// border.
  ///
  /// This is also applicable for header cells.
  ///
  /// Defaults to 16
  set padding(EdgeInsetsGeometry newValue) => this._padding = newValue;

  /// The style of the header cell in the column.
  ///
  /// Defaults to null
  ///
  /// ``` dart
  /// @override
  /// Widget build(BuildContext context) {
  ///   return Scaffold(
  ///     body: SfDataGrid(
  ///       source: _employeeDataSource,
  ///       columns: [
  ///         GridNumericColumn(mappingName: 'id', headerText: 'ID')
  ///           ..headerStyle = DataGridHeaderCellStyle(
  ///              textStyle:
  ///                TextStyle(color: Colors.red, fontWeight: FontWeight.bold)),
  ///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
  ///         GridTextColumn(mappingName: 'designation',
  ///             headerText: 'Designation'),
  ///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary')
  ///       ],
  ///     ),
  ///   );
  /// }
  ///
  /// ```
  DataGridHeaderCellStyle get headerStyle => this._headerStyle;
  DataGridHeaderCellStyle _headerStyle;

  /// The style of the header cell in the column.
  ///
  /// Defaults to null
  ///
  /// ``` dart
  /// @override
  /// Widget build(BuildContext context) {
  ///   return Scaffold(
  ///     body: SfDataGrid(
  ///       source: _employeeDataSource,
  ///       columns: [
  ///         GridNumericColumn(mappingName: 'id', headerText: 'ID')
  ///           ..headerStyle = DataGridHeaderCellStyle(
  ///              textStyle:
  ///                TextStyle(color: Colors.red, fontWeight: FontWeight.bold)),
  ///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
  ///         GridTextColumn(mappingName: 'designation',
  ///             headerText: 'Designation'),
  ///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary')
  ///       ],
  ///     ),
  ///   );
  /// }
  ///
  /// ```
  set headerStyle(DataGridHeaderCellStyle newHeaderStyle) =>
      this._headerStyle = newHeaderStyle;

  /// The style of the cells in the column except header cell.
  ///
  /// Defaults to null
  ///
  /// ``` dart
  /// @override
  /// Widget build(BuildContext context) {
  ///   return Scaffold(
  ///     body: SfDataGrid(
  ///       source: _employeeDataSource,
  ///       columns: [
  ///         GridNumericColumn(mappingName: 'id', headerText: 'ID')
  ///           ..cellStyle = DataGridCellStyle(
  ///             textStyle: TextStyle(color: Colors.red)),
  ///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
  ///         GridTextColumn(mappingName: 'designation',
  ///             headerText: 'Designation'),
  ///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary')
  ///       ],
  ///     ),
  ///   );
  /// }
  ///
  /// ```
  DataGridCellStyle get cellStyle => this._cellStyle;
  DataGridCellStyle _cellStyle;

  /// The style of the cells in the column except header cell.
  ///
  /// Defaults to null
  ///
  /// ``` dart
  /// @override
  /// Widget build(BuildContext context) {
  ///   return Scaffold(
  ///     body: SfDataGrid(
  ///       source: _employeeDataSource,
  ///       columns: [
  ///         GridNumericColumn(mappingName: 'id', headerText: 'ID')
  ///           ..cellStyle = DataGridCellStyle(
  ///              textStyle: TextStyle(color: Colors.red)),
  ///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
  ///         GridTextColumn(mappingName: 'designation',
  ///             headerText: 'Designation'),
  ///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary')
  ///       ],
  ///     ),
  ///   );
  /// }
  ///
  /// ```
  set cellStyle(DataGridCellStyle newCellStyle) =>
      this._cellStyle = newCellStyle;

  /// Whether column should be hidden.
  ///
  /// Defaults to false.
  bool get isHidden => this._isHidden;
  bool _isHidden = false;

  /// Whether column should be hidden.
  ///
  /// Defaults to false.
  set isHidden(bool newValue) => this._isHidden = newValue;

  /// Gets the formatted value based on the given format.
  ///
  /// This is applicable for column such as [GridNumericColumn]
  /// and [GridDateTimeColumn].
  String getFormattedValue(dynamic cellValue) {
    return cellValue?.toString();
  }

  /// Sets the cell type which indicates the renderer for the column.
  ///
  /// Call this method in constructor of the column when you write the custom
  /// column and renderer.
  @protected
  void setCellType(String cellType) {
    this._cellType = cellType;
  }
}

/// A column which displays the values of the string in its cells.
///
/// This column has all the required APIs to customize the widget [Text] as it
/// displays [Text] for all the cells.
///
/// ``` dart
/// @override
/// Widget build(BuildContext context) {
///   return SfDataGrid(
///     source: employeeDataSource,
///     columns: [
///     GridTextColumn(mappingName: 'name')
///     ..headerText = 'Name',
///     GridTextColumn(mappingName: 'designation')
///     ..headerText = 'Designation',
///   ],
///   );
/// }
/// ```
class GridTextColumn extends GridColumn {
  /// Creates a String column using [mappingName] and [headerText].
  GridTextColumn({@required String mappingName, String headerText}) {
    this.mappingName = mappingName;
    this.headerText = headerText;
    this._cellType = 'TextField';
    this.headerTextAlignment = Alignment.centerLeft;
    this.overflow = TextOverflow.ellipsis;
  }
}

/// A column used to display the numeric values in its cells (int, double).
///
/// [GridNumericColumn] supports [numberFormat] for formatting numeric values.
///
/// ``` dart
/// @override
/// Widget build(BuildContext context) {
///   return SfDataGrid(
///     source: employeeDataSource(),
///     columns: [
///      GridNumericColumn(mappingName: 'id')
///      ..headerText = 'ID',
///      GridNumericColumn(mappingName: 'salary')
///      ..headerText = 'Salary',
///   ],
///   );
/// }
/// ```
class GridNumericColumn extends GridColumn {
  /// [NumberFormat] to format a number in a locale-specific way.
  ///
  /// Defaults to null.
  ///
  /// ``` dart
  /// @override
  /// Widget build(BuildContext context) {
  ///   return Scaffold(
  ///     body: SfDataGrid(
  ///       source: _employeeDataSource,
  ///       columns: [
  ///         GridNumericColumn(mappingName: 'id', headerText: 'ID'),
  ///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
  ///         GridTextColumn(mappingName: 'designation',
  ///             headerText: 'Designation'),
  ///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary')
  ///           ..numberFormat =
  ///             NumberFormat.currency(locale: 'en_US', symbol: '\$')
  ///       ],
  ///     ),
  ///   );
  /// }
  /// ```
  intl.NumberFormat get numberFormat => this._numberFormat;
  intl.NumberFormat _numberFormat;

  /// [NumberFormat] to format a number in a locale-specific way.
  ///
  /// Defaults to null.
  ///
  /// ``` dart
  /// @override
  /// Widget build(BuildContext context) {
  ///   return Scaffold(
  ///     body: SfDataGrid(
  ///       source: _employeeDataSource,
  ///       columns: [
  ///         GridNumericColumn(mappingName: 'id', headerText: 'ID'),
  ///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
  ///         GridTextColumn(mappingName: 'designation',
  ///             headerText: 'Designation'),
  ///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary')
  ///           ..numberFormat =
  ///             NumberFormat.currency(locale: 'en_US', symbol: '\$')
  ///       ],
  ///     ),
  ///   );
  /// }
  /// ```
  set numberFormat(intl.NumberFormat newNumberFormat) =>
      this._numberFormat = newNumberFormat;

  /// Gets the formatted value based on the given format.
  ///
  /// This method is overridden to format the given numeric value to specified
  /// [numberFormat].
  @override
  String getFormattedValue(dynamic cellValue) {
    if (this.numberFormat != null) {
      return this.numberFormat.format(cellValue);
    }
    return cellValue.toString();
  }

  /// Creates a numeric column using [mappingName] and [headerText].
  GridNumericColumn({@required String mappingName, String headerText}) {
    this.mappingName = mappingName;
    this.headerText = headerText;
    this._cellType = 'Numeric';
    this.textAlignment = Alignment.centerRight;
    this.headerTextAlignment = Alignment.centerRight;
  }
}

/// A column that shows any custom widget inside its cells.
///
/// [GridWidgetColumn] supports loading [Image], [CheckBox] and so on in cells
/// in column.
///
/// [SfDataGrid.cellBuilder] allows you to set the widget in a column as needed.
/// ``` dart
/// @override
/// Widget build(BuildContext context) {
///   return SfDataGrid(
///     source: employeeDataSource(),
///     cellBuilder: (BuildContext context, GridColumn column, int rowIndex){
///        return employees[rowIndex].image;
///      },
///     columns: [
///      GridWidgetColumn(mappingName: image)
///      ..headerText = 'Image',
///   ],
///   );
/// }
/// ```
class GridWidgetColumn extends GridColumn {
  /// Creates a widget column using [mappingName] and [headerText].
  GridWidgetColumn({@required String mappingName, String headerText}) {
    this.mappingName = mappingName;
    this.headerText = headerText;
    this._cellType = 'Widget';
  }
}

/// A column which displays the values of DateTime in its cells.
///
/// [GridDateTimeColumn] supports [dateFormat] for formatting DateTime values.
///
/// ``` dart
/// @override
/// Widget build(BuildContext context) {
///   return SfDataGrid(
///     source: employeeDataSource(),
///     columns: [
///      GridNumericColumn(mappingName: 'id')
///      ..headerText = 'ID',
///      GridDateTimeColumn(mappingName: 'dateofjoining')
///      ..headerText = 'Date of Joining',
///   ],
///   );
/// }
/// ```
class GridDateTimeColumn extends GridColumn {
  /// [DateFormat] to format the dates in a locale-sensitive manner.
  ///
  /// Defaults to null.
  ///
  /// ``` dart
  /// Widget build(BuildContext context) {
  ///   return Scaffold(
  ///     body: SfDataGrid(
  ///       source: _employeeDataSource,
  ///       columns: [
  ///         GridNumericColumn(mappingName: 'id', headerText: 'ID'),
  ///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
  ///         GridDateTimeColumn(
  ///             mappingName: 'dateOfJoining', headerText: 'Date of Joining')
  ///           ..dateFormat = DateFormat('dd/MM/yyyy'),
  ///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary'),
  ///       ],
  ///     ),
  ///   );
  /// }
  /// ```
  intl.DateFormat get dateFormat => this._dateFormat;
  intl.DateFormat _dateFormat;

  /// [DateFormat] to format the dates in a locale-sensitive manner.
  ///
  /// Defaults to null.
  ///
  /// ``` dart
  /// Widget build(BuildContext context) {
  ///   return Scaffold(
  ///     body: SfDataGrid(
  ///       source: _employeeDataSource,
  ///       columns: [
  ///         GridNumericColumn(mappingName: 'id', headerText: 'ID'),
  ///         GridTextColumn(mappingName: 'name', headerText: 'Name'),
  ///         GridDateTimeColumn(
  ///             mappingName: 'dateOfJoining', headerText: 'Date of Joining')
  ///           ..dateFormat = DateFormat('dd/MM/yyyy'),
  ///         GridNumericColumn(mappingName: 'salary', headerText: 'Salary'),
  ///       ],
  ///     ),
  ///   );
  /// }
  /// ```
  set dateFormat(intl.DateFormat newDateTime) => this._dateFormat = newDateTime;

  final intl.DateFormat _defaultDateFormat = intl.DateFormat('dd-MM-yyyy');

  /// Gets the formatted value based on the given format.
  ///
  /// This method is overridden to format the given numeric value to specified
  /// [dateFormat].
  @override
  String getFormattedValue(dynamic cellValue) {
    if (this.dateFormat != null) {
      return this.dateFormat.format(cellValue);
    }
    return this._defaultDateFormat.format(cellValue);
  }

  /// Creates a datetime column using [mappingName] and [headerText].
  GridDateTimeColumn({@required String mappingName, String headerText}) {
    this.mappingName = mappingName;
    this.headerText = headerText;
    this._cellType = 'DateTime';
    this.textAlignment = Alignment.centerRight;
    this.headerTextAlignment = Alignment.centerRight;
  }
}
