part of datagrid;

/// Processes the row selection operation in [SfDataGrid].
///
/// You can override the available methods in this class to customize the
/// selection operation
/// in DataGrid and set the instance to the [SfDataGrid.selectionManager].
///
/// The following code shows how to change the Enter key behavior by
/// overriding the handleKeyEvent() method in RowSelectionManager,
///
/// ``` dart
/// final CustomSelectionManager _customSelectionManager = CustomSelectionManager();

/// @override
/// Widget build(BuildContext context){
///   return MaterialApp(
///       home: Scaffold(
///           body: SfDataGrid(
///             source: _employeeDataSource,
///             columns: [
///               GridNumericColumn(mappingName: 'id', headerText: 'ID'),
///               GridTextColumn(mappingName: 'name', headerText: 'Name'),
///               GridTextColumn(mappingName: 'designation', headerText: 'Designation'),
///               GridNumericColumn(mappingName: 'salary', headerText: 'Salary')
///             ],
///             selectionMode: SelectionMode.multiple,
///             navigationMode: GridNavigationMode.cell,
///             selectionManager: _customSelectionManager,
///           ))
///   );
/// }

/// class CustomSelectionManager extends RowSelectionManager{
///     @override
///     void handleKeyEvent(RawKeyEvent keyEvent) {
///         if(keyEvent.logicalKey == LogicalKeyboardKey.enter){
///             //apply your logic
///             return;
///         }

///         super.handleKeyEvent(keyEvent);
///     }
/// }
/// ```
class RowSelectionManager extends SelectionManagerBase {
  RowColumnIndex _pressedRowColumnIndex = RowColumnIndex(-1, -1);

  /// Creates the [RowSelectionManager] for [SfDataGrid] widget.
  RowSelectionManager({_DataGridStateDetails dataGridStateDetails})
      : super(dataGridStateDetails: dataGridStateDetails);

  void _applySelection(RowColumnIndex rowColumnIndex) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();

    var recordIndex = _GridIndexResolver.resolveToRecordIndex(
        dataGridSettings, rowColumnIndex.rowIndex);
    var record = _SelectionHelper.getRecord(dataGridSettings, recordIndex);

    var addedItems = [];
    var removeItems = [];

    if (dataGridSettings.selectionMode == SelectionMode.single) {
      if (this._selectedRows.contains(record)) return;

      if (dataGridSettings.onSelectionChanging != null ||
          dataGridSettings.onSelectionChanged != null) {
        addedItems.add(record);
        if (this._selectedRows.selectedRow.isNotEmpty)
          removeItems.add(this._selectedRows.selectedRow.first);
      }

      if (this._raiseSelectionChanging(
          newItems: addedItems, oldItems: removeItems)) {
        this._clearSelectedRow(dataGridSettings);
        this._addSelection(record, dataGridSettings);
        this.notifyListeners();
        this._raiseSelectionChanged(
            newItems: addedItems, oldItems: removeItems);
      }
      //If user, return false in selection changing, and we should need to update the current cell.
      else if (dataGridSettings.navigationMode == GridNavigationMode.cell) {
        this.notifyListeners();
      }
    } else if (dataGridSettings.selectionMode == SelectionMode.singleDeselect) {
      if (dataGridSettings.onSelectionChanging != null ||
          dataGridSettings.onSelectionChanged != null) {
        if (this._selectedRows.selectedRow.isNotEmpty)
          removeItems.add(this._selectedRows.selectedRow.first);

        if (!this._selectedRows.contains(record)) addedItems.add(record);
      }

      if (this._raiseSelectionChanging(
          newItems: addedItems, oldItems: removeItems)) {
        if (!this._selectedRows.contains(record)) {
          this._clearSelectedRow(dataGridSettings);
          this._addSelection(record, dataGridSettings);
        } else {
          this._clearSelectedRow(dataGridSettings);
        }
        this.notifyListeners();
        this._raiseSelectionChanged(
            newItems: addedItems, oldItems: removeItems);
      }
      //If user, return false in selection changing, and we should need to update the current cell.
      else if (dataGridSettings.navigationMode == GridNavigationMode.cell) {
        this.notifyListeners();
      }
    } else if (dataGridSettings.selectionMode == SelectionMode.multiple) {
      if (dataGridSettings.onSelectionChanging != null ||
          dataGridSettings.onSelectionChanged != null) {
        if (!this._selectedRows.contains(record)) {
          addedItems.add(record);
        } else {
          removeItems.add(record);
        }
      }

      if (this._raiseSelectionChanging(
          newItems: addedItems, oldItems: removeItems)) {
        if (!this._selectedRows.contains(record)) {
          this._addSelection(record, dataGridSettings);
        } else {
          this._removeSelection(record, dataGridSettings);
        }

        this.notifyListeners();
        this._raiseSelectionChanged(
            newItems: addedItems, oldItems: removeItems);
      }
      //If user, return false in selection changing, and we should need to update the current cell.
      else if (dataGridSettings.navigationMode == GridNavigationMode.cell) {
        this.notifyListeners();
      }
    }

    record = null;
    recordIndex = null;
  }

  void _addSelection(Object record, _DataGridSettings dataGridSettings) {
    if (!this._selectedRows.contains(record)) {
      var rowIndex =
          _SelectionHelper.resolveToRowIndex(dataGridSettings, record);
      this._setRowSelection(rowIndex, dataGridSettings, true);
      if (record != null && !this._selectedRows.selectedRow.contains(record)) {
        this._selectedRows.selectedRow.add(record);
        dataGridSettings.controller?._selectedRows?.add(record);
        this._refreshSelection();
        return;
      }
    }
  }

  void _removeSelection(Object record, _DataGridSettings dataGridSettings) {
    if (record != null && this._selectedRows.contains(record)) {
      var rowIndex =
          _SelectionHelper.resolveToRowIndex(dataGridSettings, record);
      this._setRowSelection(rowIndex, dataGridSettings, false);
      this._selectedRows.selectedRow.remove(record);
      dataGridSettings.controller?._selectedRows?.remove(record);
      rowIndex = null;
    }
  }

  void _clearSelectedRow(_DataGridSettings dataGridSettings) {
    if (this._selectedRows.selectedRow.isNotEmpty) {
      var selectedItem = this._selectedRows.selectedRow.first;
      this._removeSelection(selectedItem, dataGridSettings);
    }
  }

  void _clearSelectedRows(_DataGridSettings dataGridSettings) {
    if (this._selectedRows.selectedRow.length > 0) {
      for (int i = this._selectedRows.selectedRow.length - 1; i >= 0; i--) {
        var selectedItem = this._selectedRows.selectedRow[i];
        this._removeSelection(selectedItem, dataGridSettings);
      }
    }

    this._clearSelection(dataGridSettings);
  }

  void _setRowSelection(
      int rowIndex, _DataGridSettings dataGridSettings, bool isRowSelected) {
    if (dataGridSettings.rowGenerator.items.isEmpty) return;
    var row = dataGridSettings.rowGenerator.items
        .firstWhere((item) => item.rowIndex == rowIndex, orElse: () => null);
    if (row != null && !(row.rowRegion != RowRegion.body)) {
      row._isDirty = true;
      row.isSelectedRow = isRowSelected;
    }

    row = null;
  }

  void _clearSelection(_DataGridSettings dataGridSettings) {
    this._selectedRows.selectedRow.clear();
    dataGridSettings.controller?._selectedRows?.clear();
    dataGridSettings.controller?._selectedRow = null;
    dataGridSettings.controller?._selectedIndex = -1;
    dataGridSettings.rowGenerator.items.forEach((dataRow) {
      if (dataRow.isSelectedRow) dataRow.isSelectedRow = false;
    });

    this._clearCurrentCell(dataGridSettings);
  }

  void _clearCurrentCell(_DataGridSettings dataGridSettings) {
    dataGridSettings.currentCell._removeCurrentCell(dataGridSettings);
  }

  void _refreshSelection() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final Object _selectedRow = this._selectedRows.selectedRow.isNotEmpty
        ? this._selectedRows.selectedRow.last
        : null;
    final int _recordIndex =
        dataGridSettings.source.dataSource.indexOf(_selectedRow);
    dataGridSettings.controller?._selectedIndex = _recordIndex;
    dataGridSettings.controller?._selectedRow = _selectedRow;
  }

  void _addCurrentCell(Object record, _DataGridSettings dataGridSettings,
      {bool isSelectionChanging = false}) {
    var rowIndex = _SelectionHelper.resolveToRowIndex(dataGridSettings, record);

    if (rowIndex <= _GridIndexResolver.getHeaderIndex(dataGridSettings)) return;

    if (dataGridSettings.currentCell.columnIndex > 0) {
      dataGridSettings.currentCell._moveCurrentCellTo(dataGridSettings,
          RowColumnIndex(rowIndex, dataGridSettings.currentCell.columnIndex),
          isSelectionChanged: isSelectionChanging);
    } else {
      var firstVisibleColumnIndex =
          _GridIndexResolver.resolveToStartColumnIndex(dataGridSettings);
      dataGridSettings.currentCell._moveCurrentCellTo(
          dataGridSettings, RowColumnIndex(rowIndex, firstVisibleColumnIndex),
          isSelectionChanged: isSelectionChanging);
    }
  }

  void _onNavigationModeChanged() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    if (dataGridSettings.navigationMode == GridNavigationMode.row) {
      var currentRowColumnIndex = RowColumnIndex(
          dataGridSettings.currentCell.rowIndex,
          dataGridSettings.currentCell.columnIndex);
      this._clearCurrentCell(dataGridSettings);
      dataGridSettings.currentCell._updateBorderForMultipleSelection(
          dataGridSettings,
          nextRowColumnIndex: currentRowColumnIndex);
    } else {
      if (dataGridSettings.selectionMode != SelectionMode.none) {
        if (this._pressedRowColumnIndex.rowIndex < 0) {
          var lastRecord = dataGridSettings.controller.selectedRow;
          this._addCurrentCell(lastRecord, dataGridSettings);
        } else {
          dataGridSettings.currentCell._moveCurrentCellTo(
              dataGridSettings, this._pressedRowColumnIndex);
        }
      }
    }
  }

  void _onDataSourceChanged() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    this._clearSelection(dataGridSettings);
  }

  @override
  void handleTap(RowColumnIndex nextRowColumnIndex) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    this._pressedRowColumnIndex = nextRowColumnIndex;
    if (dataGridSettings.selectionMode == SelectionMode.none) return;
    var previousRowColumnIndex = RowColumnIndex(
        dataGridSettings.currentCell.rowIndex,
        dataGridSettings.currentCell.columnIndex);
    if (!dataGridSettings.currentCell
        ._handlePointerOperation(dataGridSettings, nextRowColumnIndex)) return;
    this._processSelection(
        dataGridSettings, nextRowColumnIndex, previousRowColumnIndex);
  }

  void _processSelection(
      _DataGridSettings dataGridSettings,
      RowColumnIndex nextRowColumnIndex,
      RowColumnIndex previousRowColumnIndex) {
    //If selectionMode is single. next current cell is going to present in the same selected row.
    // In this case, we don't update the whole data row. Instead of that we have to update next and previous current cell alone and call setState.
    // In other selection mode we will update the whole data row and it will work.
    if (dataGridSettings.selectionMode == SelectionMode.single &&
        dataGridSettings.navigationMode == GridNavigationMode.cell &&
        nextRowColumnIndex.rowIndex == previousRowColumnIndex.rowIndex &&
        nextRowColumnIndex.columnIndex != previousRowColumnIndex.columnIndex) {
      this.notifyListeners();
      return;
    }

    this._applySelection(nextRowColumnIndex);
  }

  @override
  void handleDataGridSourceChanges(
      {RowColumnIndex rowColumnIndex, String propertyName}) {
    switch (propertyName) {
      case 'selectedIndex':
        this.onSelectedIndexChanged();
        break;
      case 'selectedRow':
        this.onSelectedRowChanged();
        break;
      case 'selectedRows':
        this.onSelectedRowsChanged();
        break;
      default:
        break;
    }
  }

  @override
  void onSelectedRowChanged() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final Object newValue = dataGridSettings.controller?.selectedRow;

    if (dataGridSettings == null ||
        dataGridSettings.selectionMode == SelectionMode.none) return;

    final int recordIndex =
        dataGridSettings.source.dataSource?.indexOf(newValue);
    final int rowIndex =
        _GridIndexResolver.resolveToRowIndex(dataGridSettings, recordIndex);

    if (rowIndex < _GridIndexResolver.getHeaderIndex(dataGridSettings)) return;

    bool canClearSelections() {
      return this._selectedRows.selectedRow.isNotEmpty &&
          dataGridSettings.selectionMode != SelectionMode.multiple;
    }

    //If newValue is negative we have clear the whole selection data.
    //In multiple case we shouldn't to clear the collection as well source properties.
    if (newValue == null && canClearSelections()) {
      this._clearSelectedRow(dataGridSettings);
      this.notifyListeners();
      return;
    }

    if (!this._selectedRows.contains(newValue)) {
      //In multiple case we shouldn't to clear the collection as well source properties.
      if (canClearSelections()) {
        this._clearSelectedRow(dataGridSettings);
      }

      if (dataGridSettings.navigationMode == GridNavigationMode.cell) {
        this._addCurrentCell(newValue, dataGridSettings,
            isSelectionChanging: true);
      } else {
        var columnIndex =
            _GridIndexResolver.resolveToStartColumnIndex(dataGridSettings);
        var rowIndex =
            _SelectionHelper.resolveToRowIndex(dataGridSettings, newValue);
        dataGridSettings.currentCell
            ._updateCurrentRowColumnIndex(rowIndex, columnIndex);
      }

      this._addSelection(newValue, dataGridSettings);
      this.notifyListeners();
    }
  }

  @override
  void onSelectedIndexChanged() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final int newValue = dataGridSettings.controller.selectedIndex;

    if (dataGridSettings == null ||
        dataGridSettings.selectionMode == SelectionMode.none) return;

    if (dataGridSettings.source.dataSource.isEmpty ||
        newValue > dataGridSettings.source.dataSource.length) return;

    bool canClearSelections() {
      return this._selectedRows.selectedRow.isNotEmpty &&
          dataGridSettings.selectionMode != SelectionMode.multiple;
    }

    //If newValue is negative we have to clear the whole selection data.
    //In multiple case we shouldn't to clear the collection as well source properties.
    if (newValue == -1 && canClearSelections()) {
      this._clearSelectedRow(dataGridSettings);
      this.notifyListeners();
      return;
    }

    var record = _SelectionHelper.getRecord(dataGridSettings, newValue);
    if (record != null && !this._selectedRows.contains(record)) {
      //In multiple case we shouldn't to clear the collection as well source properties.
      if (canClearSelections()) {
        this._clearSelectedRow(dataGridSettings);
      }

      if (dataGridSettings.navigationMode == GridNavigationMode.cell) {
        this._addCurrentCell(record, dataGridSettings,
            isSelectionChanging: true);
      } else {
        var rowIndex =
            _SelectionHelper.resolveToRowIndex(dataGridSettings, record);
        var columnIndex =
            _GridIndexResolver.resolveToStartColumnIndex(dataGridSettings);
        dataGridSettings.currentCell
            ._updateCurrentRowColumnIndex(rowIndex, columnIndex);
      }

      this._addSelection(record, dataGridSettings);
      this.notifyListeners();
    }
  }

  void onSelectedRowsChanged() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final List<Object> newValue =
        dataGridSettings?.controller?.selectedRows?.toList(growable: false);

    if (dataGridSettings == null ||
        dataGridSettings.selectionMode != SelectionMode.multiple ||
        dataGridSettings.selectionMode == SelectionMode.none) {
      return;
    }

    if (newValue.isEmpty) {
      this._clearSelectedRows(dataGridSettings);
      this.notifyListeners();
      return;
    }

    this._clearSelectedRows(dataGridSettings);
    for (var record in newValue) {
      this._addSelection(record, dataGridSettings);
    }

    if (dataGridSettings.navigationMode == GridNavigationMode.cell &&
        this._selectedRows.selectedRow.isNotEmpty) {
      var lastRecord = this._selectedRows.selectedRow.last;
      this._addCurrentCell(lastRecord, dataGridSettings,
          isSelectionChanging: true);
    } else if (kIsWeb &&
        dataGridSettings.navigationMode == GridNavigationMode.row) {
      var lastRecord = this._selectedRows.selectedRow.last;
      var rowIndex =
          _SelectionHelper.resolveToRowIndex(dataGridSettings, lastRecord);
      dataGridSettings.currentCell._updateBorderForMultipleSelection(
          dataGridSettings,
          nextRowColumnIndex: RowColumnIndex(rowIndex, null));
    }

    this.notifyListeners();
  }

  @override
  void onGridSelectionModeChanged() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    if (dataGridSettings.selectionMode == SelectionMode.none &&
        this._selectedRows.selectedRow.length > 0) {
      this._clearSelectedRows(dataGridSettings);
    } else if (dataGridSettings.selectionMode != SelectionMode.none &&
        dataGridSettings.selectionMode != SelectionMode.multiple) {
      var lastRecord = dataGridSettings.controller.selectedRow;
      this._clearSelection(dataGridSettings);
      if (dataGridSettings.navigationMode == GridNavigationMode.cell) {
        if (this._pressedRowColumnIndex.rowIndex < 0) {
          this._addCurrentCell(lastRecord, dataGridSettings);
        } else {
          dataGridSettings.currentCell._moveCurrentCellTo(
              dataGridSettings, this._pressedRowColumnIndex);
        }
      }
      this._addSelection(lastRecord, dataGridSettings);
    } else if (kIsWeb &&
        dataGridSettings.selectionMode == SelectionMode.multiple) {
      var currentRowColumnIndex =
          RowColumnIndex(dataGridSettings.currentCell.rowIndex, null);
      dataGridSettings.currentCell._updateBorderForMultipleSelection(
          dataGridSettings,
          nextRowColumnIndex: currentRowColumnIndex);
    }
  }

  @override
  void _onRowColumnChanged(int recordLength, int columnLength) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final currentCell = dataGridSettings.currentCell;

    if (currentCell.rowIndex == -1 && currentCell.columnIndex == -1) return;

    var rowIndex = _GridIndexResolver.resolveToRecordIndex(
        dataGridSettings, currentCell.rowIndex);
    if (recordLength != null &&
        recordLength > 0 &&
        rowIndex >= recordLength &&
        currentCell.rowIndex != -1) {
      var startRowIndexIndex =
          this._getPreviousRowIndex(dataGridSettings, currentCell.rowIndex);
      currentCell._moveCurrentCellTo(dataGridSettings,
          RowColumnIndex(startRowIndexIndex, currentCell.columnIndex),
          needToUpdateColumn: false);
      this._refreshSelection();
    }

    var columnIndex = _GridIndexResolver.resolveToGridVisibleColumnIndex(
        dataGridSettings, currentCell.columnIndex);
    if (columnLength > 0 &&
        columnLength != null &&
        columnIndex >= columnLength &&
        currentCell.columnIndex != -1) {
      var startColumnIndex = this
          ._getPreviousColumnIndex(dataGridSettings, currentCell.columnIndex);
      currentCell._moveCurrentCellTo(dataGridSettings,
          RowColumnIndex(currentCell.rowIndex, startColumnIndex),
          needToUpdateColumn: false);
    }
  }

  @override
  void _updateSelectionController(
      {bool isSelectionModeChanged = false,
      bool isNavigationModeChanged = false,
      bool isDataSourceChanged = false}) {
    if (isDataSourceChanged) {
      this._onDataSourceChanged();
    }

    if (isSelectionModeChanged) this.onGridSelectionModeChanged();

    if (isNavigationModeChanged) {
      this._onNavigationModeChanged();
    }
  }

  //KeyNavigation
  @override
  void handleKeyEvent(RawKeyEvent keyEvent) {
    if (keyEvent.logicalKey == LogicalKeyboardKey.tab) {
      this._processKeyTab(keyEvent);
    }

    if (keyEvent.logicalKey == LogicalKeyboardKey.home) {
      final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
      this._processHomeKey(dataGridSettings, keyEvent);
    }

    if (keyEvent.logicalKey == LogicalKeyboardKey.end) {
      final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
      this._processEndKey(dataGridSettings, keyEvent);
    }

    if (keyEvent.logicalKey == LogicalKeyboardKey.pageUp) {
      this._processPageUp();
    }

    if (keyEvent.logicalKey == LogicalKeyboardKey.pageDown) {
      this._processPageDown();
    }

    if (keyEvent?.logicalKey == LogicalKeyboardKey.arrowUp) {
      this._processKeyUp(keyEvent);
    }

    if (keyEvent?.logicalKey == LogicalKeyboardKey.arrowDown ||
        keyEvent?.logicalKey == LogicalKeyboardKey.enter) {
      this._processKeyDown(keyEvent);
    }

    if (keyEvent?.logicalKey == LogicalKeyboardKey.arrowLeft) {
      final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
      if (dataGridSettings.textDirection == TextDirection.rtl)
        this._processKeyRight(dataGridSettings, keyEvent);
      else
        this._processKeyLeft(dataGridSettings, keyEvent);
    }

    if (keyEvent?.logicalKey == LogicalKeyboardKey.arrowRight) {
      final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
      if (dataGridSettings.textDirection == TextDirection.rtl)
        this._processKeyLeft(dataGridSettings, keyEvent);
      else
        this._processKeyRight(dataGridSettings, keyEvent);
    }

    if (keyEvent.logicalKey == LogicalKeyboardKey.keyA) {
      if (keyEvent.isControlPressed) {
        this._processSelectedAll();
      }
    }

    if (keyEvent?.logicalKey == LogicalKeyboardKey.space) {
      this._processSpaceKey();
    }
  }

  void _processEndKey(
      _DataGridSettings dataGridSettings, RawKeyEvent keyEvent) {
    final _CurrentCellManager currentCell = dataGridSettings.currentCell;
    final int lastCellIndex =
        _SelectionHelper.getLastCellIndex(dataGridSettings);
    final needToScrollToMinOrMaxExtent =
        dataGridSettings.container.extentWidth > dataGridSettings.viewWidth
            ? true
            : false;

    if (needToScrollToMinOrMaxExtent) {
      _SelectionHelper.scrollInViewFromLeft(dataGridSettings,
          needToScrollMaxExtent: true);
    }

    if (keyEvent.isControlPressed &&
        keyEvent.logicalKey != LogicalKeyboardKey.arrowRight) {
      final lastRowIndex =
          _SelectionHelper.getLastNavigatingRowIndex(dataGridSettings);
      _SelectionHelper.scrollInViewFromTop(dataGridSettings,
          needToScrollToMaxExtent: true);
      this._processSelectionAndCurrentCell(
          dataGridSettings, RowColumnIndex(lastRowIndex, lastCellIndex));
    } else {
      this._processSelectionAndCurrentCell(dataGridSettings,
          RowColumnIndex(currentCell.rowIndex, lastCellIndex));
    }
  }

  void _processHomeKey(
      _DataGridSettings dataGridSettings, RawKeyEvent keyEvent) {
    final _CurrentCellManager currentCell = dataGridSettings.currentCell;
    int firstCellIndex = _SelectionHelper.getFirstCellIndex(dataGridSettings);
    final needToScrollToMinOrMaxExtend =
        dataGridSettings.container.extentWidth > dataGridSettings.viewWidth
            ? true
            : false;

    if (needToScrollToMinOrMaxExtend) {
      _SelectionHelper.scrollInViewFromRight(dataGridSettings,
          needToScrollToMinExtent: true);
    }

    if (keyEvent.isControlPressed &&
        keyEvent.logicalKey != LogicalKeyboardKey.arrowLeft) {
      final firstRowIndex =
          _SelectionHelper.getFirstNavigatingRowIndex(dataGridSettings);
      _SelectionHelper.scrollInViewFromDown(dataGridSettings,
          needToScrollToMinExtent: true);
      this._processSelectionAndCurrentCell(
          dataGridSettings, RowColumnIndex(firstRowIndex, firstCellIndex));
    } else {
      this._processSelectionAndCurrentCell(dataGridSettings,
          RowColumnIndex(currentCell.rowIndex, firstCellIndex));
    }
  }

  void _processPageDown() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final _CurrentCellManager currentCell = dataGridSettings.currentCell;
    final index = _SelectionHelper.getNextPageIndex(dataGridSettings);
    if (currentCell.rowIndex != index && index != -1) {
      this._processSelectionAndCurrentCell(
          dataGridSettings, RowColumnIndex(index, currentCell.columnIndex));
    }
  }

  void _processPageUp() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final _CurrentCellManager currentCell = dataGridSettings.currentCell;
    final index = _SelectionHelper.getPreviousPageIndex(dataGridSettings);
    if (currentCell.rowIndex != index) {
      this._processSelectionAndCurrentCell(
          dataGridSettings, RowColumnIndex(index, currentCell.columnIndex));
    }
  }

  void _processKeyDown(RawKeyEvent keyEvent) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final _CurrentCellManager currentCell = dataGridSettings.currentCell;
    final int nextRowIndex =
        this._getNextRowIndex(dataGridSettings, currentCell.rowIndex);
    final lastRowIndex =
        _SelectionHelper.getLastNavigatingRowIndex(dataGridSettings);
    int nextColumnIndex = currentCell.columnIndex;

    if (nextColumnIndex <= 0)
      nextColumnIndex = _SelectionHelper.getFirstCellIndex(dataGridSettings);

    if (nextRowIndex > lastRowIndex) return;

    if (keyEvent.isControlPressed) {
      _SelectionHelper.scrollInViewFromTop(dataGridSettings,
          needToScrollToMaxExtent: true);
      this._processSelectionAndCurrentCell(
          dataGridSettings,
          RowColumnIndex(lastRowIndex, nextColumnIndex),
          keyEvent.isShiftPressed);
    } else {
      this._processSelectionAndCurrentCell(
          dataGridSettings,
          RowColumnIndex(nextRowIndex, nextColumnIndex),
          keyEvent.isShiftPressed);
    }
  }

  void _processKeyUp(RawKeyEvent keyEvent) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final currentCell = dataGridSettings.currentCell;
    final int previousRowIndex =
        this._getPreviousRowIndex(dataGridSettings, currentCell.rowIndex);

    if (previousRowIndex == currentCell.rowIndex) return;

    if (keyEvent.isControlPressed) {
      final firstRowIndex = _SelectionHelper.getFirstRowIndex(dataGridSettings);
      _SelectionHelper.scrollInViewFromDown(dataGridSettings,
          needToScrollToMinExtent: true);
      this._processSelectionAndCurrentCell(
          dataGridSettings,
          RowColumnIndex(firstRowIndex, currentCell.columnIndex),
          keyEvent.isShiftPressed);
    } else {
      this._processSelectionAndCurrentCell(
          dataGridSettings,
          RowColumnIndex(previousRowIndex, currentCell.columnIndex),
          keyEvent.isShiftPressed);
    }
  }

  void _processKeyRight(
      _DataGridSettings dataGridSettings, RawKeyEvent keyEvent) {
    if (dataGridSettings.navigationMode == GridNavigationMode.row) return;

    final currentCell = dataGridSettings.currentCell;
    var lastCellIndex = _SelectionHelper.getLastCellIndex(dataGridSettings);
    var nextCellIndex =
        this._getNextColumnIndex(dataGridSettings, currentCell.columnIndex);

    if (currentCell.rowIndex <=
            _GridIndexResolver.getHeaderIndex(dataGridSettings) ||
        nextCellIndex > lastCellIndex ||
        currentCell.columnIndex == nextCellIndex) {
      return;
    }

    if (keyEvent.isControlPressed) {
      if (dataGridSettings.textDirection == TextDirection.rtl)
        this._processHomeKey(dataGridSettings, keyEvent);
      else
        this._processEndKey(dataGridSettings, keyEvent);
    } else {
      currentCell._processCurrentCell(
          dataGridSettings, RowColumnIndex(currentCell.rowIndex, nextCellIndex),
          isSelectionChanged: true);
      if (dataGridSettings.navigationMode == GridNavigationMode.cell)
        this.notifyListeners();
    }
  }

  void _processKeyLeft(
      _DataGridSettings dataGridSettings, RawKeyEvent keyEvent) {
    if (dataGridSettings.navigationMode == GridNavigationMode.row) return;

    final currentCell = dataGridSettings.currentCell;
    final int previousCellIndex =
        this._getPreviousColumnIndex(dataGridSettings, currentCell.columnIndex);

    if (currentCell.rowIndex <=
            _GridIndexResolver.getHeaderIndex(dataGridSettings) ||
        previousCellIndex <
            _GridIndexResolver.resolveToStartColumnIndex(dataGridSettings))
      return;

    if (keyEvent.isControlPressed) {
      if (dataGridSettings.textDirection == TextDirection.rtl)
        this._processEndKey(dataGridSettings, keyEvent);
      else
        this._processHomeKey(dataGridSettings, keyEvent);
    } else {
      currentCell._processCurrentCell(dataGridSettings,
          RowColumnIndex(currentCell.rowIndex, previousCellIndex),
          isSelectionChanged: true);
      if (dataGridSettings.navigationMode == GridNavigationMode.cell)
        this.notifyListeners();
    }
  }

  void _processKeyTab(RawKeyEvent keyEvent) {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    final _CurrentCellManager currentCell = dataGridSettings.currentCell;
    final lastCellIndex = _SelectionHelper.getLastCellIndex(dataGridSettings);
    final firstCellIndex = _SelectionHelper.getFirstCellIndex(dataGridSettings);
    final int firstRowIndex =
        _SelectionHelper.getFirstRowIndex(dataGridSettings);

    if (dataGridSettings.navigationMode == GridNavigationMode.row ||
        (currentCell.rowIndex < 0 && currentCell.columnIndex < 0)) {
      this._processSelectionAndCurrentCell(
          dataGridSettings, RowColumnIndex(firstRowIndex, firstCellIndex));
      this.notifyListeners();
      return;
    }

    var needToScrollToMinOrMaxExtend =
        dataGridSettings.container.extentWidth > dataGridSettings.viewWidth
            ? true
            : false;

    if (keyEvent.isShiftPressed) {
      if (currentCell.columnIndex == firstCellIndex &&
          currentCell.rowIndex == firstRowIndex) return;

      if (currentCell.columnIndex == firstCellIndex) {
        var previousRowIndex =
            this._getPreviousRowIndex(dataGridSettings, currentCell.rowIndex);
        if (needToScrollToMinOrMaxExtend)
          _SelectionHelper.scrollInViewFromLeft(dataGridSettings,
              needToScrollMaxExtent: needToScrollToMinOrMaxExtend);
        this._processSelectionAndCurrentCell(
            dataGridSettings, RowColumnIndex(previousRowIndex, lastCellIndex));
      } else {
        this._processKeyLeft(dataGridSettings, keyEvent);
      }
    } else {
      if (currentCell.columnIndex == lastCellIndex) {
        var nextRowIndex =
            this._getNextRowIndex(dataGridSettings, currentCell.rowIndex);
        if (needToScrollToMinOrMaxExtend)
          _SelectionHelper.scrollInViewFromRight(dataGridSettings,
              needToScrollToMinExtent: needToScrollToMinOrMaxExtend);
        this._processSelectionAndCurrentCell(
            dataGridSettings, RowColumnIndex(nextRowIndex, firstCellIndex));
      } else {
        this._processKeyRight(dataGridSettings, keyEvent);
      }
    }
  }

  void _processSelectedAll() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    if (dataGridSettings.selectionMode != SelectionMode.multiple) return;

    var addedItems = [];
    var removeItems = [];
    if (dataGridSettings.onSelectionChanging != null ||
        dataGridSettings.onSelectionChanged != null) {
      addedItems.addAll(dataGridSettings.source.dataSource);
    }

    if (this
        ._raiseSelectionChanging(oldItems: removeItems, newItems: addedItems)) {
      dataGridSettings.source.dataSource.forEach((record) {
        if (!this._selectedRows.contains(record)) {
          var rowIndex =
              _SelectionHelper.resolveToRowIndex(dataGridSettings, record);
          if (rowIndex != -1 &&
              dataGridSettings.rowGenerator.items
                  .any((row) => row.rowIndex == rowIndex)) {
            this._setRowSelection(rowIndex, dataGridSettings, true);
            this._selectedRows.selectedRow.add(record);
          } else {
            this._selectedRows.selectedRow.add(record);
          }
        }
      });

      dataGridSettings.controller.selectedRows
          .addAll(dataGridSettings.source.dataSource);
      this._refreshSelection();
      dataGridSettings.container._isDirty = true;
      this.notifyListeners();
      this._raiseSelectionChanged(newItems: addedItems, oldItems: removeItems);
    }
  }

  void _processSpaceKey() {
    final _DataGridSettings dataGridSettings = this._dataGridStateDetails();
    if (dataGridSettings.selectionMode == SelectionMode.single) return;
    final currentCell = dataGridSettings.currentCell;
    this._applySelection(
        RowColumnIndex(currentCell.rowIndex, currentCell.columnIndex));
  }

  void _processSelectionAndCurrentCell(
      _DataGridSettings dataGridSettings, RowColumnIndex rowColumnIndex,
      [bool isShiftKeyPressed = false, bool isProgrammatic = false]) {
    var previousRowColumnIndex = RowColumnIndex(
        dataGridSettings.currentCell.rowIndex,
        dataGridSettings.currentCell.columnIndex);
    if (isProgrammatic) {
      dataGridSettings.currentCell
          ._moveCurrentCellTo(dataGridSettings, rowColumnIndex);
    } else {
      dataGridSettings.currentCell
          ._processCurrentCell(dataGridSettings, rowColumnIndex);
    }

    if (dataGridSettings.selectionMode == SelectionMode.multiple) {
      dataGridSettings.currentCell._updateBorderForMultipleSelection(
          dataGridSettings,
          nextRowColumnIndex: rowColumnIndex,
          previousRowColumnIndex: previousRowColumnIndex);
      if (isShiftKeyPressed)
        this._processSelection(
            dataGridSettings, rowColumnIndex, previousRowColumnIndex);
      else
        this.notifyListeners();
    }

    this._pressedRowColumnIndex = rowColumnIndex;
  }
}
