part of datagrid;

/// A leaf in the tree with value and optional sort key.
///
/// * T - _required_ - Tree value
class _GenericTreeTableEntry<T> extends _TreeTableNode
    with _TreeTableEntryBase {
  T _thisValue;

  /// Initializes a new instance of the GenericTreeTableEntry class.
  _GenericTreeTableEntry();

  /// Initializes a new instance of the GenericTreeTableEntry class.
  ///
  /// * tree - _required_ - Current tree instance
  /// * value - _required_ - Tree value
  _GenericTreeTableEntry.fromTreeTable(_TreeTable tree, T value) {
    this._thisValue = value;
    this.tree = tree;
  }

  /// Initializes a new instance of the GenericTreeTableEntry class.
  ///
  /// * tree - _required_ - Internal tree
  /// * value - _required_ - tree value
  _GenericTreeTableEntry.fromGenericTreeTable(
      _GenericTreeTable<T> tree, T value) {
    this._thisValue = value;
    this.tree = tree.internalTree;
  }

  ///Gets the value attached to this leaf.
  T get valueBase => this._thisValue;
  set valueBase(T value) {
    this._thisValue = value;
  }

  /// The sort key of this leaf.
  ///
  /// Returns the sort key of this leaf
  T getSortKeyBase() {
    return this.valueBase;
  }

  /// The minimum value (of the most-left leaf) of the branch in a sorted tree.
  ///
  /// Returns the minimum value (of the most-left leaf) of the branch in a sorted tree.
  @override
  Object getMinimum() {
    return this.getSortKey();
  }

  /// The number of child nodes (+1 for the current node).
  ///
  /// Returns the number of child nodes (+1 for the current node).
  @override
  int getCount() {
    return 1;
  }

  /// Creates a branch that can hold this entry when new leaves are inserted into the tree.
  ///
  /// * tree - _required_ - tree instance
  ///
  /// Returns the instance of newly created branch
  @override
  _TreeTableBranchBase createBranch(_TreeTable tree) {
    return _TreeTableBranch(tree);
  }

  /// The Debug / text information about the node.
  ///
  /// Returns the Debug / text information about the node.
  String getNodeInfo() {
    return getNodeInfoBase() +
        ", " +
        (this.value != null ? this.value.toString() : "null");
  }

  /// Indicates whether this is a leaf.
  ///
  /// Returns boolean value
  @override
  bool isEntry() {
    return true;
  }

  Object getSortKey() {
    return this.getSortKeyBase();
  }

  Object get value => this.valueBase;
  set value(Object value) => this.valueBase = value;
}

/// A tree table.
///
/// * T - _required_ - Generic tree table
class _GenericTreeTable<T>
    implements
        _TreeTableBase,
        _ListGenericBase<_GenericTreeTableEntry<T>>,
        _CollectionGenericBase<_GenericTreeTableEntry<T>> {
  _TreeTable _thisTree;

  /// Gets the tag that can be associated with the tree
  Object get tag => this._thisTree.tag;
  set tag(Object value) => this._thisTree.tag = value;

  /// Gets the identifier
  int get identifier => _identifier;
  int _identifier = 0;
  set identifier(int value) {
    if (value == _identifier) return;
    _identifier = value;
  }

  /// Gets the non-generic tree table with actual implementation.
  _TreeTable get internalTree => this._thisTree;

  /// Initializes a new instance of the [GenericTreeTable{T}] class.
  ///
  /// * sorted - _required_ - The boolean value denotes whether the tree is sorted or not
  _GenericTreeTable(bool sorted) {
    this._thisTree = _TreeTable(sorted)..tag = this;
  }

  /// Gets the root node.
  _TreeTableNodeBase get root => this._thisTree.root;

  /// Gets a value indicating whether thisTree is sorted or not.
  bool get sorted => this._thisTree.sorted;

  /// Gets the comparer used by sorted trees.
  Comparable get comparer => this._thisTree.comparer;
  set(Comparable value) {
    this._thisTree.comparer = value;
  }

  /// Gets a value indicating whether the tree was initialize or not.
  bool get isInitializing => this._thisTree.isInitializing;

  /// Optimizes insertion of many items when thisTree is initialized for the first time.
  void beginInit() {
    this._thisTree.beginInit();
  }

  /// Ends optimization of insertion of items when thisTree is initialized for the first time.
  void endInit() {
    this._thisTree.endInit();
  }

  /// Optimized access to a subsequent entry.
  ///
  /// * current - _required_ - current entry
  ///
  /// Returns next subsequent entry
  _GenericTreeTableEntry<T> getNextEntryBase(
      _GenericTreeTableEntry<T> current) {
    return this._thisTree.getNextEntry(current);
  }

  /// Optimized access to a subsequent entry.
  ///
  /// * current - _required_ - current entry
  ///
  /// Returns next subsequent entry
  _TreeTableEntryBase getNextEntry(_TreeTableEntryBase current) {
    return this.getNextEntryBase(current);
  }

  /// Optimized access to the previous entry.
  ///
  /// * current - _required_ - current entry
  ///
  /// Returns previous entry
  _GenericTreeTableEntry<T> getPreviousEntryBase(
      _GenericTreeTableEntry<T> current) {
    return this._thisTree.getPreviousEntry(current);
  }

  /// Optimized access to the previous entry.
  ///
  /// * current - _required_ - current entry
  ///
  /// Returns previous entry
  _TreeTableEntryBase getPreviousEntry(_TreeTableEntryBase current) {
    return this.getPreviousEntryBase(current);
  }

  /// Add the key if it is not in the collection
  ///
  /// * key - _required_ - key needs to be add
  /// * entry - _required_ - tree value
  ///
  /// Returns the instance for the tree
  _GenericTreeTableEntry<T> addIfNotExists(
      Object key, _GenericTreeTableEntry<T> entry) {
    return this._thisTree.addIfNotExists(key, entry);
  }

  /// Used to find the index of the specified key
  ///
  /// * key - _required_ - key value
  ///
  /// Returns the index of the key
  int indexOfKey(Object key) {
    return this._thisTree.indexOfKey(key);
  }

  /// Used to find the key in the collection
  ///
  /// * key - _required_ - key needs to find
  ///
  /// Returns the value corresponding to the specified key
  _GenericTreeTableEntry<T> findKey(Object key) {
    return this._thisTree.findKey(key);
  }

  /// Used to find the key approximate to the specified key
  ///
  /// * key - _required_ - key needs to be search
  ///
  /// Returns the value corresponds to the approximate key
  _GenericTreeTableEntry<T> findHighestSmallerOrEqualKey(Object key) {
    return this._thisTree.findHighestSmallerOrEqualKey(key);
  }

  /// Gets the item at the zero-based index.
  ///
  /// * index - _required_ - index value
  ///
  /// Returns the Tree value
  operator [](index) {
    return this._thisTree[index];
  }

  /// Sets the item at the zero-based index.
  operator []=(index, value) {
    this._thisTree[index] = value;
  }

  /// Inserts an item into the collection at the specified index.
  ///
  /// * index - _required_ - The zero-based index at which the item should be inserted.
  /// * item - _required_ - The item to insert. The value must not be a NULL reference (Nothing in Visual Basic).
  void insertBase(int index, _GenericTreeTableEntry<T> item) {
    this._thisTree.insert(index, item);
  }

  /// Removes the specified item from the collection.
  ///
  /// * item - _required_ - The item to remove from the collection. If the value is NULL or the item is not contained
  /// in the collection, the method will do nothing.
  ///
  /// Returns the collection after removing the specified item from the tree collection
  bool removeBase(_GenericTreeTableEntry<T> item) {
    return this._thisTree.remove(item);
  }

  /// Removes the item at the specified index of the collection.
  ///
  /// * index - _required_ - The zero-based index of the item to remove.
  void removeAtBase(int index) {
    this._thisTree.removeAt(index);
  }

  /// Returns the zero-based index of the occurrence of the item in the collection.
  ///
  /// * item - _required_ - The item to locate in the collection. The value can be a NULL reference (Nothing in Visual Basic).
  ///
  /// Returns the zero-based index of the occurrence of the item within the entire collection, if found; otherwise -1.
  int indexOfBase(_GenericTreeTableEntry<T> item) {
    return this._thisTree.indexOf(item);
  }

  /// Gets the number of items contained in the collection.
  int get count => this._thisTree.getCount();

  /// Gets a value indicating whether the collection is read-only.
  ///
  /// Returns true if the collection is read-only; otherwise, false.
  bool get isReadOnly => this._thisTree.isReadOnly;

  /// Adds a value to the end of the collection.
  ///
  /// * item - _required_ - The item to be added to the end of the collection. The value must not be a NULL reference (Nothing in Visual Basic).
  ///
  /// Returns the zero-based collection index at which the value has been added.
  int addBase(_GenericTreeTableEntry<T> item) {
    return this._thisTree.add(item);
  }

  /// Add an item to the collection
  ///
  /// * item - _required_ - tree needs to be add in an collection
  void addGeneric(_GenericTreeTableEntry<T> item) {
    this.add(item);
  }

  /// Removes all items from the collection.
  void clear() {
    this._thisTree.clear();
  }

  /// Determines if the item belongs to this collection.
  ///
  /// * item - _required_ - The object to locate in the collection. The value can be a NULL reference (Nothing in Visual Basic).
  ///
  /// Returns true if item is found in the collection; otherwise False.
  bool containsBase(_GenericTreeTableEntry<T> item) {
    if (item == null) {
      return false;
    }

    return this._thisTree.containsBase(item);
  }

  /// Copies the entire collection to a compatible one-dimensional array, starting at the specified index of the target array.
  ///
  /// * array - _required_ - The one-dimensional array that is the destination of the items copied from the  ArrayList. The array must have zero-based indexing.
  /// * index - _required_ - The zero-based index in an array at which copying begins.
  void copyToBase(List<dynamic> array, int index) {
    this._thisTree.copyTo(array, index);
  }

  /// Inserts a node at the specified index.
  ///
  /// * index - _required_ - position where to insert the value
  /// * value - _required_ - tree value need to be insert
  void insert(int index, Object value) {
    this.insertBase(index, value);
  }

  /// Removes the node with the specified value.
  ///
  /// * value - _required_ - tree value needs to be remove
  void remove(Object value) {
    this.removeBase(value);
  }

  /// Indicates whether the node belongs to this tree.
  ///
  /// * value - _required_ - value needs to be search in the tree
  ///
  /// Returns the boolean value indicating whether the node belongs to this tree.
  bool contains(Object value) {
    return this.containsBase(value);
  }

  /// Gets the index of the specified object.
  ///
  /// * value - _required_ - Value of the object.
  ///
  /// Returns the index value of the object.
  int indexOf(Object value) {
    return this.indexOfBase(value);
  }

  /// Adds the specified object to the collection.
  ///
  /// * value - _required_ - Value of the object to add.
  ///
  /// Returns the zero-based collection index at which the value has been added
  int add(Object value) {
    return this.addBase(value);
  }

  /// Gets a value indicating whether the collection has a fixed size.
  ///
  /// Returns `True` if the collection has a fixed size; otherwise, `false`.
  bool get isFixedSize => this.isFixedSizeBase;

  /// Gets a value indicating whether the collection has a fixed size.
  ///
  /// Returns `True` if the collection has a fixed size; otherwise, `false`.
  bool get isFixedSizeBase => this._thisTree.isFixedSize;

  /// Gets a value indicating whether access to the `ICollection` is synchronized (thread safe).
  ///
  /// Returns `True` if access to the [ICollection] is synchronized; otherwise, `false`. Returns `false`.
  bool get isSynchronized => false;

  /// Copies the element from this collection into an array.
  ///
  /// * array - _required_ - The destination array.
  /// * index - _required_ - The starting index in the destination array.
  @override
  void copyTo(List array, int index) {
    this.copyToBase(array, index);
  }

  /// Gets an object that can be used to synchronize access to the [ICollection].
  ///
  /// Returns an object that can be used to synchronize access to the [ICollection].
  Object get syncRoot => this.syncRootBase;

  /// Gets an object that can be used to synchronize access to the [ICollection]
  ///
  /// Returns an object that can be used to synchronize access to the [ICollection].
  Object get syncRootBase => null;

  noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}
