part of datagrid;

/// A collection of entities that is shared with a parent collection for which distances
/// need to counted.
///
///  The collection only is a subset for a specific range in
/// the parent distance collection. When you change the size of an element in this collection the change will
/// also be reflected in the parent collection and vice versa.
// Need to check whether the distance is DistanceCounterSubset or not in PixelScrollAxis's SetFooterLineCount method.
// ignore: unused_element
class _DistanceCounterSubset extends _DistanceCounterCollectionBase
    with _DistanceCounterCollectionBase {
  _DistanceCounterCollectionBase _trackDCC;
  int _count = 0;
  int _start = 0;

  /// Initializes a new instance of the DistanceCounterSubset class.
  ///
  /// * trackedParentCollection - _required_ - The parent collection for which a subset is "tracked".
  _DistanceCounterSubset(
      _DistanceCounterCollectionBase trackedParentCollection) {
    this._trackDCC = trackedParentCollection;
  }

  /// Gets the starting index of this collection in the parent collection.
  ///
  /// Returns the starting index of this collection in the parent collection.
  int get start => this._start;

  /// Sets the starting index of this collection in the parent collection.
  set start(int value) {
    this._start = value;
  }

  /// Gets the ending index of this collection in the parent collection.
  ///
  /// Returns the ending index of this collection in the parent collection.
  int get end => this._start + this._count - 1;

  /// Restores the distances in the parent collection for this subset to their default distance.
  void clear() {
    this._trackDCC.resetRange(this.start, this.end);
  }

  /// Gets the raw number of entities (lines, rows or columns).
  ///
  /// Returns the raw number of entities (lines, rows or columns).
  int get count => this._count;

  /// Sets the raw number of entities
  set count(int value) {
    if (this._count != value) {
      this._count = value;
    }
  }

  /// Gets the default distance (row height or column width) an entity spans.
  ///
  /// Returns the default distance (row height or column width) an entity spans.
  double get defaultDistance => this._trackDCC.defaultDistance;

  /// Sets the dafault distnace an entity spans.
  set defaultDistance(double value) {
    this._trackDCC.defaultDistance = value;
  }

  /// Gets the total distance all entities span (e.g. total height of all rows in grid).
  ///
  /// Returns the total distance all entities span (e.g. total height of all rows in grid).
  double get totalDistance {
    if (this.count == 0) {
      return 0;
    }

    {
      return this._trackDCC.getCumulatedDistanceAt(this.end) -
          this._trackDCC.getCumulatedDistanceAt(this._start) +
          this._trackDCC[this.end];
    }
  }

  /// Hides a specified range of entities (lines, rows or columns).
  ///
  /// * from - _required_ - The index for the first entity.
  /// * to - _required_ - The raw index for the last entity.
  /// * distance - _required_ - The distance.
  void setRange(int from, int to, double distance) {
    this._trackDCC.setRange(from + this._start, to + this._start, distance);
  }

  /// Resets the range by restoring the default distance for all entries in the specified range.
  ///
  /// * from - _required_ - The index for the first entity.
  /// * to - _required_ - The raw index for the last entity.
  void resetRange(int from, int to) {
    this._trackDCC.resetRange(from + this._start, to + this._start);
  }

  /// Gets the distance for an entity from the given index.
  ///
  /// * index - _required_ - The index for the entity.
  ///
  /// Returns the distance for an entity from the given index.
  operator [](index) {
    return this._trackDCC[index + this._start];
  }

  /// Sets the distance for an entity from the given index.
  operator []=(index, value) {
    this._trackDCC[index + this._start] = value;
  }

  /// Gets the next visible index.
  ///
  /// Skip subsequent entities for which the distance is 0.0 and return the next entity.
  ///
  /// * index - _required_ - The index.
  ///
  /// Returns the next visible index from the given index.
  int getNextVisibleIndex(int index) {
    int n = this._trackDCC.getNextVisibleIndex(index + this._start);
    if (n > this.end) {
      return -1;
    }

    return n - this._start;
  }

  /// Gets the previous visible index.
  ///
  /// Skip previous entities for which the distance is 0.0 and return the previous entity.
  ///
  /// * index - _required_ - The index.
  ///
  /// Returns the previous visible index from the given index.
  int getPreviousVisibleIndex(int index) {
    int n = this._trackDCC.getPreviousVisibleIndex(index + this._start);
    if (n < this._start) {
      return -1;
    }

    return n - this._start;
  }

  /// Gets the index of an entity in this collection for which
  /// the cumulated count of previous distances is greater than or equal to
  /// the specified cumulatedDistance. (e.g. return row index for
  /// pixel position).
  ///
  /// * cumulatedDistance - _required_ - The cumulated count of previous distances.
  ///
  /// Returns the index of an entity in this collection for which
  /// the cumulated count of previous distances is greater than or equal to
  /// the specified cumulatedDistance.
  int indexOfCumulatedDistance(double cumulatedDistance) {
    if (this.count == 0 && cumulatedDistance == 0) {
      return 0;
    }

    int n = this._trackDCC.indexOfCumulatedDistance(
        cumulatedDistance + this._trackDCC.getCumulatedDistanceAt(this._start));
    if (n > this.end || n < this._start) {
      return -1;
    }

    return n - this._start;
  }

  /// Gets the cumulated count of previous distances for the entity at the specific index.
  /// (e.g. return pixel position for a row index).
  ///
  /// * index - _required_ - The entity index.
  ///
  /// Returns the cumulated count of previous distances for the
  /// entity at the specific index.
  double getCumulatedDistanceAt(int index) {
    return this._trackDCC.getCumulatedDistanceAt(index + this._start) -
        this._trackDCC.getCumulatedDistanceAt(this.start);
  }

  /// Assigns a collection with nested entities to an item.
  ///
  /// * index - _required_ - The index.
  /// * nestedCollection - _required_ - The nested collection.
  void setNestedDistances(
      int index, _DistanceCounterCollectionBase nestedCollection) {
    this._trackDCC.setNestedDistances(index + this._start, nestedCollection);
  }

  /// Gets the nested entities at a given index.
  ///
  /// If the index does not hold a nested distances collection the method returns null.
  ///
  /// * index - _required_ - The index.
  ///
  /// Returns the nested entities at a given index or null.
  _DistanceCounterCollectionBase getNestedDistances(int index) {
    return this._trackDCC.getNestedDistances(index + this._start);
  }

  /// Gets the distance position of the next entity after a given point.
  ///
  /// * point - _required_ - The point after which the next entity is to be found.
  ///
  /// Returns the distance position of the next entity after a given point.
  double getNextScrollValue(double point) {
    double offset = this._trackDCC.getCumulatedDistanceAt(this.start);
    double d = this._trackDCC.getNextScrollValue(point + offset);
    if (d == double.nan || d < offset || d - offset > this.totalDistance) {
      return double.nan;
    }

    return d - offset;
  }

  /// Gets the distance position of the entity preceding a given point.
  ///
  /// If the point is in between entities, the starting point of the matching entity
  /// is returned.
  ///
  /// * point - _required_ - The point of the entity preceding a given point.
  ///
  /// Returns the distance position of the entity preceding a given point.
  double getPreviousScrollValue(double point) {
    double offset = this._trackDCC.getCumulatedDistanceAt(this._start);
    double d = this._trackDCC.getPreviousScrollValue(point + offset);
    if (d == double.nan || d < offset || d - offset > this.totalDistance) {
      return double.nan;
    }

    return d - offset;
  }

  /// Gets the aligned scroll value which is the starting point of the entity
  /// found at the given distance position.
  ///
  /// * point - _required_ - The point.
  ///
  /// Returns the starting point of the entity found at the given distance position.
  double getAlignedScrollValue(double point) {
    double offset =
        this._trackDCC.getPreviousScrollValue(this._start.toDouble());
    double d = this._trackDCC.getAlignedScrollValue(point + offset);
    if (d == double.nan || d < offset || d - offset > this.totalDistance) {
      return double.nan;
    }

    return d - offset;
  }

  /// This method is not supported for DistanceCounterSubset.
  ///
  /// * treeTableCounterSource - _required_ - The nested tree table visible counter source.
  void connectWithParent(_TreeTableCounterSourceBase treeTableCounterSource) {
    this.connectWithParentBase(treeTableCounterSource);
  }

  /// This method is not supported for DistanceCounterSubset.
  ///
  /// * treeTableCounterSource - _required_ - The nested tree table visible counter source.
  void connectWithParentBase(
      _TreeTableCounterSourceBase treeTableCounterSource) {
    throw Exception("Do not use DistanceCounterSubset as nested collection!");
  }

  /// Inserts entities in the collection from the given index.
  ///
  /// * insertAt - _required_ - The index of the first entity to be inserted.
  /// * count - _required_ - The number of entities to be inserted.
  void insert(int insertAt, int count) {
    this._trackDCC.insert(insertAt + this._start, count);
  }

  /// Removes entities in the collection from the given index.
  ///
  /// * removeAt - _required_ - Index of the first entity to be removed.
  /// * count - _required_ - The number of entities to be removed.
  void remove(int removeAt, int count) {
    this._trackDCC.remove(removeAt + this._start, count);
  }

  /// Gets an object that implements the `Distances` property.
  _DistanceCounterCollectionBase get distances => this._distances;

  /// Gets an distance the `Distances` property.
  _DistanceCounterCollectionBase get _distances => this;
}
