part of datagrid;

/// A collection of entities for which distances need to be counted. The
/// collection provides methods for mapping from a distance position to
/// an entity and vice versa.
///
/// For example, in a scrollable grid control you have rows with different heights.
/// Use this collection to determine the total height for all rows in the grid,
/// quickly determine the row index for a given point and also quickly determine
/// the point at which a row is displayed. This also allows a mapping between the
/// scrollbars value and the rows or columns associated with that value.
/// DistanceCounterCollection internally uses ranges for allocating
/// objects up to the modified entry with the highest index. When you modify
/// the size of an entry the collection ensures that objects are allocated
/// for all entries up to the given index. Entries that are after the modified
/// entry are assumed to have the DefaultSize and will not be allocated.
/// Ranges will only be allocated for those lines that have different sizes.
/// If you do for example only change the size of line 100 to be 10 then the collection
/// will internally create two ranges: Range 1 from 0-99 with DefaultSize and
/// Range 2 from 100-100 with size 10. This approach makes this collection
/// work very efficient with grid scenarios where often many rows have
/// the same height.
class _DistanceRangeCounterCollection extends _DistanceCounterCollectionBase {
  _DistanceLineCounterTree _rbTree;
  double _defaultDistance = 1.0;

  /// Gets the padding distance of the counter collection.
  double paddingDistance = 0;

  /// Initializes a new instance of the DistanceRangeCounterCollection class.
  _DistanceRangeCounterCollection();

  /// Initializes a new instance of the DistanceRangeCounterCollection class.
  ///
  /// * paddingDistance - _required_ - The padding distance.
  _DistanceRangeCounterCollection.fromPaddingDistance(double paddingDistance) {
    this.count = 0;
    var startPos = _DistanceLineCounter(0, 0);
    this._rbTree = _DistanceLineCounterTree(startPos, false);
    this.paddingDistance = paddingDistance;
  }

  /// Clears this instance.
  void clear() {
    this._rbTree.clear();
  }

  /// Gets the raw number of entities (lines, rows or columns).
  ///
  /// Returns the raw number of entities (lines, rows or columns).
  int get count => _count;
  int _count = 0;
  set count(int value) {
    if (value == _count) return;
    _count = value;
  }

  /// Gets the default distance (row height or column width) an entity spans.
  ///
  /// Returns the default distance (row height or column width) an entity spans.
  double get defaultDistance => this._defaultDistance;
  set defaultDistance(double value) {
    this._defaultDistance = value;
  }

  /// Gets the total distance all entities span (e.g. total height of all rows in grid).
  ///
  /// Returns the total distance all entities span (e.g. total height of all rows in grid).
  double get totalDistance {
    double delta = (this.count - this.internalCount).toDouble();
    return this.internalTotalDistance + (delta * this.defaultDistance);
  }

  /// Assigns a collection with nested entities to an item.
  ///
  /// * index - _required_ - The index.
  /// * nestedCollection - _required_ - The nested collection.
  void setNestedDistances(
      int index, _DistanceCounterCollectionBase nestedCollection) {
    this.checkRange("index", 0, this.count - 1, index);

    if (this.getNestedDistances(index) != nestedCollection) {
      if (index >= this.internalCount) {
        this.ensureTreeCount(index + 1);
      }

      _DistanceLineCounterEntry entry = this.split(index);
      this.split(index + 1);

      if (nestedCollection != null) {
        var vcs = _NestedDistanceCounterCollectionSource(
            this, nestedCollection, entry);
        entry.value = vcs;
      } else {
        entry.value = _DistanceLineCounterSource(0, 1);
      }

      entry.invalidateCounterBottomUp(true);
    }
  }

  /// Gets the nested entities at a given index. If the index does not hold
  /// a nested distances collection the method returns null.
  ///
  /// * index - _required_ - The index.
  /// Returns the nested entities at a given index or null.
  _DistanceCounterCollectionBase getNestedDistances(int index) {
    if (index >= this.internalCount) {
      return null;
    }

    this.checkRange("index", 0, this.count - 1, index);
    _LineIndexEntryAt e = this.initDistanceLine(index, false);
    var vcs = e.rbValue as _NestedDistanceCounterCollectionSource;
    if (vcs != null) {
      return vcs.nestedDistances;
    }

    return null;
  }

  /// Invalidates the nested entry of the given index.
  ///
  /// * index - _required_ - The index of the nested entry.
  void invalidateNestedEntry(int index) {
    this.checkRange("index", 0, this.count - 1, index);

    var start = this.getCumulatedDistanceAt(index);
    var end = this.getCumulatedDistanceAt(index + 1);
    var distance = end - start;
    var nested = this.getNestedDistances(index);
    if (nested != null && distance != nested.totalDistance) {
      _LineIndexEntryAt e = this.initDistanceLine(index, false);
      e.rbEntry.invalidateCounterBottomUp(true);
    }
  }

  /// Gets the distance position of the next entity after a given point.
  ///
  /// * point - _required_ - The point after which the next entity is to be found.
  ///
  /// Returns the distance position of the next entity after a given point.
  double getNextScrollValue(double point) {
    int index = this.indexOfCumulatedDistance(point);
    if (index >= this.count) {
      return double.nan;
    }

    double nestedStart = this.getCumulatedDistanceAt(index);
    double delta = point - nestedStart;
    _DistanceCounterCollectionBase nestedDcc = this.getNestedDistances(index);
    if (nestedDcc != null) {
      double r = nestedDcc.getNextScrollValue(delta);
      if (!(r == double.nan) && r >= 0 && r < nestedDcc.totalDistance) {
        return nestedStart + r;
      }
    }

    index = this.getNextVisibleIndex(index);
    if (index >= 0 && index < this.count) {
      return this.getCumulatedDistanceAt(index);
    }

    return double.nan;
  }

  /// Gets the distance position of the entity preceding a given point.
  ///
  /// If the point is in between entities, the starting point of the matching entity
  /// is returned.
  ///
  /// * point - _required_ - The point of the entity preceding a given point.
  ///
  /// Returns the distance position of the entity preceding a given point.
  double getPreviousScrollValue(double point) {
    int index = this.indexOfCumulatedDistance(point);
    double nestedStart = this.getCumulatedDistanceAt(index);
    double delta = point - nestedStart;

    if (delta > 0) {
      _DistanceCounterCollectionBase nestedDcc = this.getNestedDistances(index);
      if (nestedDcc != null) {
        double r = nestedDcc.getPreviousScrollValue(delta);
        if (!(r == double.nan) && r >= 0 && r < nestedDcc.totalDistance) {
          return nestedStart + r;
        }
      }

      return this.getCumulatedDistanceAt(index);
    }

    index = this.getPreviousVisibleIndex(index);

    if (index >= 0 && index < this.count) {
      nestedStart = this.getCumulatedDistanceAt(index);

      _DistanceCounterCollectionBase nestedDcc = this.getNestedDistances(index);
      if (nestedDcc != null) {
        delta = nestedDcc.totalDistance;
        double r = nestedDcc.getPreviousScrollValue(delta);
        if (!(r == double.nan) && r >= 0 && r < nestedDcc.totalDistance) {
          return nestedStart + r;
        }
      }

      return nestedStart;
    }

    return double.nan;
  }

  /// Gets the aligned scroll value which is the starting point of the entity
  /// found at the given distance position.
  ///
  /// * point - _required_ - The point.
  ///
  /// Returns the starting point of the entity found at the given distance position.
  double getAlignedScrollValue(double point) {
    int index = this.indexOfCumulatedDistance(point);
    double nestedStart = this.getCumulatedDistanceAt(index);
    double delta = point - nestedStart;

    if (delta > 0) {
      _DistanceCounterCollectionBase nestedDcc = this.getNestedDistances(index);
      if (nestedDcc != null) {
        double r = nestedDcc.getAlignedScrollValue(delta);
        if (!(r == double.nan) && r >= 0 && r < nestedDcc.totalDistance) {
          return nestedStart + r;
        }
      }
    }

    return this.getCumulatedDistanceAt(index);
  }

  /// Connects a nested distance collection with a parent.
  ///
  /// * treeTableCounterSource - _required_ - The `_TreeTableCounterSourceBase`
  /// representing the nested tree table visible counter source.
  void connectWithParent(_TreeTableCounterSourceBase treeTableCounterSource) {
    this._rbTree.tag = treeTableCounterSource;
  }

  /// Gets the next visible index.
  ///
  /// Skip subsequent entities for which the distance is 0.0 and return the next entity.
  ///
  /// * index - _required_ - The index.
  ///
  /// Returns the next visible index from the given index.
  int getNextVisibleIndex(int index) {
    this.checkRange("index", 0, this.count - 1, index);

    if (index >= this.internalCount) {
      return index + 1;
    }

    _LineIndexEntryAt e = this.initDistanceLine(index + 1, true);
    if (e.rbValue.singleLineDistance > 0) {
      if (index - e.rbEntryPosition.lineCount < e.rbValue.lineCount) {
        return index + 1;
      }
    }

    e.rbEntry = this._rbTree.getNextVisibleEntry(e.rbEntry);
    if (e.rbEntry == null) {
      if (this.internalCount < this.count) {
        return this.internalCount;
      }

      return -1;
    }

    e.rbEntryPosition = e.rbEntry.getCounterPosition;
    return e.rbEntryPosition.lineCount;
  }

  /// Gets the previous visible index.
  ///
  /// Skip previous entities for which the distance is 0.0 and return the previous entity.
  ///
  /// * index - _required_ - The index.
  /// Returns the previous visible index from the given index.
  int getPreviousVisibleIndex(int index) {
    this.checkRange("index", 0, this.count, index);

    if (index > this.internalCount || index == 0) {
      return index - 1;
    }

    _LineIndexEntryAt e = this.initDistanceLine(index - 1, false);
    if (e.rbValue.singleLineDistance > 0) {
      return index - 1;
    }

    e.rbEntry = this._rbTree.getPreviousVisibleEntry(e.rbEntry);
    if (e.rbEntry == null) {
      return -1;
    }

    e.rbEntryPosition = e.rbEntry.getCounterPosition;
    e.rbValue = e.rbEntry.value;
    return e.rbEntryPosition.lineCount + e.rbValue.lineCount - 1;
  }

  /// Gets the index of an entity in this collection for which
  /// the cumulated count of previous distances is greater than or equal to
  /// the specified cumulatedDistance. (e.g. return row index for
  /// pixel position).
  ///
  /// * cumulatedDistance - _required_ - The cumulated count of previous distances.
  ///
  /// Returns the index of an entity in this collection for which
  /// the cumulated count of previous distances is greater than or equal to
  /// the specified cumulatedDistance.
  int indexOfCumulatedDistance(double cumulatedDistance) {
    if (cumulatedDistance <
        (this._rbTree.getStartCounterPosition() as _DistanceLineCounter)
            .distance) {
      return -1;
    }
    return this._indexOfCumulatedDistance(cumulatedDistance);
  }

  int _indexOfCumulatedDistance(double cumulatedDistance) {
    if (this.internalCount == 0) {
      return (cumulatedDistance / this.defaultDistance).floor();
    }

    int delta = 0;
    double internalTotalDistance =
        this.internalTotalDistance - this.paddingDistance;
    if (cumulatedDistance >= internalTotalDistance) {
      delta = ((cumulatedDistance - internalTotalDistance).floor() ~/
          this.defaultDistance);
      cumulatedDistance = internalTotalDistance;
      return this.internalCount + delta;
    }

    var searchPosition = _DistanceLineCounter(cumulatedDistance, 0);
    _DistanceLineCounterEntry rbEntry = this
        ._rbTree
        .getEntryAtCounterPositionWithThreeArgs(
            searchPosition, _DistanceLineCounterKind.distance, false);
    _DistanceLineCounterSource rbValue = rbEntry.value;
    _DistanceLineCounter rbEntryPosition = rbEntry.getCounterPosition;
    if (rbValue.singleLineDistance > 0) {
      delta = ((cumulatedDistance - rbEntryPosition.distance).floor() ~/
          rbValue.singleLineDistance);
    }

    return rbEntryPosition.lineCount + delta;
  }

  /// Gets the cumulated count of previous distances for the entity at the specific index.
  /// (e.g. return pixel position for a row index).
  ///
  /// * index - _required_ - The entity index.
  ///
  /// Returns the cumulated count of previous distances for the
  /// entity at the specific index.
  double getCumulatedDistanceAt(int index) {
    return this._getCumulatedDistanceAt(index);
  }

  double _getCumulatedDistanceAt(int index) {
    this.checkRange("index", 0, this.count + 1, index);

    if (index == this.count + 1) {
      return this.totalDistance;
    }
    if (index == 0) {
      return 0;
    }
    int count = this.internalCount;
    if (count == 0) {
      return index * this.defaultDistance;
    }
    if (index >= count) {
      int delta = index - count;
      _DistanceLineCounter counter = this._rbTree.getCounterTotal();
      return counter.distance + (delta * this.defaultDistance);
    } else {
      _LineIndexEntryAt e = this.initDistanceLine(index, false);
      e.rbEntryPosition = e.rbEntry.getCounterPosition;
      return e.rbEntryPosition.distance +
          ((index - e.rbEntryPosition.lineCount) *
              e.rbValue.singleLineDistance);
    }
  }

  /// Gets the distance for an entity from the given index.
  ///
  /// * index - _required_ - The index for the entity.
  ///
  /// Returns the distance for an entity from the given index.

  operator [](index) {
    this.checkRange("index", 0, this.count - 1, index);

    if (index >= this.internalCount) {
      return this.defaultDistance;
    }

    _LineIndexEntryAt e = this.initDistanceLine(index, false);
    return e.rbValue.singleLineDistance;
  }

  operator []=(index, double value) {
    this.checkRange("index", 0, this.count - 1, index);
    if (value < 0) {
      throw Exception("value must not be negative.");
    }

    if (!(value == (this[index]))) {
      if (index >= this.internalCount) {
        this.ensureTreeCount(this.count);
      }

      _DistanceLineCounterEntry entry = this.split(index);
      this.split(index + 1);
      entry.value.singleLineDistance = value;
      entry.invalidateCounterBottomUp(true);
    }
  }

  /// Hides a specified range of entities (lines, rows or columns).
  ///
  /// * from - _required_ - The index for the first entity.
  /// * to - _required_ - The raw index for the last entity.
  /// * distance - _required_ - The distance.
  void setRange(int from, int to, double distance) {
    this.checkRange("from", 0, this.count - 1, from);
    this.checkRange("to", 0, this.count - 1, to);

    if (from == to) {
      this[from] = distance;
      return;
    }

    if (from >= this.internalCount && (distance == this._defaultDistance)) {
      return;
    }

    int count = to - from + 1;
    this.ensureTreeCount(from);
    int n = this.removeHelper(from, count);
    _DistanceLineCounterEntry rb = this.createTreeTableEntry(distance, count);
    this._rbTree.insert(n, rb);
    this.merge(rb, true);
  }

  /// Resets the range by restoring the default distance for all entries in the specified range.
  ///
  /// * from - _required_ - The index for the first entity.
  /// * to - _required_ - The raw index for the last entity.
  void resetRange(int from, int to) {
    this.checkRange("from", 0, this.count - 1, from);
    this.checkRange("to", 0, this.count - 1, to);

    if (from >= this.internalCount) {
      return;
    }

    int count = to - from + 1;
    this.setRange(from, count, this._defaultDistance);
  }

  /// Inserts entities in the collection from the given index.
  ///
  /// * insertAt - _required_ - The index of the first entity to be inserted.
  /// * count - _required_ - The number of entities to be inserted.
  void insert(int insertAt, int count) {
    this.insertBase(insertAt, count, this.defaultDistance);
  }

  /// Inserts entities in the collection from the given index.
  ///
  /// * insertAt - _required_ - The index of the first entity to be inserted.
  /// * count - _required_ - The number of entities to be inserted.
  /// * distance - _required_ - The distance to be set.
  void insertBase(int insertAt, int count, double distance) {
    this.count += count;

    if (insertAt >= this.internalCount && distance == this._defaultDistance) {
      return;
    }

    this.ensureTreeCount(insertAt);

    _LineIndexEntryAt e = this.initDistanceLine(insertAt, false);
    if (e.rbValue.singleLineDistance == distance) {
      e.rbValue.lineCount += count;
      e.rbEntry.invalidateCounterBottomUp(true);
    } else {
      _DistanceLineCounterEntry rbEntry0 = this.split(insertAt);
      _DistanceLineCounterEntry entry =
          this.createTreeTableEntry(distance, count);
      if (rbEntry0 == null) {
        this._rbTree.add(entry);
      } else {
        this._rbTree.insert(this._rbTree.indexOf(rbEntry0), entry);
      }
      this.merge(entry, true);
    }
  }

  /// Removes entities in the collection from the given index.
  ///
  /// * removeAt - _required_ - Index of the first entity to be removed.
  /// * count - _required_ - The number of entities to be removed.
  void remove(int removeAt, int count) {
    this.count -= count;

    if (removeAt >= this.internalCount) {
      return;
    }

    int n = this.removeHelper(removeAt, count);
    if (n > 0) {
      this.merge(this._rbTree[n - 1], false);
    }
  }

  _DistanceLineCounterEntry createTreeTableEntry(double distance, int count) {
    var entry = _DistanceLineCounterEntry()
      ..value = _DistanceLineCounterSource(distance, count)
      ..tree = this._rbTree;
    return entry;
  }

  int removeHelper(int removeAt, int count) {
    if (removeAt >= this.internalCount) {
      return this._rbTree.getCount();
    }

    _DistanceLineCounterEntry entry = this.split(removeAt);
    this.split(removeAt + count);
    int n = this._rbTree.indexOf(entry);

    var toDelete = List<_DistanceLineCounterEntry>();

    int total = 0;
    while (total < count && entry != null) {
      total += entry.value.lineCount;
      toDelete.add(entry);
      entry = this._rbTree.getNextEntry(entry);
    }

    for (int l = 0; l < toDelete.length; l++) {
      this._rbTree.remove(toDelete[l]);
    }

    return n;
  }

  _DistanceLineCounterEntry split(int index) {
    if (index >= this.internalCount) {
      return null;
    }

    _LineIndexEntryAt e = this.initDistanceLine(index, true);
    if (e.rbEntryPosition.lineCount != index) {
      int count1 = index - e.rbEntryPosition.lineCount;
      int count2 = e.rbValue.lineCount - count1;

      e.rbValue.lineCount = count1;

      _DistanceLineCounterEntry rbEntry2 =
          this.createTreeTableEntry(e.rbValue.singleLineDistance, count2);
      this._rbTree.insert(this._rbTree.indexOf(e.rbEntry) + 1, rbEntry2);
      e.rbEntry.invalidateCounterBottomUp(true);
      return rbEntry2;
    }

    return e.rbEntry;
  }

  void merge(_DistanceLineCounterEntry entry, bool checkPrevious) {
    _DistanceLineCounterSource value = entry.value;
    _DistanceLineCounterEntry previousEntry;
    if (checkPrevious) {
      previousEntry = this._rbTree.getPreviousEntry(entry);
    }

    _DistanceLineCounterEntry nextEntry = this._rbTree.getNextEntry(entry);

    bool dirty = false;
    if (previousEntry != null &&
        previousEntry.value.singleLineDistance == value.singleLineDistance) {
      value.lineCount += previousEntry.value.lineCount;
      this._rbTree.remove(previousEntry);
      dirty = true;
    }

    if (nextEntry != null &&
        nextEntry.value.singleLineDistance == value.singleLineDistance) {
      value.lineCount += nextEntry.value.lineCount;
      this._rbTree.remove(nextEntry);
      dirty = true;
    }

    if (dirty) {
      entry.invalidateCounterBottomUp(true);
    }
  }

  void ensureTreeCount(int count) {
    int treeCount = this.internalCount;
    int insert = count - treeCount;
    if (insert > 0) {
      _DistanceLineCounterEntry entry =
          this.createTreeTableEntry(this.defaultDistance, insert);
      this._rbTree.add(entry);
    }
  }

  int get internalCount {
    if (this._rbTree.getCount() == 0) {
      return 0;
    }

    _DistanceLineCounter counter = this._rbTree.getCounterTotal();
    return counter.lineCount;
  }

  set internalCount(int value) {
    if (value >= this.internalCount) {
      this.ensureTreeCount(value);
    }
  }

  double get internalTotalDistance {
    int treeCount = this.internalCount;
    if (treeCount == 0) {
      return this.paddingDistance;
    }

    _DistanceLineCounter counter = this._rbTree.getCounterTotal();
    return counter.distance + this.paddingDistance;
  }

  void checkRange(String paramName, int from, int to, int actualValue) {
    if (actualValue < from || actualValue > to) {
      throw Exception(
          '$paramName ,${actualValue.toString()} out of range ${from.toString()} to ${to.toString()}');
    }
  }

  _LineIndexEntryAt initDistanceLine(
      int lineIndex, bool determineEntryPosition) {
    var e = _LineIndexEntryAt()
      ..searchPosition = _DistanceLineCounter(0, lineIndex);
    e.rbEntry = this._rbTree.getEntryAtCounterPositionWithThreeArgs(
        e.searchPosition, _DistanceLineCounterKind.lines, false);
    e.rbValue = e.rbEntry.value;
    e.rbEntryPosition = null;
    if (determineEntryPosition) {
      e.rbEntryPosition = e.rbValue.lineCount > 1
          ? e.rbEntry.getCounterPosition
          : e.searchPosition;
    }

    return e;
  }
}

/// Initializes the LineIndexEntryAt class
class _LineIndexEntryAt {
  _DistanceLineCounter searchPosition;
  _DistanceLineCounterEntry rbEntry;
  _DistanceLineCounterSource rbValue;
  _DistanceLineCounter rbEntryPosition;
}

/// An object that maintains a collection of nested distances and wires
/// it to a parent distance collection.
///
/// The object is used by the DistanceCounterCollection.SetNestedDistances method to associated
/// the nested distances with an index in the parent collection.
class _NestedDistanceCounterCollectionSource
    extends _DistanceLineCounterSource {
  _DistanceCounterCollectionBase _nestedDistances;
  _DistanceCounterCollectionBase _parentDistances;
  _DistanceLineCounterEntry _entry;

  /// Initializes a new instance of the NestedDistanceCounterCollectionSource class.
  ///
  /// * parentDistances - _required_ - The parent distances.
  /// * nestedDistances - _required_ - The nested distances.
  /// * entry - _required_ - The entry.
  _NestedDistanceCounterCollectionSource(
      _DistanceCounterCollectionBase parentDistances,
      _DistanceCounterCollectionBase nestedDistances,
      _DistanceLineCounterEntry entry)
      : super(0, 1) {
    this._parentDistances = parentDistances;
    this._nestedDistances = nestedDistances;
    this.entry = _entry;

    if (nestedDistances != null) {
      nestedDistances.connectWithParent(this);
    }
  }

  /// Gets the counter entry.
  ///
  /// Returns the entry.
  _DistanceLineCounterEntry get entry => this._entry;

  /// Sets the counter entry.
  set entry(_DistanceLineCounterEntry value) {
    this._entry = value;
  }

  /// Gets the distance of a single line.
  ///
  /// Returns the single line distance.
  @override
  double get singleLineDistance =>
      this._nestedDistances != null ? this._nestedDistances.totalDistance : 0;

  /// Sets the distance of a single line.
  set singleLineDistance(double value) {
    throw Exception();
  }

  /// Gets the parent distances.
  ///
  /// Returns the parent distances.
  _DistanceCounterCollectionBase get parentDistances => this._parentDistances;

  /// Gets the nested distances.
  ///
  /// Returns the nested distances.
  _DistanceCounterCollectionBase get nestedDistances => this._nestedDistances;

  /// Marks all counters dirty in this object and parent nodes.
  @override
  void invalidateCounterBottomUp() {
    if (this.entry != null) {
      this.entry.invalidateCounterBottomUp(true);
    }
  }

  /// Returns the `_TreeTableVisibleCounter` object with counters.
  @override
  _TreeTableCounterBase getCounter() {
    return _DistanceLineCounter(
        this._nestedDistances == null ? 0 : this._nestedDistances.totalDistance,
        1);
  }

  /// Returns a string describing the state of the object.
  @override
  String toString() {
    return runtimeType.toString() +
        ("LineCount = ${this.lineCount}, SingleLineDistance = ${this._singleLineDistance} )");
  }
}

/// An object that counts objects that are marked "Visible". It implements
/// the _TreeTableCounterSourceBase interface and creates a [DistanceLineCounter].
class _DistanceLineCounterSource extends _TreeTableCounterSourceBase {
  double _singleLineDistance = 0.0;
  int _lineCount = 0;

  /// Initializes a new instance of the [DistanceLineCounterSource] class.
  ///
  /// * visibleCount - _required_ - The visible count.
  /// * lineCount - _required_ - The line count.
  _DistanceLineCounterSource(double visibleCount, int lineCount) {
    this._singleLineDistance = visibleCount;
    this._lineCount = lineCount;
  }

  /// Gets the line count.
  ///
  /// Returns the line count.
  int get lineCount => this._lineCount;

  /// Sets the line count.
  set lineCount(int value) {
    this._lineCount = value;
  }

  /// Gets the distance of a single line.
  ///
  /// Returns the single line distance.
  double get singleLineDistance => this._singleLineDistance;

  /// Sets the distance of a single line.
  set singleLineDistance(double value) {
    this._singleLineDistance = value;
  }

  /// Returns a string describing the state of the object.
  @override
  String toString() {
    return runtimeType.toString() +
        (" LineCount = ${this.lineCount}, SingleLineDistance = ${this.singleLineDistance}");
  }

  /// Returns the `_TreeTableVisibleCounter` object with counters.
  @override
  _TreeTableCounterBase getCounter() {
    return _DistanceLineCounter(
        this.singleLineDistance * this._lineCount, this._lineCount);
  }

  /// Marks all counters dirty in this object and child nodes.
  ///
  /// * notifyCounterSource - _required_ - A boolean value indicating whether to notify the counter
  /// source.
  @override
  void invalidateCounterTopDown(bool notifyCounterSource) {}

  /// Marks all counters dirty in this object and parent nodes.
  @override
  void invalidateCounterBottomUp() {}
}

/// A collection of integers used to specify various counter kinds.
class _DistanceLineCounterKind {
  /// Visible Counter.
  static final int distance = 0x8000;

  /// Line Counter.
  static final int lines = 1;

  /// Prevents a default instance of the [DistanceLineCounterKind] class from being created.
  _DistanceLineCounterKind();
}

/// A counter that counts objects that are marked "Visible".
class _DistanceLineCounter extends _TreeTableCounterBase {
  double _distance = 0.0;
  int _lineCount = 0;

  /// Initializes a new instance of the [DistanceLineCounter] class.
  ///
  /// * distance - _required_ - The distance.
  /// * lineCount - _required_ - The line count.
  _DistanceLineCounter(double distance, int lineCount) {
    this._distance = distance;
    this._lineCount = lineCount;
  }

  /// Initializes an empty DistanceLineCounter that represents zero visible elements.
  _DistanceLineCounter.empty() {
    this._distance = 0;
    this._lineCount = 0;
  }

  /// Gets the line count.
  ///
  /// Returns the line count.
  int get lineCount => this._lineCount;

  /// Gets the distance.
  ///
  /// Returns the distance.
  double get distance => this._distance;
  set distance(double value) {
    if (value == _distance) return;
    _distance = value;
  }

  /// Gets the counter kind.
  ///
  /// Returns the counter kind.
  int get kind => 0;

  /// Returns the value of the counter. A cookie specifies
  /// a specific counter type.
  ///
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns the value of the counter.
  double getValue(int cookie) {
    if (cookie == _DistanceLineCounterKind.lines) {
      return this._lineCount.toDouble();
    }
    return this._distance;
  }

  /// Combines this counter object with another counter and returns a new object. A cookie can specify
  /// a specific counter type.
  ///
  /// * other - _required_ - Counter total
  /// * cookie - _required_ - cookie value.
  ///
  /// Returns the new object
  _TreeTableCounterBase combine(_TreeTableCounterBase other, int cookie) {
    return this.combineBase(other as _DistanceLineCounter, cookie);
  }

  /// Combines the counter values of this counter object with the values of another counter object
  /// and returns a new counter object.
  ///
  /// * other - _required_ - The other line counter.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns a new object which is the combination of the counter values of this counter
  /// object with the values of another counter object.
  _DistanceLineCounter combineBase(_DistanceLineCounter other, int cookie) {
    if (other == null || other.isEmpty(_MathHelper.maxvalue)) {
      return this;
    }

    if (this.isEmpty(_MathHelper.maxvalue)) {
      return other;
    }

    double addedvalue = (this.distance + other.distance);
    return _DistanceLineCounter((addedvalue), this.lineCount + other.lineCount);
  }

  /// Compares this counter with another counter. A cookie can specify
  /// a specific counter type.
  ///
  /// * other - _required_ - The other.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns the compared value
  double compare(_TreeTableCounterBase other, int cookie) {
    return this.compareBase(other as _DistanceLineCounter, cookie);
  }

  /// Compares this counter with another counter. A cookie can specify
  /// a specific counter type.
  ///
  /// * other - _required_ - The other counter.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns a value indicating the comparison of the current object and the given
  /// object.
  double compareBase(_DistanceLineCounter other, int cookie) {
    if (other == null) {
      return 0;
    }

    int cmp = 0;

    if ((cookie & _DistanceLineCounterKind.distance) != 0) {
      cmp = this.distance.compareTo(other.distance);
    }

    if (cmp == 0 && (cookie & _DistanceLineCounterKind.lines) != 0) {
      cmp = this.lineCount.compareTo(other.lineCount);
    }

    return cmp.toDouble();
  }

  /// Indicates whether the counter object is empty. A cookie can specify
  /// a specific counter type.
  ///
  /// * cookie - _required_ - The cookie.
  ///
  /// True if the specified cookie is empty, otherwise false.

  bool isEmpty(int cookie) {
    return this.compare(_DistanceLineCounter.empty(), cookie) == 0;
  }

  /// A `string` that represents the current `object`.
  ///
  /// Returns a `string` that represents the current `object`.
  @override
  String toString() {
    return "$runtimeType LineCount = ${this.lineCount} Distance = ${this.distance}";
  }
}

/// Initializes a new instance of the DistanceLineCounterTree class.
class _DistanceLineCounterTree extends _TreeTableWithCounter {
  /// Initializes a new instance of the DistanceLineCounterTree class.
  ///
  /// * startPos - _required_ - The start position.
  /// * sorted - _required_ - A boolean value indicating whether sorted.
  _DistanceLineCounterTree(_DistanceLineCounter startPos, bool sorted)
      : super(startPos, sorted);

  /// Gets the total number of counters.
  ///
  /// Returns the total number of counters.
  _DistanceLineCounter getCounterTotal() {
    return super.getCounterTotal() as _DistanceLineCounter;
  }

  /// Gets the previous entry.
  ///
  /// * current - _required_ - The current counter entry.
  ///
  /// Returns the previous entry.
  _DistanceLineCounterEntry getPreviousEntry(Object current) {
    return super.getPreviousEntry(current) as _DistanceLineCounterEntry;
  }

  /// Gets the next entry.
  ///
  /// * current - _required_ - The current counter entry.
  ///
  /// Returns the next entry.
  _DistanceLineCounterEntry getNextEntry(Object current) {
    return super.getNextEntry(current) as _DistanceLineCounterEntry;
  }

  /// Returns an entry at the specified counter position. A cookie defines the type of counter.
  ///
  /// * searchPosition - _required_ - The search position.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns an entry at the specified counter position.
  _DistanceLineCounterEntry getEntryAtCounterPosition(
      Object searchPosition, int cookie) {
    return super.getEntryAtCounterPosition(searchPosition, cookie)
        as _DistanceLineCounterEntry;
  }

  /// Returns an entry at the specified counter position. A cookie defines the type of counter.
  ///
  /// * searchPosition - _required_ - The search position.
  /// * cookie - _required_ - The cookie.
  /// * preferLeftMost - _required_ - A boolean value indicating whether the leftmost entry should be returned if multiple tree elements have the
  /// same SearchPosition.
  ///
  /// Returns an entry at the specified counter position.

  _DistanceLineCounterEntry getEntryAtCounterPositionWithThreeArgs(
      Object searchPosition, int cookie, bool preferLeftMost) {
    return super.getEntryAtCounterPositionwithThreeParameter(
        searchPosition, cookie, preferLeftMost) as _DistanceLineCounterEntry;
  }

  /// Returns the subsequent entry in the collection for which the specific counter is not empty.
  /// A cookie defines the type of counter.
  ///
  /// * current - _required_ - The current counter entry.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns the subsequent entry in the collection for which the specific counter is not empty.
  _DistanceLineCounterEntry getNextNotEmptyCounterEntry(
      Object current, int cookie) {
    return super.getNextNotEmptyCounterEntry(current, cookie)
        as _DistanceLineCounterEntry;
  }

  /// Returns the previous entry in the collection for which the specific counter is not empty.
  /// A cookie defines the type of counter.
  ///
  /// * current - _required_ - The current counter entry.
  /// * cookie - _required_ - The cookie.
  ///
  /// Returns the previous entry in the collection for which the specific counter is not empty.
  _DistanceLineCounterEntry getPreviousNotEmptyCounterEntry(
      Object current, int cookie) {
    return super.getPreviousNotEmptyCounterEntry(current, cookie)
        as _DistanceLineCounterEntry;
  }

  /// The next entry in the collection for which CountVisible counter is not empty.
  ///
  /// * current - _required_ - The current counter entry.
  ///
  /// Returns the next entry in the collection for which CountVisible counter is not empty.
  _DistanceLineCounterEntry getNextVisibleEntry(Object current) {
    return super.getNextVisibleEntry(current) as _DistanceLineCounterEntry;
  }

  /// The previous entry in the collection for which CountVisible counter is not empty.
  ///
  /// * current - _required_ - The current counter entry.
  ///
  /// Returns the previous entry in the collection for which CountVisible counter is not empty.
  _DistanceLineCounterEntry getPreviousVisibleEntry(Object current) {
    return super.getPreviousVisibleEntry(current) as _DistanceLineCounterEntry;
  }

  /// Gets the distance line counter entry for the given index.
  ///
  /// * index - _required_ - The index of the counter entry needed.
  ///
  /// Returns the distance line counter entry for the given index.
  operator [](index) {
    return super[index] as _DistanceLineCounterEntry;
  }

  /// Sets the distance line counter entry for the given index.
  operator []=(index, value) {
    super[index] = value;
  }

  /// Inserts a `DistanceLineCounterEntry` object at the specified index.
  ///
  /// * index - _required_ - The index.
  /// * value - _required_ - The object to be inserted.
  void insert(int index, Object value) {
    super.insert(index, value);
  }

  /// Removes the given object from the tree.
  ///
  /// * value - _required_ - The object to be removed.
  @override
  bool remove(Object value) {
    return super.remove(value);
  }

  /// Indicates whether the given object belongs to the tree.
  ///
  /// * value - _required_ - The object to be queried.
  ///
  /// Returns `True` if tree contains the specified object, otherwise `false`.

  bool contains(Object value) {
    return super.contains(value);
  }

  /// Gets the position of an object in the tree.
  ///
  /// * value - _required_ - The object whose index is to be obtained.
  ///
  /// Returns the position of an object in the tree.
  int indexOf(Object value) {
    return super.indexOf(value);
  }

  /// Appends the given object.
  ///
  /// * value - _required_ - The object to be appended.
  ///
  /// Returns the position of the object appended.
  int add(Object value) {
    return super.add(value);
  }
}

/// Initializes a new instance of the [DistanceLineCounterTree] class.
class _DistanceLineCounterEntry extends _TreeTableWithCounterEntry {
  /// Gets the distance line counter source.
  ///
  /// Returns the distance line counter source.
  _DistanceLineCounterSource get value =>
      super.value as _DistanceLineCounterSource;

  /// Sets the distance line counter source.
  set value(Object value) {
    super.value = value;
  }

  /// The cumulative position of this node.
  ///
  /// Returns the cumulative position of this node.
  _DistanceLineCounter get getCounterPosition {
    return super.getCounterPosition as _DistanceLineCounter;
  }
}
