part of datagrid;

/// The LineScrollAxis implements scrolling only for whole lines.
///
/// You can hide lines and LineScrollAxis provides a mapping mechanism between the
/// index of the line and the scroll index and vice versa. Hidden lines
/// are not be counted when the scroll index is determined for a line.
/// The LineScrollAxis does not support scrolling in between lines (pixel scrolling).
/// This can be of advantage if you have a large number of lines with varying
/// line sizes. In such case the LineScrollAxis does not need to maintain
/// a collection that tracks line sizes whereas the `PixelScrollAxis` does need to.
class _LineScrollAxis extends _ScrollAxisBase {
  /// distances holds the visible lines. Each visible line
  /// has a distance of 1.0. Hidden lines have a distance of 0.0.
  _DistanceCounterCollectionBase _distances;
  double _headerExtent = 0.0;
  double _footerExtent = 0.0;
  double _defaultLineSize = 0.0;
  double _viewSize = 0.0;

  /// Initializes a new instance of the LineScrollAxis class.
  ///
  /// * scrollBar - _required_ - The state of the scrollbar.
  /// * scrollLinesHost - _required_ - The scroll lines host.
  _LineScrollAxis(_ScrollBarBase scrollBar, _LineSizeHostBase scrollLinesHost)
      : super(scrollBar, scrollLinesHost) {
    var distancesHost = scrollLinesHost as _DistancesHostBase;
    this._distances = distancesHost != null
        ? distancesHost.distances
        : new _DistanceRangeCounterCollection();
    if (scrollLinesHost != null) {
      scrollLinesHost.initializeScrollAxis(this);
    }
    this._distances.defaultDistance = 1.0;
  }

  /// Gets the line count.
  ///
  /// Returns the line count.
  @override
  int get lineCount => this._distances.count;

  /// Sets the line count.
  set lineCount(int value) {
    if (this.lineCount != value) {
      this._distances.count = value;
      this.updateScrollbar();
    }
  }

  /// Gets the default size of lines.
  ///
  /// Returns the default size of lines.
  @override
  double get defaultLineSize => this._defaultLineSize;

  /// Sets the default size of lines.
  set defaultLineSize(double value) {
    if (this.defaultLineSize != value) {
      this._defaultLineSize = value;
      this.updateDistances();
      this.updateScrollbar();
    }
  }

  /// Gets a value indicating whether this axis supports pixel scrolling.
  ///
  /// Returns `true` if this instance supports pixel scrolling; otherwise, `false`.
  bool get isPixelScroll => false;

  /// Updates the line size for visible lines to be "1" for LineScrollAxis
  void updateDistances() {
    int repeatSizeCount;

    for (int index = 0; index < this.lineCount; index++) {
      bool hide = scrollLinesHost.getHidden(index, repeatSizeCount)[0];
      repeatSizeCount = scrollLinesHost.getHidden(index, repeatSizeCount)[1];
      var value = hide == true ? 0 : 1.0;
      int rangeTo =
          getRangeToHelper(index, this.lineCount - 1, repeatSizeCount);
      this._distances.setRange(index, rangeTo, value);
      index = rangeTo;
    }
  }

  /// Gets the distances collection which is used internally for mapping from a point position to
  /// a line index and vice versa.
  ///
  /// Returns the distances collection which is used internally for mapping from a point position
  /// to a line index and vice versa.
  _DistanceCounterCollectionBase get distances => this._distances;

  /// Gets the total extent of all line sizes.
  ///
  /// Returns the total extent of all line sizes.
  double get totalExtent => this.distances.totalDistance;

  /// Gets the header extent. This is total height (or width) of the header lines.
  ///
  /// Returns the header extent.
  double get headerExtent => this._headerExtent;

  /// Sets the header line count.
  ///
  /// * value - _required_ - The value.
  @override
  void setHeaderLineCount(int value) {
    double size = 0.0;
    int lines = 0;
    while (lines < value) {
      size += this.getLineSize(lines++);
    }
    this._headerExtent = size;
  }

  /// Gets the footer extent. This is total height (or width) of the footer lines.
  ///
  /// Returns the footer extent.
  @override
  double get footerExtent => this._footerExtent;

  /// Sets the footer line count.
  ///
  /// * value - _required_ - The value.
  @override
  void setFooterLineCount(int value) {
    double size = 0.0;
    int lines = 0;
    int index = this.lineCount - 1;
    while (lines < value) {
      size += this.getLineSize(index--);
      lines++;
    }

    this._footerExtent = size;
  }

  /// Gets the index of the first visible Line in the Body region.
  ///
  /// Returns the index of the first visible Line in the Body region.
  @override
  int get scrollLineIndex => this.scrollBarValueToLineIndex(scrollBar.value);

  /// Sets the index of the first visible Line in the Body region.
  set scrollLinesIndex(int value) {
    this.setScrollLineIndex(value, 0.0);
  }

  /// Gets the index of the scroll line.
  ///
  /// * scrollLineIndex - _required_ - Index of the scroll line.
  /// * scrollLineDelta - _required_ - The scroll line delta.
  /// * isRightToLeft - _required_ - The boolean value used to calculate visible columns in right to left mode.
  @override
  List getScrollLineIndex(int scrollLineIndex, double scrollLineDelta,
      [bool isRightToLeft = false]) {
    scrollLineIndex = this.scrollBarValueToLineIndex(scrollBar.value);
    scrollLineDelta = 0.0;
    return [scrollLineIndex, scrollLineDelta];
  }

  /// Sets the index of the scroll line.
  ///
  /// * scrollLineIndex - _required_ - The index of the scroll line.
  /// * scrollLineDelta - _required_ - The scroll line delta.
  @override
  void setScrollLineIndex(int scrollLineIndex, double scrollLineDelta) {
    scrollLineIndex =
        min(max(this._distances.count - 1, 0), max(0, scrollLineIndex));
    scrollBar.value = this.lineIndexToScrollBarValue(scrollLineIndex);
    this.resetVisibleLines();
  }

  /// Sets the hidden state of the lines.
  ///
  /// * from - _required_ - The start index of the line.
  /// * to - _required_ - The end index of the line.
  /// * hide - _required_ - A boolean value indicating whether to hide the lines. if set to true - [hide].
  @override
  void setLineHiddenState(int from, int to, bool hide) {
    this._distances.setRange(from, to, hide ? 0.0 : 1.0);
  }

  /// Sets the size of the lines for the given range of lines. Will do nothing for a `LineScrollAxis`.
  ///
  /// * from - _required_ - The start index of the line.
  /// * to - _required_ - The end index of the line.
  /// * size - _required_ - The line size.
  @override
  void setLineSize(int from, int to, double size) {}

  /// Gets the view size of the (either height or width) of the parent control.
  ///
  /// Normally the ViewSize is the same as `ScrollAxisBase.RenderSize`. Only if the parent control
  /// has more space then needed to display all lines, the ViewSize will be less. In
  /// such case the ViewSize is the total height for all lines.
  ///
  /// Returns the size of the view.
  double get viewSize {
    if (scrollBar.value + scrollBar.largeChange > scrollBar.maximum) {
      return this.viewSize;
    } else {
      return this.renderSize;
    }
  }

  int scrollBarValueToLineIndex(double sbValue) {
    return this._distances.indexOfCumulatedDistance(sbValue);
  }

  double lineIndexToScrollBarValue(int lineIndex) {
    return this._distances.getCumulatedDistanceAt(lineIndex);
  }

  /// Initialize scrollbar properties from line count in header, footer and body.
  @override
  void updateScrollbar() {
    this.setHeaderLineCount(this.headerLineCount);
    this.setFooterLineCount(this.footerLineCount);

    _ScrollBarBase sb = scrollBar;
    bool isMinimum = sb.minimum == sb.value;
    sb.minimum = this.headerLineCount.toDouble();
    var maximum = this._distances.totalDistance - footerLineCount;
    sb.maximum = max(maximum, 0);
    sb.smallChange = 1;

    sb.largeChange = this.determineLargeChange().toDouble();
    sb.value =
        isMinimum ? sb.minimum : max(sb.minimum, min(sb.maximum, sb.value));
  }

  int determineLargeChange() {
    double sbValue = scrollBar.maximum;
    double abortSize = scrollPageSize;
    int count = 0;
    this._viewSize = 0;

    while (sbValue >= scrollBar.minimum) {
      int lineIndex = this.scrollBarValueToLineIndex(sbValue);
      double size = getLineSize(lineIndex);
      if (this._viewSize + size > abortSize) {
        break;
      }

      count++;
      sbValue--;
      this._viewSize += size;
    }

    this._viewSize += this.footerExtent + this.headerExtent;

    return count;
  }

  /// Gets the index of the previous scroll line.
  ///
  /// * index - _required_ - The current index of the scroll line.
  /// Returns the index of the previous scroll line.
  @override
  int getPreviousScrollLineIndex(int index) {
    if (this._distances.count > index) {
      return this._distances.getPreviousVisibleIndex(index);
    } else {
      return 0;
    }
  }

  /// Gets the index of the next scroll line.
  ///
  /// * index - _required_ - The current index of the scroll line.
  /// Returns the index of the next scroll line.
  @override
  int getNextScrollLineIndex(int index) {
    if (this._distances.count > index) {
      return this._distances.getNextVisibleIndex(index);
    } else {
      return 0;
    }
  }

  /// Scrolls to next page.
  @override
  void scrollToNextPage() {
    int index = visiblePointToLineIndexWithTwoArgs(scrollPageSize, true);
    if (index == this.scrollLineIndex) {
      index = this.getNextScrollLineIndex(index);
    }

    this.scrollLineIndex = index;
  }

  /// Scrolls to previous page.
  @override
  void scrollToPreviousPage() {
    int index =
        this.intPreviousPageLineIndex(scrollPageSize, this.scrollLineIndex);
    if (index == this.scrollLineIndex) {
      index = this.getPreviousScrollLineIndex(index);
    }

    this.scrollLineIndex = index;
  }

  int intPreviousPageLineIndex(double p, int index) {
    int result = index;
    while (p > 0) {
      result = index;
      index = this.getPreviousScrollLineIndex(index);
      if (index == -1) {
        return -1;
      }

      p -= this.getLineSize(index);
    }

    return result;
  }

  /// Scrolls to next line.
  @override
  void scrollToNextLine() {
    this.scrollLineIndex = this.getNextScrollLineIndex(this.scrollLineIndex);
  }

  /// Scrolls to previous line.
  @override
  void scrollToPreviousLine() {
    this.scrollLineIndex =
        this.getPreviousScrollLineIndex(this.scrollLineIndex);
  }

  /// Aligns the scroll line.
  @override
  void alignScrollLine() {}

  /// An array with 3 ranges indicating the first and last point for the given lines in each region.
  ///
  /// * first - _required_ - The index of the first line.
  /// * last - _required_ - The index of the last line.
  /// * allowEstimatesForOutOfViewLines - _required_ - if set to true allow estimates for out of view lines.
  ///
  /// Returns An array with 3 ranges indicating the first and last point for the given lines in each region.
  @override
  List<_DoubleSpan> rangeToRegionPoints(
      int first, int last, bool allowEstimatesForOutOfViewLines) {
    List<_DoubleSpan> result = List<_DoubleSpan>();
    for (int n = 0; n < 3; n++) {
      var region;
      if (n == 0) {
        region = _ScrollAxisRegion.header;
      } else if (n == 1) {
        region = _ScrollAxisRegion.body;
      } else if (n == 2) {
        region = _ScrollAxisRegion.footer;
      }
      result.insert(
          n,
          this.rangeToPoints(
              region, first, last, allowEstimatesForOutOfViewLines));
    }

    return result;
  }

  /// The first and last point for the given lines in a region.
  ///
  /// * region - _required_ - The scroll axis region.
  /// * first - _required_ - The index of the first line.
  /// * last - _required_ - The index of the last line.
  /// * allowEstimatesForOutOfViewLines - _required_ - if set to true allow estimates for out of view lines.
  ///
  /// Returns the first and last point for the given lines in a region.
  @override
  _DoubleSpan rangeToPoints(_ScrollAxisRegion region, int first, int last,
      bool allowEstimatesForOutOfViewLines) {
    bool firstVisible, lastVisible;
    _VisibleLineInfo firstLine, lastLine;

    List lineValues = this.getLinesAndVisibility(
        first, last, true, firstVisible, lastVisible, firstLine, lastLine);
    firstVisible = lineValues[0];
    lastVisible = lineValues[1];
    firstLine = lineValues[2];
    lastLine = lineValues[3];
    if (firstLine == null || lastLine == null) {
      return _DoubleSpan.empty();
    }

    if (allowEstimatesForOutOfViewLines) {
      switch (region) {
        case _ScrollAxisRegion.header:
          if (!firstLine.isHeader) {
            return _DoubleSpan.empty();
          }

          break;
        case _ScrollAxisRegion.footer:
          if (!lastLine.isFooter) {
            return _DoubleSpan.empty();
          }

          break;
        case _ScrollAxisRegion.body:
          if (firstLine.isFooter || lastLine.isHeader) {
            return _DoubleSpan.empty();
          }

          break;
      }

      return _DoubleSpan(firstLine.origin, lastLine.corner);
    } else {
      switch (region) {
        case _ScrollAxisRegion.header:
          {
            if (!firstLine.isHeader) {
              return _DoubleSpan.empty();
            }

            if (!lastVisible || !lastLine.isHeader) {
              double corner = firstLine.corner;
              for (int n = firstLine.lineIndex + 1; n <= last; n++) {
                corner += this.getLineSize(n);
              }

              return _DoubleSpan(firstLine.origin, corner);
            }

            return _DoubleSpan(firstLine.origin, lastLine.corner);
          }

        case _ScrollAxisRegion.footer:
          {
            if (!lastLine.isFooter) {
              return _DoubleSpan.empty();
            }

            if (!firstVisible || !firstLine.isFooter) {
              double origin = lastLine.origin;
              for (int n = lastLine.lineIndex - 1; n >= first; n--) {
                origin -= this.getLineSize(n);
              }

              return _DoubleSpan(origin, lastLine.corner);
            }

            return _DoubleSpan(firstLine.origin, lastLine.corner);
          }

        case _ScrollAxisRegion.body:
          {
            if (firstLine.isFooter || lastLine.isHeader) {
              return _DoubleSpan.empty();
            }

            double origin = firstLine.origin;
            if (!firstVisible || firstLine.region != _ScrollAxisRegion.body) {
              origin = this.headerExtent;
              for (int n = this.scrollLineIndex - 1; n >= first; n--) {
                origin -= this.getLineSize(n);
              }
            }

            double corner = lastLine.corner;
            if (!lastVisible || lastLine.region != _ScrollAxisRegion.body) {
              corner = lastBodyVisibleLine.corner;
              for (int n = lastBodyVisibleLine.lineIndex + 1; n <= last; n++) {
                corner += this.getLineSize(n);
              }
            }

            return _DoubleSpan(origin, corner);
          }
      }

      return _DoubleSpan(firstLine.origin, lastLine.corner);
    }
  }

  /// This method is called in response to a MouseWheel event.
  ///
  /// * delta - _required_ - The delta.
  @override
  void mouseWheel(int delta) {
    if (delta > 0) {
      this.scrollToPreviousLine();
    } else {
      this.scrollToNextLine();
    }
  }

  /// Scrolls the line into viewable area.
  ///
  /// * lineIndex - _required_ - The index of the line.
  /// * lineSize - _required_ - The size of the line.
  /// * isRightToLeft - _required_ - The boolean value used to calculate visible columns in right to left mode.
  @override
  void scrollInView(int lineIndex, double lineSize, bool isRightToLeft) {
    _VisibleLinesCollection lines = getVisibleLines();
    _VisibleLineInfo line = lines.getVisibleLineAtLineIndex(lineIndex);
    double delta = 0;

    if (line != null) {
      if (!line.isClipped || line.isFooter || line.isHeader) {
        return;
      }

      if (line.isClippedOrigin && !line.isClippedCorner) {
        delta = -1;
      } else if (!line.isClippedOrigin && line.isClippedCorner) {
        double y = line.size - line.clippedSize;
        double visibleScrollIndex = scrollBar.value;
        while (y > 0 && visibleScrollIndex < scrollBar.maximum) {
          delta++;
          visibleScrollIndex++;
          y -= this
              .getLineSize(this.scrollBarValueToLineIndex(visibleScrollIndex));
        }
      }
    } else {
      double visibleScrollIndex = this.lineIndexToScrollBarValue(lineIndex);

      if (visibleScrollIndex > scrollBar.value) {
        int scrollIndexLinex = this.intPreviousPageLineIndex(
            scrollPageSize - getLineSize(lineIndex), lineIndex);
        visibleScrollIndex = this.lineIndexToScrollBarValue(scrollIndexLinex);
      }

      delta = visibleScrollIndex - scrollBar.value;
    }

    if (delta != 0) {
      scrollBar.value += delta;
    }

    super.scrollInView(lineIndex, lineSize, isRightToLeft);
  }

  /// Called when lines were inserted in ScrollLinesHost.
  ///
  /// * insertAt - _required_ - Index of the first inserted line.
  /// * count - _required_ - The count.
  @override
  void onLinesInserted(int insertAt, int count) {
    int to = insertAt + count - 1;
    int repeatSizeCount;
    for (int index = insertAt; index <= to; index++) {
      bool hide = scrollLinesHost.getHidden(index, repeatSizeCount)[0];
      repeatSizeCount = scrollLinesHost.getHidden(index, repeatSizeCount)[1];
      var value = hide == true ? 0 : 1.0;
      int rangeTo = getRangeToHelper(index, to, repeatSizeCount);
      this.distances.setRange(index, rangeTo, value);
      index = rangeTo;
    }
  }
}
