part of datagrid;

/// ScrollAxisBase is an abstract base class and implements scrolling
/// logic for both horizontal and vertical scrolling in a `Syncfusion.GridCommon.ScrollAxis.PixelScrollAxis.Distances`.
/// Logical units in the ScrollAxisBase are called "Lines". With the
/// `Syncfusion.GridCommon.ScrollAxis.PixelScrollAxis.distances` a line represents rows in a grid
/// and with `Syncfusion.GridCommon.ScrollAxis.PixelScrollAxis.Distances` a line represents columns in a grid.
///
/// ScrollAxisBase has support for frozen header and footer lines, maintaining a
/// scroll position and updating and listening to scrollbars. It also maintains
/// a collection of `VisibleLineInfo` items for all the lines that are
/// visible in the viewing area. ScrollAxisBase wires itself with a
/// `ScrollLinesHost` and reacts to changes in line count,
/// line sizes, hidden state and default line size.

typedef _ChangedCallback = void Function(_ScrollChangedArgs scrollChangedArgs);

abstract class _ScrollAxisBase {
  double _renderSize = 0.0;
  _ScrollBarBase _scrollBar;
  _LineSizeHostBase _scrollLinesHost;
  bool _layoutDirty = false;
  int _lineResizeIndex = -1;
  double _lineResizeSize = 0;
  _VisibleLinesCollection _visibleLines = _VisibleLinesCollection();
  double _lastScrollValue = -1;
  _DoubleSpan _clip = _DoubleSpan.empty();
  bool _ignoreScrollBarPropertyChange = false;
  bool _inGetVisibleLines = false;
  bool _allBodyLinesShown = false;
  int _lastBodyLineIndex = -1;

  /// Gets the name.
  ///
  /// Returns the name.
  String get name => _name;
  String _name;

  /// Sets the name.
  set name(String value) {
    if (value == _name) return;
    _name = value;
  }

  /// Occurs when a scroll axis changed.
  _ChangedCallback onChangedCallback;

  /// Initializes a new instance of the ScrollAxisBase class.
  ///
  /// * scrollBar The scroll bar state.
  /// * scrollLinesHost The scroll lines host.
  _ScrollAxisBase(_ScrollBarBase scrollBar, _LineSizeHostBase scrollLinesHost) {
    if (scrollBar == null) {
      throw Exception();
    }
    this.scrollBar = scrollBar;
    this._scrollLinesHost = scrollLinesHost;
    this.wireScrollLinesHost();
  }

  /// Gets a value indicating whether this axis supports pixel scrolling.
  ///
  /// @value
  /// `true` if this instance supports pixel scrolling, otherwise `false`.
  bool get isPixelScroll => _isPixelScroll;
  bool _isPixelScroll;

  // Support for sharing axis with a parent axis
  /// Gets the index of the first line in a parent axis.
  ///
  /// This is used for shared
  /// or nested scroll axis (e.g. a nested grid with shared axis in a covered cell).
  ///
  /// Returns the index of the first line.
  int get startLineIndex => _startLineIndex;
  int _startLineIndex;

  /// Sets the index of the first line in a parent axis.
  set startLineIndex(int value) {
    if (value == _startLineIndex) return;
    _startLineIndex = value;
  }

  /// Gets the scroll bar state.
  ///
  /// Returns the scroll bar state.
  _ScrollBarBase get scrollBar => _scrollBar;
  set scrollBar(_ScrollBarBase value) {
    if (value == _scrollBar) return;
    _scrollBar = value;
  }

  /// Updates the scroll bar.
  ///
  /// * ignorePropertyChange A boolean value indicating whether to ignore the
  /// property change.
  void updateScrollBar(bool ignorePropertyChange) {
    bool b = this._ignoreScrollBarPropertyChange;
    this._ignoreScrollBarPropertyChange |= ignorePropertyChange;
    try {
      this.updateScrollbar();
    } finally {
      this._ignoreScrollBarPropertyChange = b;
    }
  }

  void updateScrollbar();

  /// Gets the line count.
  ///
  /// Returns the line count.
  int get lineCount => _lineCount;
  int _lineCount;

  /// Sets the line count.
  set lineCount(int value) {
    if (value == _lineCount) return;
    _lineCount = value;
  }

  /// Gets the default size of lines.
  ///
  /// Returns the default size of lines.
  double get defaultLineSize => _defaultLineSize;
  double _defaultLineSize;

  /// Sets the default size of lines.
  set defaultLineSize(double value) {
    if (value == _defaultLineSize) return;
    _defaultLineSize = value;
  }

  /// Sets the hidden state of the lines.
  ///
  /// * from The start index of the line.
  /// * to The end index of the line.
  /// * hide A boolean value indicating whether to hide the lines. if set to true - [hide].
  void setLineHiddenState(int from, int to, bool hide);

  /// Sets the size of the lines for the given range of lines. Will do nothing for a [LineScrollAxis].
  ///
  /// * from The start index of the line.
  /// * to The end index of the line.
  /// * size The line size.
  void setLineSize(int from, int to, double size);

  /// Gets the size from `ScrollLinesHost` or if the line is being resized then get temporary value
  /// previously set with `SetLineResize`
  ///
  /// * index The index.
  /// * repeatSizeCount The number of subsequent values with same size.
  /// Returns the size from `ScrollLinesHost` or if the line is being resized then
  /// get temporary value previously set with `SetLineResize`
  List getLineSizeWithTwoArgs(int index, int repeatSizeCount) {
    repeatSizeCount = 1;
    if (index == this._lineResizeIndex) {
      return [this._lineResizeSize, repeatSizeCount];
    }

    if (this.scrollLinesHost == null) {
      repeatSizeCount = _MathHelper.maxvalue;
      return [this.defaultLineSize, repeatSizeCount];
    }

    return [
      this.getScrollLinesHostSize(index, repeatSizeCount)[0],
      this.getScrollLinesHostSize(index, repeatSizeCount)[1]
    ];
  }

  /// Gets the size of the line.
  ///
  /// * index The index of the line.
  /// Returns the size of the line.
  double getLineSize(int index) {
    int repeatSizeCount;
    return this.getLineSizeWithTwoArgs(index, repeatSizeCount)[0];
  }

  /// Gets the maximum range
  ///
  /// * n start index
  /// * to end index
  /// * repeatSizeCount repeat count value
  /// Returns the minimum index value
  int getRangeToHelper(int n, int to, int repeatSizeCount) {
    if (repeatSizeCount == _MathHelper.maxvalue) {
      return to;
    }

    return min(to, n + repeatSizeCount - 1);
  }

  /// Set temporary value for a line size during a resize operation without committing
  /// value to ScrollLinesHost.
  ///
  /// * index The index of the line.
  /// * size The size of the line.

  void setLineResize(int index, double size) {
    this._lineResizeIndex = index;
    this._lineResizeSize = size;
    this.setLineSize(index, index, size);
    this.markDirty();
    raiseChanged(_ScrollChangedAction.lineResized);
  }

  /// Resets temporary value for line size after a resize operation.
  void resetLineResize() {
    int repeatSizeCount = 0;
    if (this._lineResizeIndex >= 0 && this._scrollLinesHost != null) {
      this.setLineSize(
          this._lineResizeIndex,
          this._lineResizeIndex,
          this.getScrollLinesHostSize(
              this._lineResizeIndex, repeatSizeCount)[0]);
    }

    this._lineResizeIndex = -1;
    this._lineResizeSize = 0;
    this.markDirty();
    raiseChanged(_ScrollChangedAction.lineResized);
  }

  void rangeChangedCallback(_RangeChangedArgs rangeChangedArgs) {
    for (int n = rangeChangedArgs.from; n <= rangeChangedArgs.to; n++) {
      int repeatSizeCount = 0;
      double size = getScrollLinesHostSize(n, repeatSizeCount)[0];
      repeatSizeCount = getScrollLinesHostSize(n, repeatSizeCount)[1];
      int rangeTo =
          this.getRangeToHelper(n, rangeChangedArgs.to, repeatSizeCount);
      this.setLineSize(n, rangeTo, size);
      n = rangeTo;
    }
    // Also check whether I need to re-hide any of the rows.
    this.hiddenRangeChangedCallback(_HiddenRangeChangedArgs.fromArgs(
        rangeChangedArgs.from, rangeChangedArgs.to, false));
    this.markDirty();
    raiseChanged(_ScrollChangedAction.lineResized);
  }

  void linesRemovedCallback(_LinesRemovedArgs linesRemovedArgs) {
    onLinesRemoved(linesRemovedArgs.removeAt, linesRemovedArgs.count);
    raiseChanged(_ScrollChangedAction.linesRemoved);
  }

  void linesInsertedCallback(_LinesInsertedArgs linesInsertedArgs) {
    onLinesInserted(linesInsertedArgs.insertAt, linesInsertedArgs.count);
    raiseChanged(_ScrollChangedAction.linesInserted);
  }

  void defaultLineSizeChangedCallback(
      _DefaultLineSizeChangedArgs defaultLineSizeChangedArgs) {
    this.defaultLineSize = this.scrollLinesHost.getDefaultLineSize();
    this.markDirty();
    this.updateScrollBar(true);
    raiseChanged(_ScrollChangedAction.defaultLineSizeChanged);
  }

  void hiddenRangeChangedCallback(
      _HiddenRangeChangedArgs hiddenRangeChangedArgs) {
    for (int n = hiddenRangeChangedArgs.from;
        n <= hiddenRangeChangedArgs.to;
        n++) {
      int repeatSizeCount = 0;
      bool hide = this.scrollLinesHost.getHidden(n, repeatSizeCount)[0];
      repeatSizeCount = this.scrollLinesHost.getHidden(n, repeatSizeCount)[1];
      int rangeTo =
          this.getRangeToHelper(n, hiddenRangeChangedArgs.to, repeatSizeCount);
      this.setLineHiddenState(n, rangeTo, hide);
      n = rangeTo;
    }

    this.markDirty();
    this.updateScrollBar(true);
    raiseChanged(_ScrollChangedAction.hiddenLineChanged);
  }

  void scrollChangedCallback(_ScrollChangedArgs scrollChangedArgs) {}
  void footerLineChangedCallback() {
    this.setFooterLineCount(this.scrollLinesHost.getFooterLineCount());
    this.markDirty();
    raiseChanged(_ScrollChangedAction.footerLineCountChanged);
  }

  void headerLineChangedCallback() {
    this.setFooterLineCount(this.scrollLinesHost.getFooterLineCount());
    this.markDirty();
    raiseChanged(_ScrollChangedAction.footerLineCountChanged);
  }

  void lineCountChangedCallback() {
    this.lineCount = this.scrollLinesHost.getLineCount();
    this.markDirty();
    this.updateScrollBar(true);
    raiseChanged(_ScrollChangedAction.lineCountChanged);
  }

  /// Gets the size from `ScrollLinesHost` or if the line is being resized then get temporary value
  /// previously set with `SetLineResize`. If size is negative then `DefaultLineSize` is returned.
  ///
  /// * index The index of the line.
  /// * repeatSizeCount The number of subsequent values with same size.
  ///
  /// The size from `ScrollLinesHost` or if the line is being resized then get temporary value
  /// previously set with `SetLineResize`. If size is negative then `DefaultLineSize` is returned.

  List getScrollLinesHostSize(int index, int repeatSizeCount) {
    double size = this.scrollLinesHost.getSize(index, repeatSizeCount)[0];
    repeatSizeCount = this.scrollLinesHost.getSize(index, repeatSizeCount)[1];
    if (size < 0) {
      size = this.defaultLineSize;
    }

    return [size, repeatSizeCount];
  }

  /// Gets the scroll lines host.
  ///
  /// Returns the scroll lines host.
  _LineSizeHostBase get scrollLinesHost => this._scrollLinesHost;

  void wireScrollLinesHost() {
    if (this.scrollLinesHost == null) {
      return;
    }
    this.scrollLinesHost.onDefaultLineSizeChanged =
        defaultLineSizeChangedCallback;
    this.scrollLinesHost.onLineHiddenChanged = hiddenRangeChangedCallback;
    this.scrollLinesHost.onLineSizeChanged = rangeChangedCallback;
    this.scrollLinesHost.onLinesInserted = linesInsertedCallback;
    this.scrollLinesHost.onLinesRemoved = linesRemovedCallback;
    this.scrollLinesHost.onFooterLineCountChanged = headerLineChangedCallback;
    this.scrollLinesHost.onFooterLineCountChanged = footerLineChangedCallback;
    this.scrollLinesHost.onLineCountChanged = lineCountChangedCallback;
  }

  void unwireScrollLinesHost() {
    if (this.scrollLinesHost == null) {
      return;
    }
    this.scrollLinesHost.onDefaultLineSizeChanged =
        defaultLineSizeChangedCallback;
    this.scrollLinesHost.onLineHiddenChanged = hiddenRangeChangedCallback;
    this.scrollLinesHost.onLineSizeChanged = rangeChangedCallback;
    this.scrollLinesHost.onLinesInserted = linesInsertedCallback;
    this.scrollLinesHost.onLinesRemoved = linesRemovedCallback;
    this.scrollLinesHost.onFooterLineCountChanged = headerLineChangedCallback;
    this.scrollLinesHost.onFooterLineCountChanged = footerLineChangedCallback;
    this.scrollLinesHost.onLineCountChanged = lineCountChangedCallback;
  }

  /// Called when lines were removed in ScrollLinesHost.
  ///
  /// * removeAt Index of the first removed line.
  /// * count The count.
  void onLinesRemoved(int removeAt, int count) {}

  /// Called when lines were inserted in ScrollLinesHost.
  ///
  /// * insertAt Index of the first inserted line.
  /// * count The count.
  void onLinesInserted(int insertAt, int count) {}

  /// Sets the size (either height or width) of the parent control.
  ///
  /// Returns the size of the parent control.
  double get renderSize => this._renderSize;

  /// Sets the size (either height or width) of the parent control.
  set renderSize(double value) {
    if (this._renderSize != value) {
      this._renderSize = value;
      this.markDirty();
      this.updateScrollBar(true);
    }
  }

  /// Gets the size (either height or width) of the parent control excluding the
  /// area occupied by Header and Footer. This size is used for scrolling down
  /// or up one page.
  ///
  /// Returns the size of the parent control.
  double get scrollPageSize =>
      this.renderSize - this.headerExtent - this.footerExtent;

  /// Gets the clipping region. Depending on the orientation of
  /// the axis, this is either the left and right or top and bottom
  /// values of the clipping rectangle in the parent control.
  ///
  /// Returns the clip value.
  _DoubleSpan get clip => this._clip;

  /// Sets the clipping region. Depending on the orientation of
  /// the axis, this is either the left and right or top and bottom
  /// values of the clipping rectangle in the parent control.
  set clip(_DoubleSpan value) {
    this.clip = value;
  }

  /// Gets the view size of the (either height or width) of the parent control. Normally
  /// the ViewSize is the same as `RenderSize`. Only if the parent control
  /// has more space then needed to display all lines, the ViewSize will be less. In
  /// such case the ViewSize is the total height for all lines.
  ///
  /// Returns the view size of the (either height or width) of the parent control.
  double get viewSize => _viewSize;
  double _viewSize = 0.0;

  /// Gets the header extent. This is total height (or width) of the header lines.
  ///
  /// Returns the header extent.
  double get headerExtent => _headerExtent;
  double _headerExtent = 0.0;

  /// Gets the header line count.
  ///
  /// Returns the header line count.
  int get headerLineCount {
    if (this.scrollLinesHost == null) {
      return 0;
    }

    return this.scrollLinesHost.getHeaderLineCount();
  }

  /// Sets the header line count.
  ///
  /// * value The value.
  void setHeaderLineCount(int value);

  /// Gets the footer extent. This is total height (or width) of the footer lines.
  ///
  /// Returns the footer extent.
  double get footerExtent => _footerExtent;
  double _footerExtent = 0.0;

  /// Gets the footer line count.
  ///
  /// Returns the footer line count.
  int get footerLineCount {
    if (this.scrollLinesHost == null) {
      return 0;
    }

    return this.scrollLinesHost.getFooterLineCount();
  }

  /// Sets the footer line count.
  ///
  /// * value The value.
  void setFooterLineCount(int value);

  /// Scroll = First Visible Body Line
  ///
  /// Gets the index of the first visible Line in the body region.
  ///
  /// Returns the index of the scroll line.
  int get scrollLineIndex => _scrollLineIndex;
  int _scrollLineIndex = 0;

  /// Sets the index of the first visible Line in the body region.
  set scrollLineIndex(int value) {
    if (value == _scrollLineIndex) return;
    _scrollLineIndex = value;
  }

  /// Gets the index of the scroll line in RTL Mode
  ///
  /// * scrollLineIndex Index of the scroll line.
  /// * scrollLineOffset The scroll line offset.
  /// * isRightToLeft The boolean value used to calculate visible columns in right to left mode.
  List getScrollLineIndex(int scrollLineIndex, double scrollLineOffset,
      [bool isRightToLeft = false]);

  /// Sets the index of the scroll line.
  ///
  /// * scrollLineIndex The index of the scroll line.
  /// * scrollLineOffset The scroll line offset.
  void setScrollLineIndex(int scrollLineIndex, double scrollLineOffset);

  /// Gets the index of the previous scroll line.
  ///
  /// * lineIndex The current index of the line.
  /// Returns the index of the previous scroll line.
  int getPreviousScrollLineIndex(int lineIndex);

  /// Gets the index of the next scroll line.
  ///
  /// * lineIndex The current index of the line.
  /// Returns the index of the next scroll line.
  int getNextScrollLineIndex(int lineIndex);

  /// Scrolls to next page.
  void scrollToNextPage();

  /// Scrolls to previous page.
  void scrollToPreviousPage();

  /// Scrolls to next line.
  void scrollToNextLine();

  /// Scrolls to previous line.
  void scrollToPreviousLine();

  /// Aligns the scroll line.
  void alignScrollLine();

  /// Scrolls the line into viewable area.
  /// * lineIndex Index of the line.
  /// * lineSize Size of the line.
  /// * isRightToLeft The boolean value indicates the right to left mode.
  void scrollInView(int lineIndex, double lineSize, bool isRightToLeft) {}

  /// Scrolls the line into viewable area.
  ///
  /// * lineIndex The index of the line.
  /// * isRightToLeft The boolean value used to calculate visible columns in right to left mode.s
  void scrollInViewwithTwoArgs(int lineIndex, bool isRightToLeft) {
    scrollInView(lineIndex, getLineSize(lineIndex), isRightToLeft);
  }

  /// This method is called in response to a MouseWheel event.
  ///
  /// * delta The delta.
  void mouseWheel(int delta);

  /// Makes the layout dirty.
  ///
  void markDirty() {
    this._layoutDirty = true;
  }

  //code to handle issue in DT 77714
  static double ePSILON =
      2.2204460492503131e-016; /* smallest such that 1.0+EPSILON != 1.0 */

  static bool strictlyLessThan(double d1, double d2) {
    return !(d1 > d2 || areClose(d1, d2));
  }

  static bool areClose(double d1, double d2) {
    if (d1 == d2) {
      return true;
    }

    double eps = (d1.abs() + d2.abs() + 10.0) * ePSILON;
    return (d1 - d2).abs() < eps;
  }

  /// Freezes the visible lines.
  void freezeVisibleLines() {
    this._inGetVisibleLines = true;
  }

  /// Unfreezes the visible lines.
  void unfreezeVisibleLines() {
    this._inGetVisibleLines = false;
  }

  /// Gets the visible lines collection.
  ///
  /// * isRightToLeft The boolean value used to calculate visible columns in right to left mode.
  /// Returns the visible lines collection.
  _VisibleLinesCollection getVisibleLines([bool isRightToLeft = false]) {
    if (!isRightToLeft) {
      if (this._inGetVisibleLines) {
        return this._visibleLines;
      }

      this._inGetVisibleLines = true;
      try {
        if (this._layoutDirty) {
          this.setHeaderLineCount(this.headerLineCount);
          this.setFooterLineCount(this.footerLineCount);

          this._lastScrollValue = -1;
          this._layoutDirty = false;
          updateScrollBar(true);
        }
        if (this._visibleLines.length == 0 ||
            this._lastScrollValue != this.scrollBar.value) {
          this._visibleLines.clear();

          int visibleIndex = 0;
          int scrollLineIndex = 0;
          double scrollOffset = 0.0;
          int headerLineCount = this.headerLineCount;
          List scrollLineValues =
              this.getScrollLineIndex(scrollLineIndex, scrollOffset);
          scrollLineIndex = scrollLineValues[0];
          scrollOffset = scrollLineValues[1];
          int firstFooterLine = this.lineCount - this.footerLineCount;
          double footerStartPoint = this.renderSize - this.footerExtent;
          int index;

          // Header
          double point = 0;
          int lastHeaderLineIndex = -1;
          for (index = 0;
              index != -1 &&
                  strictlyLessThan(point, this.headerExtent) &&
                  index < firstFooterLine &&
                  index < headerLineCount;
              index = this.getNextScrollLineIndex(index)) {
            double size = this.getLineSize(index);
            this._visibleLines.add(_VisibleLineInfo(
                visibleIndex++, index, size, point, 0, true, false));
            point += size;
            lastHeaderLineIndex = index;
          }

          this._visibleLines.firstBodyVisibleIndex = this._visibleLines.length;

          _VisibleLineInfo lastScrollableLine;
          // Body
          point = this.headerExtent;
          int firstBodyLineIndex =
              max(scrollLineIndex, lastHeaderLineIndex + 1);
          for (index = firstBodyLineIndex;
              index != -1 &&
                  strictlyLessThan(point, footerStartPoint) &&
                  index < firstFooterLine;
              index = this.getNextScrollLineIndex(index)) {
            double size = this.getLineSize(index);
            this._visibleLines.add(lastScrollableLine = _VisibleLineInfo(
                visibleIndex++,
                index,
                size,
                point,
                scrollOffset,
                false,
                false));
            point += size - scrollOffset;
            scrollOffset =
                0; // reset scrollOffset after first line. Subsequent lines will start at given point.
          }

          if (lastScrollableLine == null) {
            this._allBodyLinesShown = true;
            this._lastBodyLineIndex = -1;
          } else {
            this._allBodyLinesShown = index >= firstFooterLine;
            this._lastBodyLineIndex = lastScrollableLine.lineIndex;
          }

          this._visibleLines.firstFooterVisibleIndex =
              this._visibleLines.length;

          // Footer
          point = max(this.headerExtent, this.viewSize - this.footerExtent);
          for (index = firstFooterLine;
              index != -1 &&
                  strictlyLessThan(point, this.renderSize) &&
                  index < this.lineCount;
              index = this.getNextScrollLineIndex(index)) {
            if (lastScrollableLine != null) {
              lastScrollableLine.clippedCornerExtent =
                  lastScrollableLine.corner - point;
              lastScrollableLine = null;
            }

            double size = this.getLineSize(index);
            this._visibleLines.add(_VisibleLineInfo(
                visibleIndex++, index, size, point, 0, false, true));
            point += size;
          }

          if (lastScrollableLine != null) {
            lastScrollableLine.clippedCornerExtent =
                lastScrollableLine.corner - point;
            lastScrollableLine = null;
          }

          this._lastScrollValue = this.scrollBar.value;

          if (this._visibleLines.length > 0) {
            this._visibleLines[this._visibleLines.length - 1].isLastLine = true;
          }
          try {
            // throws exception when a line is duplicate
            this._visibleLines.getVisibleLineAtLineIndex(0);
          } catch (Exception) {
            if (!this.indebug) {
              this.indebug = true;
              this._visibleLines.clear();
              try {
                this.getVisibleLines();
              } finally {
                this.indebug = false;
              }
            }
          }
        }
      } finally {
        this._inGetVisibleLines = false;
      }
      return this._visibleLines;
    } else {
      if (_inGetVisibleLines) return _visibleLines;
      _inGetVisibleLines = true;
      try {
        if (_layoutDirty) {
          setHeaderLineCount(headerLineCount);
          setFooterLineCount(footerLineCount);
          _lastScrollValue = -1;
          _layoutDirty = false;
          updateScrollBar(true);
        }

        if (_visibleLines.length == 0 || _lastScrollValue != scrollBar.value) {
          _visibleLines.clear();
          int visibleIndex = 0;
          int scrollLineIndex = 0;
          double scrollOffset = 0.0;
          int _headerLineCount = headerLineCount;
          List scrollLineValues =
              this.getScrollLineIndex(scrollLineIndex, scrollOffset, true);
          scrollLineIndex = scrollLineValues[0];
          scrollOffset = scrollLineValues[1];
          int firstFooterLine = lineCount - footerLineCount;
          int index;
          double footerStartPoint = renderSize + clip.start;

          // Header
          double point = 0;
          int lastHeaderLineIndex = -1;
          for (index = 0;
              index != -1 &&
                  strictlyLessThan(point, headerExtent) &&
                  index < firstFooterLine &&
                  index < _headerLineCount;
              index = getNextScrollLineIndex(index)) {
            double size = this.getLineSize(index);
            _visibleLines.add(_VisibleLineInfo(visibleIndex++, index, size,
                footerStartPoint - size - point, 0, true, false));
            footerStartPoint -= size;
            lastHeaderLineIndex = index;
          }

          _visibleLines.firstBodyVisibleIndex = _visibleLines.length;
          _VisibleLineInfo lastScrollableLine;

          // Body
          point = headerExtent;
          int firstBodyLineIndex =
              max(scrollLineIndex, lastHeaderLineIndex + 1);

          point = footerStartPoint - clip.start;

          if (clip.start > 0 && scrollOffset < clip.start) {
            scrollOffset =
                getLineSize(scrollLineIndex - 1) - (clip.start - scrollOffset);
            scrollLineIndex -= 1;
            point += clip.start;
          }

          firstBodyLineIndex = max(scrollLineIndex, lastHeaderLineIndex + 1);
          for (index = firstBodyLineIndex;
              index != -1 &&
                  strictlyLessThan(
                      point - 1, renderSize + clip.start - headerExtent) &&
                  (point > footerExtent) &&
                  (index < firstFooterLine && index >= headerLineCount);
              index = getNextScrollLineIndex(index)) {
            double size = this.getLineSize(index);
            _visibleLines.add(lastScrollableLine = _VisibleLineInfo(
                visibleIndex++,
                index,
                size,
                point - size + scrollOffset,
                scrollOffset,
                false,
                false));
            point -= size - scrollOffset;
            scrollOffset = 0;
          }

          if (lastScrollableLine == null) {
            _allBodyLinesShown = true;
            _lastBodyLineIndex = -1;
          } else {
            _allBodyLinesShown = index >= firstFooterLine;
            _lastBodyLineIndex = lastScrollableLine.lineIndex;
          }

          _visibleLines.firstFooterVisibleIndex = _visibleLines.length;

          if (footerLineCount > 0) {
            if (this.renderSize < scrollBar.maximum + footerExtent) {
              point = min(renderSize + clip.start - headerExtent,
                  clip.start + footerExtent);
              for (index = firstFooterLine;
                  index != -1 &&
                      strictlyLessThan(
                          point - 1,
                          (this.renderSize < scrollBar.maximum
                              ? this.clip.start + footerExtent
                              : renderSize - headerExtent)) &&
                      index < lineCount;
                  index = getNextScrollLineIndex(index)) {
                if (lastScrollableLine != null) {
                  lastScrollableLine.clippedCornerExtent =
                      lastScrollableLine.corner - point;
                  lastScrollableLine = null;
                }

                double size = this.getLineSize(index);
                _visibleLines.add(_VisibleLineInfo(
                    visibleIndex++, index, size, point - size, 0, false, true));
                point -= size;
              }
            } else {
              for (index = firstFooterLine;
                  index != -1 &&
                      strictlyLessThan(
                          point - 1,
                          (this.renderSize < scrollBar.maximum
                              ? this.clip.start + footerExtent
                              : renderSize - headerExtent)) &&
                      index < lineCount;
                  index = getNextScrollLineIndex(index)) {
                if (lastScrollableLine != null) {
                  lastScrollableLine.clippedCornerExtent =
                      lastScrollableLine.corner - point;
                  lastScrollableLine = null;
                }

                double size = this.getLineSize(index);
                _visibleLines.add(_VisibleLineInfo(
                    visibleIndex++, index, size, point - size, 0, false, true));
                point -= size;
              }
            }
          }

          if (lastScrollableLine != null) {
            lastScrollableLine.clippedCornerExtent =
                lastScrollableLine.corner - point;
            lastScrollableLine = null;
          }

          _lastScrollValue = scrollBar.value;
          if (_visibleLines.length > 0)
            _visibleLines[_visibleLines.length - 1].isLastLine = true;

          try {
            // throws exception when a line is duplicate
            _visibleLines.getVisibleLineAtLineIndex(0);
          } catch (Exception) {
            if (!indebug) {
              indebug = true;
              _visibleLines.clear();
              try {
                getVisibleLines(true);
              } finally {
                indebug = false;
              }
            }
          }
        }
      } finally {
        _inGetVisibleLines = false;
      }
      return _visibleLines;
    }
  }

  bool indebug = false;

  /// Resets the visible lines collection.
  void resetVisibleLines() {
    this._visibleLines.clear();
  }

  /// Gets the visible line index for a point in the display.
  ///
  /// * point The point.
  /// * allowOutsideLines Set this true if point can be below corner of last line.
  /// Returns the visible line index for a point in the display.
  int visiblePointToLineIndexWithTwoArgs(double point, bool allowOutsideLines) {
    if (allowOutsideLines) {
      point = max(point, 0);
    }

    _VisibleLineInfo line = this.getVisibleLines().getVisibleLineAtPoint(point);
    if (line != null && (allowOutsideLines || point <= line.corner)) {
      return line.lineIndex;
    }

    return -1;
  }

  /// Gets the visible line index for a point in the display.
  ///
  /// * point The point in the display for which the line index is to be obtained.
  /// Returns the visible line index for a point in the display.
  int visiblePointToLineIndex(double point) {
    return this.visiblePointToLineIndexWithTwoArgs(point, true);
  }

  /// Gets the visible line for a point in the display.
  ///
  /// * point The point in the display for which the visible line is to be obtained.
  /// * allowOutSideLines boolean value
  /// * isRightToLeft The boolean value used to calculate visible columns in right to left mode.
  /// Returns the visible line for a point in the display.
  _VisibleLineInfo getVisibleLineAtPoint(double point,
      [bool allowOutSideLines = false, bool isRightToLeft = false]) {
    if (!isRightToLeft) {
      if (allowOutSideLines) {
        point = max(point, 0);
      }
      var lineInfo = this.getVisibleLines().getVisibleLineAtPoint(point);
      if (lineInfo != null && (allowOutSideLines || point <= lineInfo.corner)) {
        return lineInfo;
      }
    } else {
      if (allowOutSideLines) point = max(point, 0);

      _VisibleLinesCollection collection = getVisibleLines(true);
      collection.reversed;
      var lineInfo = collection.getVisibleLineAtPoint(point);
      collection.reversed;
      if (lineInfo != null && (allowOutSideLines || point <= lineInfo.corner))
        return lineInfo;
    }

    return null;
  }

  /// Gets the visible line that displays the line with the given absolute line index.
  ///
  /// * lineIndex Index of the line.
  /// The visible line that displays the line with the given absolute line index.

  _VisibleLineInfo getVisibleLineAtLineIndex(int lineIndex) {
    return this.getVisibleLines().getVisibleLineAtLineIndex(lineIndex);
  }

  /// Gets the visible line that displays the line with the given absolute line index. If the
  /// line is outside the view and you specify `allowCreateEmptyLineIfNotVisible` then
  /// the method will create an empty line and initializes its line index and line size.
  ///
  /// * lineIndex Index of the line.
  /// * allowCreateEmptyLineIfNotVisible if set to true and if the
  /// line is outside the view then
  /// the method will create an empty line and initializes its LineIndex and LineSize.
  /// Returns the visible line that displays the line with the given absolute line index.
  _VisibleLineInfo getVisibleLineAtLineIndexWithTwoArgs(
      int lineIndex, bool allowCreateEmptyLineIfNotVisible) {
    _VisibleLineInfo line = this.getVisibleLineAtLineIndex(lineIndex);
    if (line == null && allowCreateEmptyLineIfNotVisible) {
      double size = this.getLineSize(lineIndex);
      line = _VisibleLineInfo(_MathHelper.maxvalue, lineIndex, size,
          this.renderSize + 1, size, false, false);
    }

    return line;
  }

  /// Gets a boolean value indicating whether the line with the given absolute line index is
  /// visible.
  ///
  /// * lineIndex The index of the line.
  /// A boolean value indicating whether the line with the given absolute line index is
  /// visible.

  bool isLineVisible(int lineIndex) {
    return this.getVisibleLines().getVisibleLineAtLineIndex(lineIndex) !=
            null ||
        ((lineIndex > this._lastBodyLineIndex) && this._allBodyLinesShown);
  }

  /// Gets a boolean value indicating whether any of the lines with the given absolute line index
  /// are visible.
  ///
  /// * lineIndex1 The line index 1.
  /// * lineIndex2 The line index 2.
  /// A boolean value indicating whether any of the lines with the given absolute line index
  /// are visible.
  bool anyVisibleLines(int lineIndex1, int lineIndex2) {
    return this._visibleLines.anyVisibleLines(lineIndex1, lineIndex2);
  }

  /// Gets the origin and corner points of body region.
  ///
  /// * origin The origin.
  /// * corner The corner.
  void getOriginAndCornerOfBodyRegion(double origin, double corner) {
    int scrollLineIndex = 0;
    double scrollOffset = 0.0;
    scrollLineIndex = this.getScrollLineIndex(scrollLineIndex, scrollOffset)[0];
    scrollOffset = this.getScrollLineIndex(scrollLineIndex, scrollOffset)[1];
    double arrangeSize = this.renderSize;
    double adjustedFooterExtent =
        this.adjustFooterExtentToAvoidGap(this.footerExtent, arrangeSize);
    origin = this.headerExtent - scrollOffset;
    corner = arrangeSize - adjustedFooterExtent;
  }

  /// Adjusts the footer extent to avoid gap between last visible line of body region
  /// and first line of footer in case the view is larger than the height/width of all
  /// lines.
  ///
  /// * footerSize Size of the footer.
  /// * arrangeSize Size of the arrange.
  /// Returns the
  double adjustFooterExtentToAvoidGap(double footerSize, double arrangeSize) {
    // Adjust start of footer to avoid gap after last row.
    if (this.viewSize < arrangeSize) {
      footerSize += arrangeSize - this.viewSize;
    }

    if (footerSize + this.headerExtent > arrangeSize) {
      footerSize = max(0, arrangeSize - this.headerExtent);
    }

    return footerSize;
  }

  /// Gets the view corner which is the point after the last visible line
  /// of the body region.
  ///
  /// Returns the view corner which is the point after the last visible line of the body region.
  double get viewCorner {
    int scrollLineIndex = 0;
    double scrollOffset = 0.0;
    scrollLineIndex = this.getScrollLineIndex(scrollLineIndex, scrollOffset)[0];
    scrollOffset = this.getScrollLineIndex(scrollLineIndex, scrollOffset)[1];
    double arrangeSize = this.renderSize;
    double adjustedFooterExtent =
        this.adjustFooterExtentToAvoidGap(this.footerExtent, arrangeSize);

    return arrangeSize - adjustedFooterExtent;
  }

  /// Return indexes for VisibleLinesCollection for area identified by section.
  ///
  /// * section The integer value indicating the section : 0 - Header, 1 - Body, 2 - Footer
  /// * start The start index.
  /// * end The end index.
  List getVisibleSection(int section, int start, int end) {
    _VisibleLinesCollection visibleLines = this.getVisibleLines();
    switch (section) {
      case 0:
        start = 0;
        end = visibleLines.firstBodyVisibleIndex - 1;
        break;
      case 1:
        start = visibleLines.firstBodyVisibleIndex;
        end = visibleLines.firstFooterVisibleIndex - 1;
        break;
      case 2:
        start = visibleLines.firstFooterVisibleIndex;
        end = visibleLines.length - 1;
        break;
      default:
        start = end = -1;
        break;
    }
    return [start, end];
  }

  /// Returns the first and last VisibleLine.LineIndex for area identified by section.
  ///
  /// * section The integer value indicating the section : 0 - Header, 1 - Body, 2 - Footer
  /// Returns the first and last VisibleLine.LineIndex for area identified by section.
  _Int32Span getVisibleLinesRange(int section) {
    _VisibleLinesCollection visibleLines = this.getVisibleLines();
    int start = 0, end = 0;
    start = this.getVisibleSection(section, start, end)[0];
    start = this.getVisibleSection(section, start, end)[1];
    return _Int32Span(
        visibleLines[start].lineIndex, visibleLines[end].lineIndex);
  }

  /// Return indexes for VisibleLinesCollection for area identified by section.
  ///
  /// * section The integer value indicating the section : 0 - Header, 1 - Body, 2 - Footer
  /// * start The start index.
  /// * end The end index.
  void getVisibleSectionWithThreeArgs(
      _ScrollAxisRegion section, int start, int end) {
    this.getVisibleSection(section as int, start, end);
  }

  /// Returns the clipping area for the specified visible lines. Only if `VisibleLineInfo.IsClippedOrigin` is true for
  /// first line or if `VisibleLineInfo.IsClippedCorner` is true for last line then the area will be clipped. Otherwise
  /// the whole area from 0 to `RenderSize` is returned.
  ///
  /// * firstLine The first visible line.
  /// * lastLine The last visible line.
  /// Returns the clipping area for the specified visible lines.
  _DoubleSpan getBorderRangeClipPoints(
      _VisibleLineInfo firstLine, _VisibleLineInfo lastLine) {
    if (!firstLine.isClippedOrigin && !lastLine.isClippedCorner) {
      return _DoubleSpan(0, this.renderSize);
    }

    if (firstLine.isClippedOrigin && !lastLine.isClippedCorner) {
      return firstLine.clippedOrigin < this.renderSize
          ? _DoubleSpan(firstLine.clippedOrigin, this.renderSize)
          : _DoubleSpan(this.renderSize, firstLine.clippedOrigin);
    }

    if (!firstLine.isClippedOrigin && lastLine.isClippedCorner) {
      return _DoubleSpan(0, lastLine.clippedCorner);
    }

    return _DoubleSpan(firstLine.clippedOrigin, lastLine.clippedCorner);
  }

  /// Gets the line near the given corner point. Use this method for hit-testing row or
  /// column lines for resizing cells.
  ///
  /// * point The point.
  /// * hitTestPrecision The hit test precision in points.
  /// * isRightToLeft The boolean value used to calculate visible columns in right to left mode.
  /// Returns the visible line.
  _VisibleLineInfo getLineNearCorner(
      double point, double hitTestPrecision, _CornerSide side,
      [bool isRightToLeft = false]) {
    return getLineNearCorner(
        point, hitTestPrecision, _CornerSide.both, isRightToLeft);
  }

  /// Gets the line near the given corner point. Use this method for hit-testing row or
  /// column lines for resizing cells.
  ///
  /// * point The point.
  /// * hitTestPrecision The hit test precision in points.
  /// * side The hit test corner.
  /// * isRightToLeft The boolean value indicates the right to left mode.
  /// Returns the visible line.
  _VisibleLineInfo getLineNearCornerWithFourArgs(
      double point, double hitTestPrecision, _CornerSide side,
      [bool isRightToLeft = false]) {
    if (!isRightToLeft) {
      _VisibleLinesCollection lines = this.getVisibleLines();
      _VisibleLineInfo visibleLine = lines.getVisibleLineAtPoint(point);
      if (visibleLine != null) {
        double d;

        // Close to
        for (int n = max(0, visibleLine.visibleIndex); n < lines.length; n++) {
          visibleLine = lines[n];

          d = visibleLine.clippedOrigin - point;

          if ((d > hitTestPrecision) ||
              (d > 0 &&
                  (side == _CornerSide.right || side == _CornerSide.bottom))) {
            return null;
          } else if ((d).abs() <= hitTestPrecision &&
              side != _CornerSide.left) {
            if (visibleLine.visibleIndex == 0) {
              return null;
            } else {
              return lines[visibleLine.visibleIndex - 1];
            }
          } else if (visibleLine.size - (d).abs() <= hitTestPrecision &&
              side == _CornerSide.left) {
            return lines[visibleLine.visibleIndex];
          }
        }

        // last line - check corner instead of origin.
        d = visibleLine.clippedCorner - point;
        if ((d).abs() <= hitTestPrecision) {
          return lines[visibleLine.visibleIndex];
        }
      }
    } else {
      _VisibleLinesCollection lines = getVisibleLines(true);
      _VisibleLineInfo visibleLine =
          this.getVisibleLineAtPoint(point, false, isRightToLeft);

      if (visibleLine != null) {
        double d;
        // Close to
        for (int n = max(0, visibleLine.visibleIndex); n < lines.length; n++) {
          visibleLine = lines[n];

          d = point - visibleLine.clippedOrigin;

          if ((d > hitTestPrecision &&
                  d < visibleLine.size - hitTestPrecision) ||
              (d > 0 &&
                  d < visibleLine.size - hitTestPrecision &&
                  (side == _CornerSide.right || side == _CornerSide.bottom)) ||
              (d < 0 && (side == _CornerSide.left))) return null;

          if (d > visibleLine.size - hitTestPrecision &&
              visibleLine.visibleIndex > 0)
            return lines[visibleLine.visibleIndex - 1];

          if ((d).abs() <= hitTestPrecision)
            return lines[visibleLine.visibleIndex];

          if (((d).abs() + hitTestPrecision) >= visibleLine.clippedSize)
            return visibleLine;
        }
        // last line - check corner instead of origin.
        d = visibleLine.clippedCorner - point;
        if ((d).abs() <= hitTestPrecision)
          return lines[visibleLine.visibleIndex];
      }
    }
    return null;
  }

  /// Returns points for given absolute line indexes.
  ///
  /// * firstIndex The first index.
  /// * lastIndex The last index.
  /// * allowAdjust if set to true return the first visible line if firstIndex
  /// is above viewable area or return last visible line if lastIndex is after viewable area
  /// (works also for header and footer).
  ///
  /// * firstVisible if set to true indicates the line with index
  /// firstIndex is visible in viewable area.
  /// * lastVisible if set to true indicates the line with index
  /// lastIndex is visible in viewable area..
  /// * firstLine The first line or null if allowAdjust is false and line
  /// is not in viewable area.
  /// * lastLine The last line or null if allowAdjust is false and line
  /// is not in viewable area.
  List getLinesAndVisibility(
      int firstIndex,
      int lastIndex,
      bool allowAdjust,
      bool firstVisible,
      bool lastVisible,
      _VisibleLineInfo firstLine,
      _VisibleLineInfo lastLine) {
    _VisibleLinesCollection visibleLines = this.getVisibleLines();

    if (firstIndex < 0) {
      firstIndex = 0;
    }

    // Invalid Line
    if (firstIndex < 0 || firstIndex >= this.lineCount) {
      firstVisible = false;
      firstLine = null;
    }

    // Header
    else if (firstIndex < this.headerLineCount) {
      firstVisible = true;
      firstLine = visibleLines.getVisibleLineNearLineIndex(firstIndex);
    }

    // Footer
    else if (firstIndex >= this.firstFooterLineIndex) {
      firstVisible = true;
      firstLine = visibleLines.getVisibleLineNearLineIndex(firstIndex);
    }

    // After Header and Before Scroll Position
    else if (firstIndex < this.scrollLineIndex) {
      firstVisible = false;
      firstLine = allowAdjust
          ? this.getVisibleLineAtLineIndex(this.scrollLineIndex)
          : null;
    }

    // After Scroll Position and Before Footer
    else if (firstIndex > this.lastBodyVisibleLineIndex) {
      firstVisible = false;
      if (allowAdjust && this.isFooterVisible) {
        firstLine = visibleLines[visibleLines.firstFooterVisibleIndex];
      } else {
        firstLine = null;
      }
    }
    // Regular line (Body) - Visible and not a Header or Footer.
    else {
      firstVisible = true;
      firstLine = visibleLines.getVisibleLineNearLineIndex(firstIndex);
    }

    if (lastIndex >= this.lineCount) {
      lastIndex = this.lineCount - 1;
    }

    // Invalid Line
    if (lastIndex < 0 || lastIndex >= this.lineCount) {
      lastVisible = false;
      lastLine = null;
    }

    // Header
    else if (lastIndex < this.headerLineCount) {
      lastVisible = true;
      lastLine = visibleLines.getVisibleLineNearLineIndex(lastIndex);
    }

    // Footer
    else if (lastIndex >= this.firstFooterLineIndex) {
      lastVisible = true;
      lastLine = visibleLines.getVisibleLineNearLineIndex(lastIndex);
    }

    // After Header and Before Scroll Position
    else if (lastIndex < this.scrollLineIndex) {
      lastVisible = false;
      if (!firstVisible &&
          firstIndex <
              this.scrollLineIndex) // maybe - in case you want right border to look through ...: && lastIndex+1 < ScrollLineIndex)
      {
        firstLine = null;
        lastLine = null;
      } else {
        if (allowAdjust && this.headerLineCount > 0) {
          lastLine = visibleLines[visibleLines.firstBodyVisibleIndex - 1];
        } else {
          lastLine = null;
        }
      }
    }

    // After Scroll Position and Before Footer
    else if (lastIndex > this.lastBodyVisibleLineIndex) {
      lastVisible = false;
      if (!firstVisible && firstIndex > this.lastBodyVisibleLineIndex) {
        firstLine = null;
        lastLine = null;
      } else {
        lastLine = allowAdjust ? this.lastBodyVisibleLine : null;
      }
    }

    // Regular line (Body) - Visible and not a Header or Footer.
    else {
      lastVisible = true;
      lastLine = visibleLines.getVisibleLineNearLineIndex(lastIndex);
    }
    return [firstVisible, lastVisible, firstLine, lastLine];
  }

  /// Gets the visible lines clip points (clipped origin of first line and clipped
  /// corner of last line). If both lines are above or below viewable area an empty
  /// span is returned. If lines are both above and below viewable are then the
  /// range for all viewable lines is returned.
  ///
  /// * firstIndex The first index.
  /// * lastIndex The last index.
  ///
  /// The visible lines clip points (clipped origin of first line and clipped
  /// corner of last line).

  _DoubleSpan getVisibleLinesClipPoints(int firstIndex, int lastIndex) {
    bool firstVisible, lastVisible;
    _VisibleLineInfo firstLine, lastLine;

    this.getLinesAndVisibility(firstIndex, lastIndex, true, firstVisible,
        lastVisible, firstLine, lastLine);
    if (firstLine == null || lastLine == null) {
      return _DoubleSpan.empty();
    }

    return _DoubleSpan(firstLine.clippedOrigin, lastLine.clippedCorner);
  }

  /// Gets a value indicating whether footer lines are visible.
  ///
  /// @value
  /// true if footer lines are visible, otherwise false.
  bool get isFooterVisible {
    _VisibleLinesCollection visibleLines = this.getVisibleLines();
    return visibleLines.firstFooterVisibleIndex < visibleLines.length;
  }

  /// Gets the index of the first footer line.
  ///
  /// Returns the index of the first footer line.
  int get firstFooterLineIndex => this.lineCount - this.footerLineCount;

  /// Gets the last visible line.
  ///
  /// Returns the last visible line.
  _VisibleLineInfo get lastBodyVisibleLine {
    _VisibleLinesCollection visibleLines = this.getVisibleLines();
    if (visibleLines.length == 0 ||
        visibleLines.lastBodyVisibleIndex > visibleLines.length) {
      return null;
    }

    return visibleLines[visibleLines.lastBodyVisibleIndex];
  }

  /// Gets the index of the last visible line.
  ///
  /// Returns the index of the last visible line.
  int get lastBodyVisibleLineIndex {
    _VisibleLinesCollection visibleLines = this.getVisibleLines();
    if (visibleLines.length == 0 ||
        visibleLines.lastBodyVisibleIndex > visibleLines.length) {
      return -1;
    }

    return visibleLines[visibleLines.lastBodyVisibleIndex].lineIndex;
  }

  /// Gets the clip points for a region.
  ///
  /// * region The region for which the clip points is to be obtained.
  /// * isRightToLeft The boolean value used to calculate visible columns in right to left mode.
  /// Returns the clip points for a region.
  _DoubleSpan getClipPoints(_ScrollAxisRegion region,
      {bool isRightToLeft = false}) {
    _VisibleLinesCollection lines = this.getVisibleLines();
    int start, end;
    start = this.getVisibleSection(region as int, start, end)[0];
    end = this.getVisibleSection(region as int, start, end)[1];
    if (start == end &&
        region == _ScrollAxisRegion.body &&
        lines[end].clippedOrigin > 0 &&
        lines[start].clippedCorner > 0) {
      return _DoubleSpan(lines[start].clippedOrigin, lines[end].clippedCorner);
    }

    if (end < start) {
      return _DoubleSpan.empty();
    }
    if (isRightToLeft)
      return _DoubleSpan(lines[start].clippedCorner, lines[end].clippedOrigin);
    else
      return _DoubleSpan(lines[start].clippedOrigin, lines[end].clippedCorner);
  }

  /// Returns an array with 3 ranges indicating the first and last point for the given lines in each region.
  ///
  /// * first The index of the first line.
  /// * last The index of the last line.
  /// * allowEstimatesForOutOfViewLines if set to true allow estimates for out of view lines.
  /// Returns An array with 3 ranges indicating the first and last point for the given lines in each region.
  List<_DoubleSpan> rangeToRegionPoints(
      int first, int last, bool allowEstimatesForOutOfViewLines);

  /// Gets the first and last point for the given lines in a region.
  ///
  /// * region The region.
  /// * first The index of the first line.
  /// * last The index of the last line.
  /// * allowEstimatesForOutOfViewLines if set to true allow estimates for out of view lines.
  /// Returns the first and last point for the given lines in a region.
  _DoubleSpan rangeToPoints(_ScrollAxisRegion region, int first, int last,
      bool allowEstimatesForOutOfViewLines);

  /// Raises the `Changed` event.
  ///
  /// * action scroll action
  void raiseChanged(_ScrollChangedAction action) {
    if (onChangedCallback != null) {
      var scrollChangedArgs = _ScrollChangedArgs.fromArgs(action);
      onChangedCallback(scrollChangedArgs);
    }
  }

  /// Determines the line one page down from the given line.
  ///
  /// * lineIndex The index of the current line.
  /// Returns the line index of the line one page down from the given line.
  int getNextPage(int lineIndex) {
    double extent = 0;
    double pageExtent = this.scrollPageSize - this.getLineSize(lineIndex);
    int count = this.lineCount;
    while (extent < pageExtent && lineIndex < count && lineIndex != -1) {
      var index = this.getNextScrollLineIndex(lineIndex);
      if (index < 0) {
        break;
      }

      lineIndex = index;
      extent += this.getLineSize(index);
    }

    return lineIndex;
  }

  /// Determines the line one page up from the given line.
  ///
  /// * lineIndex The index of the current line.
  /// Returns the line index of the line one page up from the given line.
  int getPreviousPage(int lineIndex) {
    double extent = 0;
    double pageExtent = this.scrollPageSize - this.getLineSize(lineIndex);
    while (extent < pageExtent && lineIndex > 0) {
      lineIndex = this.getPreviousScrollLineIndex(lineIndex);
      extent += this.getLineSize(lineIndex);
    }

    return lineIndex;
  }
}

/// Corner side enumeration.
enum _CornerSide {
  /// Includes both Left and right side or top and bottom side.
  both,

  /// Left side alone.
  left,

  /// Right side alone.
  right,

  /// Bottom side alone.
  bottom
}
